/***************************Copyright-DO-NOT-REMOVE-THIS-LINE**
 * CONDOR Copyright Notice
 *
 * See LICENSE.TXT for additional notices and disclaimers.
 *
 * Copyright (c)1990-1998 CONDOR Team, Computer Sciences Department, 
 * University of Wisconsin-Madison, Madison, WI.  All Rights Reserved.  
 * No use of the CONDOR Software Program Source Code is authorized 
 * without the express consent of the CONDOR Team.  For more information 
 * contact: CONDOR Team, Attention: Professor Miron Livny, 
 * 7367 Computer Sciences, 1210 W. Dayton St., Madison, WI 53706-1685, 
 * (608) 262-0856 or miron@cs.wisc.edu.
 *
 * U.S. Government Rights Restrictions: Use, duplication, or disclosure 
 * by the U.S. Government is subject to restrictions as set forth in 
 * subparagraph (c)(1)(ii) of The Rights in Technical Data and Computer 
 * Software clause at DFARS 252.227-7013 or subparagraphs (c)(1) and 
 * (2) of Commercial Computer Software-Restricted Rights at 48 CFR 
 * 52.227-19, as applicable, CONDOR Team, Attention: Professor Miron 
 * Livny, 7367 Computer Sciences, 1210 W. Dayton St., Madison, 
 * WI 53706-1685, (608) 262-0856 or miron@cs.wisc.edu.
****************************Copyright-DO-NOT-REMOVE-THIS-LINE**/

#include "condor_common.h"
#include "condor_debug.h"
#include "string_list.h"
#include "condor_classad.h"
#include "condor_io.h"
#include "file_transfer.h"
#include "condor_attributes.h"
#include "condor_commands.h"
#include "basename.h"
#include "directory.h"
#include "condor_config.h"
#include "condor_ckpt_name.h"
#include "condor_string.h"
#include "util_lib_proto.h"
#include "daemon.h"
#include "daemon_types.h"

#define COMMIT_FILENAME ".ccommit.con"

// Filenames are case insensitive on Win32, but case sensitive on Unix
#ifdef WIN32
#	define file_strcmp _stricmp
#	define file_contains contains_anycase
#else
#	define file_strcmp strcmp
#	define file_contains contains
#endif

TranskeyHashTable* FileTransfer::TranskeyTable = NULL;
TransThreadHashTable *FileTransfer::TransThreadTable = NULL;
int FileTransfer::CommandsRegistered = FALSE;
int FileTransfer::SequenceNum = 0;
int FileTransfer::ReaperId = -1;


// Utils from the util_lib that aren't prototyped
extern "C" {
	int		get_random_int();
	int		set_seed( int );
}

struct upload_info {
	FileTransfer *myobj;
};

struct download_info {
	FileTransfer *myobj;
};

// Hash function for pid table.
static int compute_transkey_hash(const MyString &key, int numBuckets) 
{
	return ( key.Hash() % numBuckets );
}

static int compute_transthread_hash(const int &pid, int numBuckets) 
{
	return ( pid % numBuckets );
}

FileTransfer::FileTransfer()
{
	Iwd = NULL;
	InputFiles = NULL;
	OutputFiles = NULL;
	IntermediateFiles = NULL;
	SpooledIntermediateFiles = NULL;
	FilesToSend = NULL;
	ExecFile = NULL;
	TransSock = NULL;
	TransKey = NULL;
	SpoolSpace = NULL;
	TmpSpoolSpace = NULL;
	user_supplied_key = FALSE;
	upload_changed_files = false;
	last_download_time = 0;
	ActiveTransferTid = -1;
	TransferStart = 0;
	ClientCallback = 0;
	TransferPipe[0] = TransferPipe[1] = -1;
	bytesSent = 0.0;
	bytesRcvd = 0.0;
	m_final_transfer_flag = FALSE;
#ifdef WIN32
	perm_obj = NULL;
#endif
	desired_priv_state = PRIV_UNKNOWN;
	want_priv_change = false;
	did_init = false;
}

FileTransfer::~FileTransfer()
{
	if (ActiveTransferTid >= 0) {
		dprintf(D_ALWAYS, "FileTransfer object destructor called during "
				"active transfer.  Cancelling transfer.\n");
		daemonCore->Kill_Thread(ActiveTransferTid);
		TransThreadTable->remove(ActiveTransferTid);
		ActiveTransferTid = -1;
	}
	if (TransferPipe[0] >= 0) close(TransferPipe[0]);
	if (TransferPipe[1] >= 0) close(TransferPipe[1]);
	if (Iwd) free(Iwd);
	if (ExecFile) free(ExecFile);
	if (SpoolSpace) free(SpoolSpace);
	if (TmpSpoolSpace) free(TmpSpoolSpace);
	if (InputFiles) delete InputFiles;
	if (OutputFiles) delete OutputFiles;
	if (IntermediateFiles) delete IntermediateFiles;
	if (SpooledIntermediateFiles) delete SpooledIntermediateFiles;
	// Note: do _not_ delete FileToSend!  It points to OutputFile or Intermediate.
	if (TransSock) free(TransSock);
	if (TransKey) {
		// remove our key from the hash table
		if ( TranskeyTable ) {
			MyString key(TransKey);
			TranskeyTable->remove(key);
			if ( TranskeyTable->getNumElements() == 0 ) {
				// if hash table is empty, delete table as well
				delete TranskeyTable;
				TranskeyTable = NULL;
				// delete our Thread table as well, which should be empty now
				delete TransThreadTable;
				TransThreadTable = NULL;
			}
		}		
		// and free the key as well
		free(TransKey);
	}	
#ifdef WIN32
	if (perm_obj) delete perm_obj;
#endif

}

int
FileTransfer::Init( ClassAd *Ad, bool want_check_perms, priv_state priv ) 
{
	char buf[ATTRLIST_MAX_EXPRESSION];

	if( did_init ) {
			// no need to except, just quietly return success
		return 1;
	}

	dprintf(D_FULLDEBUG,"entering FileTransfer::Init\n");

	desired_priv_state = priv;
    if ( priv == PRIV_UNKNOWN ) {
        want_priv_change = false;
    } else {
        want_priv_change = true;
    }

	if (Iwd) {
		EXCEPT("FileTransfer::Init called twice!");
	}

	if (ActiveTransferTid >= 0) {
		EXCEPT("FileTransfer::Init called during active transfer!");
	}

	if (!TranskeyTable) {
		// initialize our hashtable
		if (!(TranskeyTable = new TranskeyHashTable(7, compute_transkey_hash)))
		{
			// failed to allocate our hashtable ?!?!
			return 0;
		}
		
	}

	if (!TransThreadTable) {
		// initialize our thread hashtable
		if (!(TransThreadTable =
			  new TransThreadHashTable(7, compute_transthread_hash))) {
			// failed to allocate our hashtable ?!?!
			return 0;
		}
	}


	// Note: we must register commands here instead of our constructor 
	// to ensure that daemonCore object has been initialized before we 
	// call Register_Command.
	if ( !CommandsRegistered ) {
		CommandsRegistered = TRUE;
		daemonCore->Register_Command(FILETRANS_UPLOAD,"FILETRANS_UPLOAD",
				(CommandHandler)&FileTransfer::HandleCommands,
				"FileTransfer::HandleCommands()",NULL,WRITE);
		daemonCore->Register_Command(FILETRANS_DOWNLOAD,"FILETRANS_DOWNLOAD",
				(CommandHandler)&FileTransfer::HandleCommands,
				"FileTransfer::HandleCommands()",NULL,WRITE);
		ReaperId = daemonCore->Register_Reaper("FileTransfer::Reaper",
							(ReaperHandler)&FileTransfer::Reaper,
							"FileTransfer::Reaper()",NULL);
		if (ReaperId == 1) {
			EXCEPT("FileTransfer::Reaper() can not be the default reaper!\n");
		}

		// we also need to initialize the random number generator.  since
		// this only has to happen once, and we will only be in this section
		// of the code once (because the CommandsRegistered flag is static),
		// initialize the C++ random number generator here as well.
		set_seed( time(NULL) + (unsigned)this + (unsigned)Ad );
	}

	if (Ad->EvaluateAttrString(ATTR_TRANSFER_KEY, buf, ATTRLIST_MAX_EXPRESSION) != 1) {
		char tempbuf[80];
		// classad did not already have a TRANSFER_KEY, so
		// generate a new one.  It must be unique and not guessable.
		sprintf(tempbuf,"%x#%x%x%x",++SequenceNum,(unsigned)time(NULL),
			get_random_int(),get_random_int());
		TransKey = strdup(tempbuf);
		user_supplied_key = FALSE;
		Ad->InsertAttr(ATTR_TRANSFER_KEY,TransKey);

		// since we generated the key, it is only good on our socket.
		// so update TRANSFER_SOCK now as well.
		char *mysocket = daemonCore->InfoCommandSinfulString();
		ASSERT(mysocket);
		Ad->InsertAttr(ATTR_TRANSFER_SOCKET,mysocket);
	} else {
		// Here the ad we were given already has a Transfer Key.
		TransKey = strdup(buf);
		user_supplied_key = TRUE;
	}

	// user must give us an initial working directory.
	if (!Ad->EvaluateAttrString(ATTR_JOB_IWD, buf, ATTRLIST_MAX_EXPRESSION)) {
		return 0;
	}
	Iwd = strdup(buf);

	// if the user want us to check file permissions, pull out the Owner
	// from the classad and instantiate a perm object.
	if ( want_check_perms ) {
		if (!Ad->EvaluateAttrString(ATTR_OWNER, buf, ATTRLIST_MAX_EXPRESSION)) {
			// no owner specified in ad
			return 0;		
		}
#ifdef WIN32
		// lookup the domain
		char ntdomain[80];
		char *p_ntdomain = ntdomain;
		if (!Ad->EvaluateAttrString(ATTR_NT_DOMAIN, ntdomain)) {
			// no nt domain specified in the ad; assume local account
			p_ntdomain = NULL;
		}
		perm_obj = new perm();
		if ( !perm_obj->init(buf,p_ntdomain) ) {
			// could not find the owner on this system; perm object
			// already did a dprintf so we don't have to.
			delete perm_obj;
			perm_obj = NULL;
			return 0;
		} 
#endif
	}

	// Set InputFiles to be ATTR_TRANSFER_INPUT_FILES plus 
	// ATTR_JOB_INPUT, ATTR_JOB_CMD.
	if (Ad->EvaluateAttrString(ATTR_TRANSFER_INPUT_FILES, buf, ATTRLIST_MAX_EXPRESSION)) {
		InputFiles = new StringList(buf,",");
	} else {
		InputFiles = new StringList(NULL,",");
	}
	if (Ad->EvaluateAttrString(ATTR_JOB_INPUT,buf, ATTRLIST_MAX_EXPRESSION)) {
		// only add to list if not NULL_FILE (i.e. /dev/null)
		if ( file_strcmp(buf,NULL_FILE) != 0 ) {			
			if ( !InputFiles->file_contains(buf) )
				InputFiles->append(buf);			
		}
	}
	if ( IsServer() && Ad->EvaluateAttrString(ATTR_JOB_CMD, buf, ATTRLIST_MAX_EXPRESSION) ) {
		// stash the executable name for comparison later, so
		// we know that this file should be called condor_exec on the
		// client machine.  if an executable for this cluster exists
		// in the spool dir, use it instead.
		char *source;
		char *Spool;
		int Cluster = 0;
		int Proc = 0;

		Ad->EvaluateAttrInt(ATTR_CLUSTER_ID, Cluster);
		Ad->EvaluateAttrInt(ATTR_PROC_ID, Proc);
		Spool = param("SPOOL");
		if ( Spool ) {

			SpoolSpace = strdup( gen_ckpt_name(Spool,Cluster,Proc,0) );
			TmpSpoolSpace = (char*)malloc( strlen(SpoolSpace) + 10 );
			sprintf(TmpSpoolSpace,"%s.tmp",SpoolSpace);

			priv_state old_priv = set_condor_priv();
			
			if( (mkdir(SpoolSpace,0777) < 0) ) {
				// mkdir can return 17 = EEXIST (dirname exists) or 2 = ENOENT (path not found)
				dprintf( D_FULLDEBUG, 
						 "FileTransfer::Init(): mkdir(%s) failed, errno: %d\n",
						 SpoolSpace, errno );
			}
			if( (mkdir(TmpSpoolSpace,0777) < 0) ) {
				dprintf( D_FULLDEBUG, 
						 "FileTransfer::Init(): mkdir(%s) failed, errno: %d\n",
						 TmpSpoolSpace, errno );
			}
			source = gen_ckpt_name(Spool,Cluster,ICKPT,0);
			free(Spool);
			if ( access(source,F_OK | X_OK) >= 0 ) {
				// we can access an executable in the spool dir
				ExecFile = strdup(source);
			}
			set_priv( old_priv );

		}
		if ( !ExecFile ) {
			// apparently the executable is not in the spool dir.
			// so we must make certain the user has permission to read
			// this file; if so, we can record it as the Executable to send.
#ifdef WIN32
			if ( perm_obj && (perm_obj->read_access(buf) != 1) ) {
				// we do _not_ have permission to read this file!!
				dprintf(D_ALWAYS,
					"FileTrans: permission denied reading %s\n",buf);
				return 0;
			}
#endif
			ExecFile = strdup(buf);
		}

		// If we don't already have this on our list of things to transfer, 
		// and we haven't set TRANSFER_EXECTUABLE to false, send it along.
		// If we didn't set TRANSFER_EXECUTABLE, default to true 

		bool xferExec;
		if( !Ad->EvaluateAttrBool(ATTR_TRANSFER_EXECUTABLE, xferExec)) {
			xferExec=true;
		}

		if ( xferExec && !InputFiles->contains(ExecFile) ) {
			InputFiles->append(ExecFile);	
		}	
	}

	// Set OutputFiles to be ATTR_TRANSFER_OUTPUT_FILES if specified.
	// If not specified, set it to send whatever files have changed.
	// Also add in ATTR_JOB_OUPUT plus ATTR_JOB_ERROR.
	if (Ad->EvaluateAttrString(ATTR_TRANSFER_OUTPUT_FILES, buf, ATTRLIST_MAX_EXPRESSION)) {
		OutputFiles = new StringList(buf,",");
	} else {
		// send back new/changed files after the run
		upload_changed_files = true;
	}
	// and now check stdout/err
	if (Ad->EvaluateAttrString(ATTR_JOB_OUTPUT, buf, ATTRLIST_MAX_EXPRESSION)) {
		// only add to list if not NULL_FILE (i.e. /dev/null)
		if ( file_strcmp(buf,NULL_FILE) != 0 ) {
			if ( OutputFiles ) {
				if ( !OutputFiles->file_contains(buf) )
					OutputFiles->append(buf);
			} else
				OutputFiles = new StringList(buf,",");
		}
	}
	if (Ad->EvaluateAttrString(ATTR_JOB_ERROR, buf, ATTRLIST_MAX_EXPRESSION)) {
		// only add to list if not NULL_FILE (i.e. /dev/null)
		if ( file_strcmp(buf,NULL_FILE) != 0 ) {
			if ( OutputFiles ) {
				if ( !OutputFiles->file_contains(buf) )
					OutputFiles->append(buf);
			} else
				OutputFiles = new StringList(buf,",");
		}
	}


	if (Ad->EvaluateAttrString(ATTR_TRANSFER_SOCKET, buf, ATTRLIST_MAX_EXPRESSION)) {
		return 0;		
	}
	TransSock = strdup(buf);


	// If we are acting as the server side and we are uploading
	// any changed files, make a list of "intermediate" files
	// stored in our spool space (i.e. if transfer_files=ALWAYS).
	// This list is stored in the ClassAd which is sent to the 
	// client side, so that when the client does a final transfer
	// it can send changed files from that run + all intermediate
	// files.  -Todd Tannenbaum <tannenba@cs.wisc.edu> 6/8/01
	buf[0] = '\0';
	if ( IsServer() && upload_changed_files ) {
		CommitFiles();
		MyString filelist;
		const char* current_file = NULL;
		bool print_comma = false;
		Directory spool_space( SpoolSpace, PRIV_CONDOR );
		while ( (current_file=spool_space.Next()) ) {
			if ( print_comma ) {
				filelist += ",";
			} else {
				print_comma = true;
			}
			filelist += current_file;			
		}
		if ( print_comma ) {
			// we know that filelist has at least one entry, so
			// insert it as an attribute into the ClassAd which
			// will get sent to our peer.
			sprintf(buf,"\"%s\"",filelist.Value());
			Ad->InsertAttr( ATTR_TRANSFER_INTERMEDIATE_FILES, buf );
			dprintf(D_FULLDEBUG,"%s\n",buf);
		}
	}
	if ( IsClient() && upload_changed_files ) {
        string s;
        if (Ad->EvaluateAttrString(ATTR_TRANSFER_INTERMEDIATE_FILES, s)) {
            if ( s.length() > 0 ) {
                SpooledIntermediateFiles = strnewp(s.data());
            }
		}
		dprintf(D_FULLDEBUG,"%s=\"%s\"\n",
				ATTR_TRANSFER_INTERMEDIATE_FILES,buf);
	}
	

	// if we are acting as the server side, insert this key 
	// into our hashtable if it is not already there.
	if ( IsServer() ) {
		MyString key(TransKey);
		FileTransfer *transobject;
		if ( TranskeyTable->lookup(key,transobject) < 0 ) {
			// this key is not in our hashtable; insert it
			if ( TranskeyTable->insert(key,this) < 0 ) {
				dprintf(D_ALWAYS,
					"FileTransfer::Init failed to insert key in our table\n");
				return 0;
			}
		} else {
			// this key is already in our hashtable; this is a programmer error!
			EXCEPT("FileTransfer: Duplicate TransferKeys!");
		}
	}

	did_init = true;
	return 1;
}

int
FileTransfer::DownloadFiles(bool blocking)
{
	int ret_value;
	ReliSock sock;

	dprintf(D_FULLDEBUG,"entering FileTransfer::DownloadFiles\n");

	if (ActiveTransferTid >= 0) {
		EXCEPT("FileTransfer::DownloadFiles called during active transfer!\n");
	}

	// Make certain Init() was called.
	if ( Iwd == NULL ) {
		EXCEPT("FileTransfer: Init() never called");
	}

	// This method should only be called on the client side, so if
	// we are the server side, there is a programmer error -- do EXCEPT.
	if ( IsServer() ) {
		EXCEPT("FileTransfer: DownloadFiles called on server side");
	}

	Daemon d(TransSock);

    if ( !sock.connect(TransSock,0) ) {
        EXCEPT("Unable to connect to server %s\n", TransSock);
    }

    d.startCommand(FILETRANS_UPLOAD, &sock, 0);

	sock.encode();

	if ( !sock.code(TransKey) ||
		!sock.end_of_message() ) {
		return 0;
	}

	ret_value = Download(&sock,blocking);

	// If Download was successful (it returns 1 on success) and
	// upload_changed_files is true, then we must record the current
	// time in last_download_time so in UploadFiles we have a timestamp
	// to compare.  If it is a non-blocking download, we do all this
	// in the thread reaper.
	if ( blocking && ret_value == 1 && upload_changed_files ) {
		time(&last_download_time);
		// Now sleep for 1 second.  If we did not do this, then jobs
		// which run real quickly (i.e. less than a second) would not
		// have their output files uploaded.  The real reason we must
		// sleep here is time_t is only at the resolution on 1 second.
		sleep(1);
	}

	return ret_value;
}

int
FileTransfer::UploadFiles(bool blocking, bool final_transfer)
{
    ReliSock sock;

	StringList changed_files(NULL,",");

	dprintf(D_FULLDEBUG,
		"entering FileTransfer::UploadFiles (final_transfer=%d)\n",
		final_transfer ? 1 : 0);

	if (ActiveTransferTid >= 0) {
		EXCEPT("FileTransfer::UpLoadFiles called during active transfer!\n");
	}

	// Make certain Init() was called.
	if ( Iwd == NULL ) {
		EXCEPT("FileTransfer: Init() never called");
	}

	// This method should only be called on the client side, so if
	// we are the server side, there is a programmer error -- do EXCEPT.
	if ( IsServer() ) {
		EXCEPT("FileTransfer: UploadFiles called on server side");
	}

	// set flag saying if this is the last upload (i.e. job exited)
	m_final_transfer_flag = final_transfer ? 1 : 0;

	if (IntermediateFiles) delete(IntermediateFiles);
	IntermediateFiles = NULL;
	FilesToSend = NULL;

	if ( upload_changed_files && last_download_time > 0 ) {
		// Here we will upload only files in the Iwd which have changed
		// since we downloaded last.  We only do this if 
		// upload_changed_files it true, and if last_download_time > 0
		// which means we have already downloaded something.

		// If this is the file transfer, be certain to send back
		// not only the files which have been modified during this run,
		// but also the files which have been modified during
		// previous runs (i.e. the SpooledIntermediateFiles).
		if ( final_transfer && SpooledIntermediateFiles ) {
			IntermediateFiles = 
				new StringList(SpooledIntermediateFiles,",");
			FilesToSend = IntermediateFiles;
		}
	
		Directory* dir;
		if( want_priv_change ) {
			dir = new Directory( Iwd, desired_priv_state );
		} else {
			dir = new Directory( Iwd );
		}
		const char *f;
		while ( (f=dir->Next()) ) {
			// don't send back condor_exec.exe
			if ( file_strcmp(f,CONDOR_EXEC)==MATCH )
				continue;
			if ( file_strcmp(f,"condor_exec.bat")==MATCH )
				continue;

			// for now, skip all subdirectory names until we add
			// subdirectory support into FileTransfer.
			if ( dir->IsDirectory() )
				continue;
							
			// if this file is has been modified since last download,
			// add it to the list of files to transfer.
			if ( dir->GetModifyTime() > last_download_time ) {
				dprintf( D_FULLDEBUG, 
						 "Sending changed file %s, mod=%ld, dow=%ld\n",	
						 f, dir->GetModifyTime(), last_download_time );
				if (!IntermediateFiles) {
					// Initialize it with intermediate files
					// which we already have spooled.  We want to send
					// back these files + any that have changed this time.
					IntermediateFiles = new StringList(NULL,",");
					FilesToSend = IntermediateFiles;
				}
				// now append changed file to list only if not already there 
				if ( IntermediateFiles->file_contains(f) == FALSE ) {
					IntermediateFiles->append(f);
				}
			}
		}
		delete dir;
	} else {
		// Here we want to upload the files listed in OutputFiles.
		FilesToSend = OutputFiles;
	}
	
	// Optimization: files_to_send now contains the files to upload.
	// If files_to_send is NULL, then we have nothing to send, so
	// we can return with SUCCESS immedidately.
	if ( FilesToSend == NULL ) {
		return 1;
	}

	Daemon d(TransSock);

    if ( !sock.connect(TransSock,0) ) {
        EXCEPT("Unable to connect to server %s\n", TransSock);
    }

	d.startCommand(FILETRANS_DOWNLOAD, &sock);

	sock.encode();

	if ( !sock.code(TransKey) ||
		!sock.end_of_message() ) {
		return 0;
	}

	dprintf( D_FULLDEBUG,
			 "FileTransfer::UploadFiles: sent TransKey=%s\n", TransKey );
	int retval = Upload(&sock,blocking);

	return( retval );
}

int
FileTransfer::HandleCommands(Service *, int command, Stream *s)
{
	FileTransfer *transobject;
	char transkeybuf[_POSIX_PATH_MAX];
	char *transkey = transkeybuf;

	dprintf(D_FULLDEBUG,"entering FileTransfer::HandleCommands\n");

	if ( s->type() != Stream::reli_sock ) {
		// the FileTransfer object only works on TCP, not UDP
		return 0;
	}
	ReliSock *sock = (ReliSock *) s;

	// turn off timeouts on sockets, since our peer could get suspended
	// (like in the case of the starter sending files back to the shadow)
	sock->timeout(0);

	if (!sock->code(transkey) ||
		!sock->end_of_message() ) {
		dprintf(D_FULLDEBUG,
			    	"FileTransfer::HandleCommands failed to read transkey\n");
		return 0;
	}
	dprintf(D_FULLDEBUG,
					"FileTransfer::HandleCommands read transkey=%s\n",transkey);

	MyString key(transkey);
	if ( (TranskeyTable == NULL) || 
		 (TranskeyTable->lookup(key,transobject) < 0) ) {		
		// invalid transkey sent; send back 0 for failure
		sock->snd_int(0,1);	// sends a "0" then an end_of_record
		dprintf(D_FULLDEBUG,"transkey is invalid!\n");
		// sleep for 5 seconds to prevent brute-force attack on guessing key
		sleep(5);
		return FALSE;
	}

	switch (command) {
		case FILETRANS_UPLOAD:
			// We want to upload all files listed as InputFiles,
			// but first append to InputFiles any files in our SpoolSpace.
			// And before we do that, call CommitFiles() to finish any
			// previous commit which may have been prematurely aborted.
			{
			transobject->CommitFiles();
			Directory spool_space( transobject->SpoolSpace,
								   PRIV_CONDOR );
			while ( spool_space.Next() ) {
				transobject->InputFiles->append( spool_space.GetFullPath() );
			}
			transobject->FilesToSend = transobject->InputFiles;
			transobject->Upload(sock,true);		// blocking = true for now...
			}
			break;
		case FILETRANS_DOWNLOAD:
			transobject->Download(sock,true);	// blocking = true for now...
			break;
		default:
			dprintf(D_ALWAYS,
				"FileTransfer::HandleCommands: unrecognized command %d\n",
				command);
			return 0;
			break;
	}

	return 1;
	// return KEEP_STREAM;
}


int
FileTransfer::Reaper(Service *, int pid, int exit_status)
{
	FileTransfer *transobject;
	if ( TransThreadTable->lookup(pid,transobject) < 0) {
		dprintf(D_ALWAYS, "unknown pid %d in FileTransfer::Reaper!\n", pid);
		return FALSE;
	}
	transobject->ActiveTransferTid = -1;
	TransThreadTable->remove(pid);

	transobject->Info.duration = time(NULL)-transobject->TransferStart;
	transobject->Info.in_progress = false;
	if (WEXITSTATUS(exit_status) > 0) {
		dprintf(D_ALWAYS, "File transfer completed successfully.\n");
		transobject->Info.success = true;
	} else {
		dprintf(D_ALWAYS, "File transfer failed (status=%d).\n",exit_status);
		transobject->Info.success = false;
	}

	read(transobject->TransferPipe[0], (char *)&transobject->Info.bytes,
		 sizeof(int));
	close(transobject->TransferPipe[0]);
	close(transobject->TransferPipe[1]);
	transobject->TransferPipe[0] = transobject->TransferPipe[1] = -1;

	// If Download was successful (it returns 1 on success) and
	// upload_changed_files is true, then we must record the current
	// time in last_download_time so in UploadFiles we have a timestamp
	// to compare.  If it is a non-blocking download, we do all this
	// in the thread reaper.
	if ( transobject->Info.success && transobject->upload_changed_files &&
		 transobject->IsClient() && transobject->Info.type == DownloadFilesType ) {
		time(&(transobject->last_download_time));
		// Now sleep for 1 second.  If we did not do this, then jobs
		// which run real quickly (i.e. less than a second) would not
		// have their output files uploaded.  The real reason we must
		// sleep here is time_t is only at the resolution on 1 second.
		sleep(1);
	}

	if (transobject->ClientCallback) {
		dprintf(D_FULLDEBUG,
				"Calling client FileTransfer handler function.\n");
		((transobject->ClientCallbackClass)->*(transobject->ClientCallback))(transobject);
	}

	return TRUE;
}

int
FileTransfer::Download(ReliSock *s, bool blocking)
{
	dprintf(D_FULLDEBUG,"entering FileTransfer::Download\n");
	
	if (ActiveTransferTid >= 0) {
		EXCEPT("FileTransfer::Download called during active transfer!\n");
	}

	Info.duration = 0;
	Info.type = DownloadFilesType;
	Info.success = true;
	Info.in_progress = true;
	TransferStart = time(NULL);

	if (blocking) {

		Info.bytes = DoDownload((ReliSock *)s);
		Info.duration = time(NULL)-TransferStart;
		Info.success = (Info.bytes >= 0);
		Info.in_progress = false;
		return Info.success;

	} else {

		// make a pipe to communicate with our thread
		if (pipe(TransferPipe) < 0) {
			dprintf(D_ALWAYS, "pipe failed with errno %d in "
					"FileTransfer::Upload\n", errno);
			return FALSE;
		}

		download_info *info = (download_info *)malloc(sizeof(download_info));
		info->myobj = this;
		ActiveTransferTid = daemonCore->
			Create_Thread((ThreadStartFunc)&FileTransfer::DownloadThread,
						  (void *)info, s, ReaperId);
		if (ActiveTransferTid == FALSE) {
			dprintf(D_ALWAYS,
					"Failed to create FileTransfer DownloadThread!\n");
			ActiveTransferTid = -1;
			free(info);
			return FALSE;
		}
		// daemonCore will free(info) when the thread exits
		TransThreadTable->insert(ActiveTransferTid, this);

	}
	
	return 1;
}

int
FileTransfer::DownloadThread(void *arg, Stream *s)
{
	dprintf(D_FULLDEBUG,"entering FileTransfer::DownloadThread\n");
	FileTransfer * myobj = ((download_info *)arg)->myobj;
	int total_bytes = myobj->DoDownload((ReliSock *)s);
	write(myobj->TransferPipe[1],	// write-end of pipe
				(char *)&total_bytes, sizeof(int));
	return (total_bytes >= 0);
}


/*
  Define a macro to restore our priv state (if needed) and return.  We
  do this so we don't leak priv states in functions where we need to
  be in our desired state.
*/

#define return_and_resetpriv(i)                     \
    if( saved_priv != PRIV_UNKNOWN )                \
        _set_priv(saved_priv,__FILE__,__LINE__,1);  \
    return i;


int
FileTransfer::DoDownload(ReliSock *s)
{
	int reply, bytes, total_bytes = 0;
	char filename[_POSIX_PATH_MAX];
	char* p_filename = filename;
	char fullname[_POSIX_PATH_MAX];
	int final_transfer;

	priv_state saved_priv = PRIV_UNKNOWN;

	// we want to tell get_file() to perform an fsync (i.e. flush to disk)
	// the files we download if we are the client & we will need to upload
	// the changed files later on.  why do we need this fsync, you ask?  	
	// because we figure out which files have changed by looking at the file's
	// modification time.  on some operating systems, the file modification time
	// is _not_ when it was really modified, but when the modifications are actually
	// commited to disk.  thus we must fsync in order to make certain we do not think
	// that files like condor_exec.exe have been modified, etc. -Todd <tannenba@cs>
	bool want_fsync = ( (IsClient()) && upload_changed_files );

	dprintf(D_FULLDEBUG,"entering FileTransfer::DoDownload sync=%d\n",
					want_fsync ? 1 : 0);

	s->decode();

	// find out if this is the final download.  if so, we put the files
	// into the user's Iwd instead of our SpoolSpace.
	if( !s->code(final_transfer) ) {
		return_and_resetpriv( -1 );
	}
	if( !s->end_of_message() ) {
		return_and_resetpriv( -1 );
	}	
	for (;;) {
		if( !s->code(reply) ) {
			return_and_resetpriv( -1 );
		}
		if( !s->end_of_message() ) {
			return_and_resetpriv( -1 );
		}
		if( !reply ) {
			break;
		}
		if( !s->code(p_filename) ) {
			return_and_resetpriv( -1 );
		}
		if( final_transfer || IsClient() ) {
			sprintf(fullname,"%s%c%s",Iwd,DIR_DELIM_CHAR,filename);
			if( want_priv_change ) {
				saved_priv = set_priv( desired_priv_state );
			}
#ifdef WIN32
			// check for write permission on this file, if we are supposed to check
			if ( perm_obj && (perm_obj->write_access(fullname) != 1) ) {
				// we do _not_ have permission to write this file!!
				dprintf(D_ALWAYS,"DoDownload: Permission denied to write file %s!\n",
					fullname);
				dprintf(D_FULLDEBUG,"DoDownload: exiting at %d\n",__LINE__);
				return_and_resetpriv( -1 );
			}
#endif
		} else {
			sprintf(fullname,"%s%c%s",TmpSpoolSpace,DIR_DELIM_CHAR,filename);
				// We *need* to be in condor_priv for this.
			saved_priv = set_condor_priv();
		}
	
		// On WinNT and apparently, some Unix, too, even doing an
		// fsync on the file does not get rid of the lazy-write
		// behavior which in the modification time being set a few
		// seconds into the future.  Instead of sleeping for 30+
		// seconds here in the starter & thus delaying the start of
		// the job, we call _utime to manually set the modification
		// time of the file we just wrote backwards in time by a few
		// minutes!  MLOP!! Since we are doing this, we may as well
		// not bother to fsync every file.
		if( ((bytes = s->get_file(fullname)) < 0) ) {
			return_and_resetpriv( -1 );
		}
		if ( want_fsync ) {
			struct utimbuf timewrap;

			time_t current_time = time(NULL);
			timewrap.actime = current_time;		// set access time to now
			timewrap.modtime = current_time - 180;	// set modify time to 3 min ago

			utime(fullname,&timewrap);
		}

		if( !s->end_of_message() ) {
			return_and_resetpriv( -1 );
		}
		total_bytes += bytes;
	}

	bytesRcvd += total_bytes;

	if ( !final_transfer && IsServer() ) {
		char buf[ATTRLIST_MAX_EXPRESSION];
		int fd;
		// we just stashed all the files in the TmpSpoolSpace.
		// write out the commit file.

		sprintf(buf,"%s%c%s",TmpSpoolSpace,DIR_DELIM_CHAR,COMMIT_FILENAME);
#if defined(WIN32)
		if ((fd = ::open(buf, O_WRONLY | O_CREAT | O_TRUNC | 
			_O_BINARY | _O_SEQUENTIAL, 0644)) < 0)
#else
		if ((fd = ::open(buf, O_WRONLY | O_CREAT | O_TRUNC, 0644)) < 0)
#endif
		{
			dprintf(D_ALWAYS, 
				"FileTransfer::DoDownload failed to write commit file\n");
			return_and_resetpriv( -1 );
		}

		memset(buf,0,sizeof(buf));
		// copy in list of files to remove here
		::write(fd,buf,1+strlen(buf));
		::close(fd);

		// Now actually perform the commit.
		CommitFiles();
	}

	return_and_resetpriv( total_bytes );
}

void
FileTransfer::CommitFiles()
{
	char buf[_POSIX_PATH_MAX];
	char newbuf[_POSIX_PATH_MAX];
	const char *file;

	if ( IsClient() ) {
		return;
	}

		// We *need* to be condor_priv for all of this
	Directory tmpspool( TmpSpoolSpace, PRIV_CONDOR );
	priv_state saved_priv = set_condor_priv();

	sprintf(buf,"%s%c%s",TmpSpoolSpace,DIR_DELIM_CHAR,COMMIT_FILENAME);
	if ( access(buf,F_OK) >= 0 ) {
		// the commit file exists, so commit the files.
		while ( (file=tmpspool.Next()) ) {
			// don't commit the commit file!
			if ( file_strcmp(file,COMMIT_FILENAME) == MATCH )
				continue;
			sprintf(buf,"%s%c%s",TmpSpoolSpace,DIR_DELIM_CHAR,file);
			sprintf(newbuf,"%s%c%s",SpoolSpace,DIR_DELIM_CHAR,file);
			if ( rotate_file(buf,newbuf) < 0 ) {
				EXCEPT("FileTransfer CommitFiles Failed -- What Now?!?!");
			}
		}
		// TODO: remove files specified in commit file
	}

	// We have now commited the files in tmpspool, if we were supposed to.
	// So now blow away tmpspool.
	tmpspool.Remove_Entire_Directory();
	set_priv( saved_priv );
}

int
FileTransfer::Upload(ReliSock *s, bool blocking)
{
	dprintf(D_FULLDEBUG,"entering FileTransfer::Upload\n");

	if (ActiveTransferTid >= 0) {
		EXCEPT("FileTransfer::Upload called during active transfer!\n");
	}

	Info.duration = 0;
	Info.type = UploadFilesType;
	Info.success = true;
	Info.in_progress = true;
	TransferStart = time(NULL);

	if (blocking) {

		Info.bytes = DoUpload((ReliSock *)s);
		Info.duration = time(NULL)-TransferStart;
		Info.success = (Info.bytes >= 0);
		Info.in_progress = false;
		return Info.success;

	} else {

		// make a pipe to communicate with our thread
		if (pipe(TransferPipe) < 0) {
			dprintf(D_ALWAYS, "pipe failed with errno %d in "
					"FileTransfer::Upload\n", errno);
			return FALSE;
		}

		upload_info *info = (upload_info *)malloc(sizeof(upload_info));
		info->myobj = this;
		ActiveTransferTid = daemonCore->
			Create_Thread((ThreadStartFunc)&FileTransfer::UploadThread,
						  (void *)info, s, ReaperId);
		if (ActiveTransferTid == FALSE) {
			dprintf(D_ALWAYS, "Failed to create FileTransfer UploadThread!\n");
			free(info);
			ActiveTransferTid = -1;
			return FALSE;
		}
		// daemonCore will free(info) when the thread exits
		TransThreadTable->insert(ActiveTransferTid, this);
	}
		
	return 1;
}

int
FileTransfer::UploadThread(void *arg, Stream *s)
{
	dprintf(D_FULLDEBUG,"entering FileTransfer::UploadThread\n");
	FileTransfer * myobj = ((upload_info *)arg)->myobj;
	int total_bytes = myobj->DoUpload((ReliSock *)s);
	write(myobj->TransferPipe[1],	// write end
			(char *)&total_bytes, sizeof(int));
	return (total_bytes >= 0);
}

int
FileTransfer::DoUpload(ReliSock *s)
{
	char *filename;
	char *basefilename;
	char fullname[_POSIX_PATH_MAX];
	int bytes, total_bytes=0;
	bool is_the_executable;
	StringList * filelist = FilesToSend;
	
	dprintf(D_FULLDEBUG,"entering FileTransfer::DoUpload\n");

	priv_state saved_priv = PRIV_UNKNOWN;
	if( want_priv_change ) {
		saved_priv = set_priv( desired_priv_state );
	}

	s->encode();

	// tell the server if this is the final transfer our not.
	// if it is the final transfer, the server places the files
	// into the user's Iwd.  if not, the files go into SpoolSpace.
	if( !s->code(m_final_transfer_flag) ) {
		return_and_resetpriv( -1 );
	}
	if( !s->end_of_message() ) {
		return_and_resetpriv( -1 );
	}
	if ( filelist ) {
		filelist->rewind();
	}

	while ( filelist && (filename=filelist->next()) ) {

		dprintf(D_FULLDEBUG,"DoUpload: send file %s\n",filename);

		if( !s->snd_int(1,FALSE) ) {
			dprintf(D_FULLDEBUG,"DoUpload: exiting at %d\n",__LINE__);
			return_and_resetpriv( -1 );
		}
		if( !s->end_of_message() ) {
			dprintf(D_FULLDEBUG,"DoUpload: exiting at %d\n",__LINE__);
			return_and_resetpriv( -1 );
		}

		if ( ExecFile && ( file_strcmp(ExecFile,filename)==0 ) ) {
			// this file is the job executable
			is_the_executable = true;
			basefilename = (char *)CONDOR_EXEC ;
		} else {
			// this file is _not_ the job executable
			is_the_executable = false;
			basefilename = basename(filename);
		}

		if( !s->code(basefilename) ) {
			dprintf(D_FULLDEBUG,"DoUpload: exiting at %d\n",__LINE__);
			return_and_resetpriv( -1 );
		}

		if( filename[0] != '/' && filename[0] != '\\' && filename[1] != ':' ){
			sprintf(fullname,"%s%c%s",Iwd,DIR_DELIM_CHAR,filename);
		} else {
			strcpy(fullname,filename);
		}

		// check for read permission on this file, if we are supposed to check.
		// do not check the executable, since it is likely sitting in the SPOOL
		// directory.
#ifdef WIN32
		if( perm_obj && !is_the_executable &&
			(perm_obj->read_access(fullname) != 1) ) {
			// we do _not_ have permission to read this file!!
			dprintf(D_ALWAYS,"DoUpload: Permission denied to read file %s!\n",
				fullname);
			dprintf(D_FULLDEBUG,"DoUpload: exiting at %d\n",__LINE__);
			return_and_resetpriv( -1 );
		}
#endif

		if( ((bytes = s->put_file(fullname)) < 0) ) {
			/* if we fail to transfer a file, EXCEPT so the other side can */
			/* try again. SC2000 hackery. _WARNING_ - I think Keller changed */
			/* all of this. -epaulson 11/22/2000 */
			EXCEPT("DoUpload: Failed to send file %s, exiting at %d\n",
				fullname,__LINE__);
			return_and_resetpriv( -1 );
		}

		if( !s->end_of_message() ) {
			dprintf(D_FULLDEBUG,"DoUpload: exiting at %d\n",__LINE__);
			return_and_resetpriv( -1 );
		}

		total_bytes += bytes;
	}

	// tell our peer we have nothing more to send
	s->snd_int(0,TRUE);

	dprintf(D_FULLDEBUG,"DoUpload: exiting at %d\n",__LINE__);

	bytesSent += total_bytes;

	return_and_resetpriv( total_bytes );
}

int
FileTransfer::Suspend()
{
	int result = TRUE;	// return TRUE if there currently is no thread

	if (ActiveTransferTid != -1 ) {
		result = daemonCore->Suspend_Thread(ActiveTransferTid);
	}

	return result;
}

int
FileTransfer::Continue()
{
	int result = TRUE;	// return TRUE if there currently is no thread

	if (ActiveTransferTid != -1 ) {
		result = daemonCore->Continue_Thread(ActiveTransferTid);
	}

	return result;
}


bool
FileTransfer::addOutputFile( const char* filename )
{
	if( ! OutputFiles ) {
		return false;
	}
	if( OutputFiles->contains(filename) ) {
		return true;
	}
	OutputFiles->append( filename );
	return true;
}


