#include "condor_api.h"
#include "condor_classad.h"
#include "classadList.h"
#include "condor_common.h"
#include "condor_config.h"
#include "../condor_daemon_core.V6/condor_daemon_core.h"
#include "daemon.h"
#include "string_list.h"
#include "cluster_server.h"
#include "condor_auth.h"

ClusterServer::ClusterServer()
{
	expireTransactionsPeriod = paramExpireTransactionsPeriod();
	expireReservationsPeriod = paramExpireReservationsPeriod();
	truncateLogsPeriod       = paramTruncateLogsPeriod();
	refreshHostAdsPeriod     = paramRefreshHostAdsPeriod();
	conToCanon				 = NULL;
	hostnamesToManage        = paramHostList();
	hostLogFile              = paramHostLogFile();
	transLogFile             = paramTransLogFile();
	condor_offCmd			 = paramcondor_offCmd();
	condor_onCmd			 = paramcondor_onCmd();
	condor_statusCmd		 = paramcondor_statusCmd();
	
	managedHosts = NULL;
	transactions = NULL;

	obtainHostAds();
	obtainTransactions();

	ClassAd *general;
	if( !(general = transactions->GetClassAd("general")) ) {
		dprintf(D_ALWAYS, "Note: not able to get general ad, starting transid "
				"at 1\n");
		nextTransID = 1;
	} else if( !general->EvaluateAttrInt("NextID", nextTransID) ) {
		dprintf(D_ALWAYS,
				"Note: not able to get last transid, starting at 1\n");
		nextTransID = 1;
	}

	expireReservations();
	expireTransactions();
	truncateLogs();
}


ClusterServer::~ClusterServer()
{
	daemonCore->Cancel_Timer(expireTransactionsTmr);
	daemonCore->Cancel_Timer(expireReservationsTmr);
	daemonCore->Cancel_Timer(truncateLogsTmr);
	daemonCore->Cancel_Timer(refreshHostAdsTmr);

	if( hostnamesToManage ) delete hostnamesToManage;
	if( managedHosts      ) delete managedHosts;
	if( transactions      ) delete transactions;
	if( conToCanon        ) delete conToCanon;

		// these char*'s were given by param
	if( hostLogFile		  ) free(hostLogFile);
	if( transLogFile	  ) free(transLogFile);
	if( condor_offCmd	  ) free(condor_offCmd);
	if( condor_onCmd	  ) free(condor_onCmd);
	if( condor_statusCmd  ) free(condor_statusCmd);
}


void
ClusterServer::registerCmds()
{
	daemonCore->Register_Command(CT_QUERY_TRANSACTION, "query_transaction",
				 (CommandHandlercpp)&ClusterServer::handleQueryTransaction,
				 "query_transaction", this, READ);

	daemonCore->Register_Command(CT_BEGIN_TRANSACTION, "begin_transaction",
				 (CommandHandlercpp)&ClusterServer::handleBeginTransaction,
				 "begin_transaction", this, WRITE);

	daemonCore->Register_Command(CT_END_TRANSACTION, "end_transaction",
				 (CommandHandlercpp)&ClusterServer::handleEndTransaction,
				 "end_transaction", this, WRITE);

	daemonCore->Register_Command(CT_CHECK, "check",
			   	 (CommandHandlercpp)&ClusterServer::handleCheck,
				 "check", this, READ);

	daemonCore->Register_Command(CT_RETURN, "return",
				 (CommandHandlercpp)&ClusterServer::handleReturn,
				 "return", this, WRITE);

	daemonCore->Register_Command(CT_RESERVE, "reserve",
				 (CommandHandlercpp)&ClusterServer::handleReserve,
				 "reserve", this, WRITE);

	daemonCore->Register_Command(CT_FORGET, "forget",
				 (CommandHandlercpp)&ClusterServer::handleForget,
				 "forget", this, WRITE);

	daemonCore->Register_Command(CT_REMEMBER, "remember",
				 (CommandHandlercpp)&ClusterServer::handleRemember,
				 "remember", this, WRITE);
#ifdef KILLTEST
	daemonCore->Register_Command(CT_KILL, "kill",
				 (CommandHandlercpp)&ClusterServer::handleKill,
				 "kill", this, WRITE);
#endif
}


void
ClusterServer::registerTmrs()
{
		// the handles returned by Register_Timer are retained for the benfit
		// of methods like setTruncateLogsPeriod
	expireTransactionsTmr = daemonCore->Register_Timer(
				 expireTransactionsPeriod, expireTransactionsPeriod,
				 (Eventcpp)&ClusterServer::expireTransactions,
				 "expire_transactions", this);

	expireReservationsTmr = daemonCore->Register_Timer(
				 expireReservationsPeriod, expireReservationsPeriod,
				 (Eventcpp)&ClusterServer::expireReservations,
				 "expire_reservations", this);

	truncateLogsTmr = daemonCore->Register_Timer(
				 truncateLogsPeriod, truncateLogsPeriod,
				 (Eventcpp)&ClusterServer::truncateLogs,
				 "truncate_logs", this);

	refreshHostAdsTmr = daemonCore->Register_Timer(
				 refreshHostAdsPeriod, refreshHostAdsPeriod,
				 (Eventcpp)&ClusterServer::refreshHostAds,
				 "refresh_host_ads", this);
}


int
ClusterServer::handleQueryTransaction( int, Stream *s )
{
	dprintf(D_FULLDEBUG, "We got a query_transaction\n");

  	int protocolVersion = negotiateProtocolVersion(s);
  	if( protocolVersion < 0 ) {
  		dprintf(D_ALWAYS, "Error negotiating protocol version\n");
  		return FALSE;
  	}

		// first get the transaction ID
	int transid = receiveTransactionID(s);
	if( transid < 0 ) {
		return FALSE;
	}

	int response;		// what to send back to the client

		// now look up the classad of that ID and see
		// what's up
	ClassAd *currentTransaction;
	char buffer[32];
	sprintf(buffer, "%d", transid);

	if( !(currentTransaction = transactions->GetClassAd(buffer)) ) {
			// this transaction does not exist!
		dprintf(D_ALWAYS, "transaction ID found to be invalid\n");
		response = -1;
	} else if( !currentTransaction->EvaluateAttrInt("State", response) ) {
		dprintf(D_FULLDEBUG, "attribute state not found in transaction");
		response = -1;
	}

	s->encode();
	if( !s->code(response) ) {
		dprintf(D_ALWAYS, "Error sending response to query transaction\n");
	}
	if( !s->end_of_message() ) {
		dprintf(D_ALWAYS, "Error ending message\n");
	}

	return FALSE;
}


int
ClusterServer::handleBeginTransaction( int, Stream *s )
{
	dprintf(D_FULLDEBUG, "We got a begin_transaction\n");

  	int protocolVersion = negotiateProtocolVersion(s);
  	if( protocolVersion < 0 ) {
  		dprintf(D_ALWAYS, "Error negotiating protocol version\n");
  		return FALSE;
  	}

#ifdef NOSECTEST
	const char* fulqualname = "wilbur";
#else
	const char* fulqualname = ((ReliSock*) s)->getFullyQualifiedUser();
#endif
	if( !fulqualname ) {
		dprintf(D_ALWAYS, "Error authenticating client, getFullyQualifiedUser"
				" returned null\n");
		return FALSE;
	}

	char* username = strdup(fulqualname);

		// ugly hack to get just the CN field out of full principle
		// this should be better
	char* userchar = username;
	if( !strncmp(userchar, "/CN=", 4) ) {
		while( strncmp(userchar, "/CN=", 4) && *userchar != 0 ) userchar++;
		if( *userchar != 0 ) username = userchar + 4;
	}
		// get part before '@', e.g., get bethenco from
		// bethenco@your.domain
	userchar = username;
	while( *userchar != '@' && *userchar != 0 ) userchar++;
	*userchar = 0;

	dprintf(D_FULLDEBUG, "client is: %s\n", username);

	ClassAd *generalUpdates = new ClassAd;
	if( !generalUpdates->InsertAttr("NextID", nextTransID + 1) ) {
		dprintf(D_ALWAYS, "Error updating NextID\n");
	}

	if( !transactions->UpdateClassAd("general", generalUpdates) &&
		!transactions->AddClassAd("general", generalUpdates) ) {
			dprintf(D_ALWAYS, "Error updating/adding ClassAd in collection\n");
	}

	ClassAd *newTransaction = new ClassAd;
	if( !newTransaction ) {
		dprintf(D_ALWAYS, "Error creating ClassAd for new transaction\n");
	}
	if( !newTransaction->InsertAttr("Username", string(username)) ) {
		dprintf(D_ALWAYS,
				"Error inserting Username attribute into new ClassAd\n");
	}
	if( !newTransaction->InsertAttr("ID", nextTransID) ) {
		dprintf(D_ALWAYS, "Error inserting ID attribute into new ClassAd\n");
	}
		// they've requested a transaction but we haven't sent them an id yet
	if( !newTransaction->InsertAttr("State", TR_STATE_REQUESTED) ) {
		dprintf(D_ALWAYS,
				"Error inserting State attribute into new ClassAd\n");
	}
		// so far this request includes no action
	if( !newTransaction->InsertAttr("Request", TR_REQ_NOTHING) ) {
		dprintf(D_ALWAYS,
				"Error inserting Request attribute into new ClassAd\n");
	}
		// the time of the request (now)
	if( !newTransaction->Insert("Time", Literal::MakeAbsTime()) ) {
		dprintf(D_ALWAYS,
				"Error inserting Time attribute into new ClassAd\n");
	}

  		// protocol version (we don't need to renegotiate every time)
  	if( !newTransaction->InsertAttr("ClusterProtocolVersion",protocolVersion)){
  		dprintf(D_ALWAYS, "Error inserting ClusterProtocolVersion attribute "
  				"into new ClassAd\n");
  	}

	string adstring;
	pp.Unparse(adstring, newTransaction);
	dprintf(D_FULLDEBUG,
			"Created new transaction: \n%s\n\n", adstring.c_str());

	char buffer[32];
	sprintf(buffer, "%d", nextTransID);
	if( !transactions->AddClassAd(buffer, newTransaction) ) {
		dprintf(D_ALWAYS, "Error adding new ClassAd to collection\n");
	}

	s->encode();
	if( !s->code(nextTransID) ) {
		dprintf(D_ALWAYS, "Error sending transaction ID to client\n");
	}
	if( !s->end_of_message() ) {
		dprintf(D_ALWAYS, "Error closing connection with client\n");
	}

	nextTransID++;

	return FALSE;
}


int
ClusterServer::handleEndTransaction( int, Stream *s )
{
	dprintf(D_FULLDEBUG, "We got an end_transaction\n");

		// first get the transaction ID
	int transid = receiveTransactionID(s);
	if( transid < 0 ) {
		return FALSE;
	}

		// now look up the classad of that ID and see
		// what's up
	ClassAd *currenttransaction;
	char buffer[32];
	sprintf(buffer, "%d", transid);

	if( !(currenttransaction = transactions->GetClassAd(buffer)) ) {
				// this transaction does not exist!
		dprintf(D_ALWAYS, "Error! Invalid transaction ID\n");
		return FALSE;
	}

	ClassAd *updates = new ClassAd;

	if( !updates ) {
		dprintf(D_ALWAYS, "Error, ClassAd not created!\n");
	}

	if( !updates->InsertAttr("State", TR_STATE_CLOSED) ) {
		dprintf(D_ALWAYS, "Error updating State\n");
	}

	if( !transactions->UpdateClassAd(buffer, updates) ) {
		dprintf(D_ALWAYS, "Error updating ClassAd in collection\n");
	}

	string adstring;
	pp.Unparse(adstring, currenttransaction);
	dprintf(D_FULLDEBUG, "Modified transaction: \n%s\n\n", adstring.c_str());

	return FALSE;
}


int
ClusterServer::handleCheck( int, Stream* s )
{
	dprintf(D_FULLDEBUG, "We got a check\n");

  	int protocolVersion = negotiateProtocolVersion(s);
  	if( protocolVersion < 0 ) {
  		dprintf(D_ALWAYS, "Error negotiating protocol version\n");
  		return FALSE;
  	}

		// note: the check command doesn't do anything with transactions
		// because it is not important

		// now get request ad

	ClassAd *requestad;

	if( !(requestad = receiveClassAd(s)) ) {
		return FALSE;
	}

	string adstring;
	pp.Unparse(adstring, requestad);
	dprintf(D_FULLDEBUG, "The request ad is this: \n%s\n\n", adstring.c_str());

		// then we find out how many of our classads match it
	LocalCollectionQuery query;
	query.Bind(managedHosts);
	ExprTree *requestrequirements = requestad->Lookup("Requirements");
	if( !requestrequirements ) {
		cp.ParseExpression("true", requestrequirements);
	}
	if( !query.Query(string("root"), requestrequirements) ) {
		dprintf(D_ALWAYS, "Error making query\n");
	}

	string currentkey;
	int matchcount = 0;
	query.ToFirst();
	if( query.Current(currentkey) ) {
		matchcount++;
		while( query.Next(currentkey) ) {
			matchcount++;
		}
	}

		// send number of matches
	s->encode();
	s->code(matchcount);
	s->end_of_message();

		// if there weren't matches, done
	if( !matchcount ) {
		delete requestad;
		return FALSE;
	}

		// otherwise send the classads
	ClassAd *matchedad;
	for( query.ToFirst(), query.Current(currentkey); !query.IsAfterLast();
		 query.Next(currentkey) ) {
		if( !(matchedad = managedHosts->GetClassAd(currentkey)) ) {
			dprintf(D_ALWAYS,
					"Error, unable to retrieve ClassAd from collection\n");
		} else if( !putClassAd(s, *matchedad ) ) {
			dprintf(D_FULLDEBUG, "put returned false!\n");
		}
	}

	delete requestad;
	return FALSE;
}


int
ClusterServer::handleReturn( int, Stream* s )
{
	return doReturnOrReserve(s, FALSE);
}


int
ClusterServer::handleReserve( int, Stream* s )
{
	return doReturnOrReserve(s, TRUE);
}


int
ClusterServer::handleForget( int, Stream* s )
{
	return doForgetOrRemember(s, TRUE);
}


int
ClusterServer::handleRemember( int, Stream* s )
{
	return doForgetOrRemember(s, FALSE);
}

#ifdef KILLTEST
int
ClusterServer::handleKill( int, Stream * )
{
	daemonCore->Shutdown_Graceful(daemonCore->getpid());
	return FALSE;
}
#endif

int
ClusterServer::expireTransactions()
{
		// get the old transaction records and throw them away
	LocalCollectionQuery query;
	query.Bind(transactions);
	ExprTree *havetobeold;
	ExprTree *othertime = cp.ParseExpression("other.Time");
	Value whatisold;
  
	whatisold.SetRelativeTimeValue(86400);		// one day

	havetobeold = Operation::MakeOperation(
					Operation::LESS_THAN_OP,
					Operation::MakeOperation(Operation::ADDITION_OP,
											 othertime,
											 Literal::MakeLiteral(whatisold)),
					Literal::MakeAbsTime());
  

	if( !query.Query(string("root"), havetobeold) ) {
		dprintf(D_ALWAYS, "Error making query\n");
	}

	string currentkey;
	int matchcount = 0;
	query.ToFirst();
	if( query.Current(currentkey) ) {
		matchcount++;
		if( !transactions->RemoveClassAd(currentkey) ) {
			dprintf(D_ALWAYS, "Error removing old transaction\n");
		}
		while( query.Next(currentkey) ) {
			matchcount++;
			if( !transactions->RemoveClassAd(currentkey) ) {
				dprintf(D_ALWAYS, "Error removing old transaction\n");
			}
		}
	}

	if (matchcount) {
		dprintf(D_ALWAYS, "%d old transactions expired\n", matchcount);
	}

	return FALSE;
}


int
ClusterServer::expireReservations()
{
	LocalCollectionQuery query;
	query.Bind(managedHosts);

	ExprTree *havetobeold = cp.ParseExpression("other.ReservedBy != \"\""
									 "&& other.ReservedUntil < CurrentTime()");

	if( !query.Query(string("root"), havetobeold) ) {
		dprintf(D_ALWAYS, "Error making query\n");
	}

	string currentkey;
	int matchcount = 0;
	ClassAd *matchedad;
	ClassAd *updates;
	query.ToFirst();
		// then turn Condor on/off on the matches
	if( query.Current(currentkey) ) {
		for( query.ToFirst(), query.Current(currentkey);
			 !query.IsAfterLast();
			 query.Next(currentkey) ) {
			matchcount++;
    
			if( !(matchedad = managedHosts->GetClassAd(currentkey)) ) {
				dprintf(D_ALWAYS,
						"Error, unable to retrieve ClassAd from collection\n");
				break;
			}

			updates = new ClassAd;
      
			updates->InsertAttr("ReservedBy", string(""));
    
			string hoststring;
			if( !matchedad->EvaluateAttrString("Machine", hoststring) ) {
				dprintf(D_ALWAYS, "Error, ClassAd has no machine attribute\n");
			}
			string username;
			if( !matchedad->EvaluateAttrString("ReservedBy", username) ) {
				dprintf(D_ALWAYS,
						"Error, ClassAd has no ReservedBy attribute\n");
			}
    
			StringList* hoststodo = new StringList;
			hoststodo->append(hoststring.c_str());
			enactReservation(hoststodo, username, TRUE);
			delete hoststodo;
   
			managedHosts->UpdateClassAd(currentkey, updates);
		}

	}
  
	if (matchcount) {
		dprintf(D_ALWAYS, "%d old reservations expired\n", matchcount);
	}

	return FALSE;
}


int
ClusterServer::truncateLogs()
{
	dprintf(D_ALWAYS, "truncating the logs\n");
	if( !transactions->TruncateLog() ) {
		dprintf(D_ALWAYS, "Error truncating log file\n");
	}
	if( !managedHosts->TruncateLog() ) {
		dprintf(D_ALWAYS, "Error truncating log file\n");
	}
	return FALSE;
}


int
ClusterServer::refreshHostAds()
{
	dprintf(D_ALWAYS, "refreshing ad list\n");
	
		// get list of cluster_server managed hosts out of config
	StringList* newHostnamesToManage;
	if( !(newHostnamesToManage = paramHostList()) ) {
		dprintf(D_ALWAYS, "Unable to get list of hosts to manage.\n");
		return FALSE;
	}
	hostnamesToManage = newHostnamesToManage;
  
	if( !obtainHostAds() ) {
		dprintf(D_ALWAYS, "Can't fill our collection of host ads.\n");
	}

	return FALSE;
}


int
ClusterServer::getExpireTransactionsPeriod()
{
	return expireTransactionsPeriod;
}


void
ClusterServer::setExpireTransactionsPeriod( int newperiod )
{
	expireTransactionsPeriod = newperiod;

	daemonCore->Cancel_Timer(expireTransactionsTmr);
	expireTransactionsTmr = daemonCore->Register_Timer(
				 expireTransactionsPeriod / 2, expireTransactionsPeriod,
				 (Eventcpp)&ClusterServer::expireTransactions,
				 "expire_transactions", this);
}


int
ClusterServer::getExpireReservationsPeriod()
{
	return expireReservationsPeriod;
}


void
ClusterServer::setExpireReservationsPeriod( int newperiod )
{
	expireReservationsPeriod = newperiod;

	daemonCore->Cancel_Timer(expireReservationsTmr);
	expireReservationsTmr = daemonCore->Register_Timer(
				 expireReservationsPeriod / 2, expireReservationsPeriod,
				 (Eventcpp)&ClusterServer::expireReservations,
				 "expire_reservations", this);
}


int
ClusterServer::getTruncateLogsPeriod()
{
	return truncateLogsPeriod;
}


void
ClusterServer::setTruncateLogsPeriod( int newperiod )
{
	truncateLogsPeriod = newperiod;
	
	daemonCore->Cancel_Timer(truncateLogsTmr);
	truncateLogsTmr = daemonCore->Register_Timer(
				 truncateLogsPeriod / 2, truncateLogsPeriod,
				 (Eventcpp)&ClusterServer::truncateLogs,
				 "truncate_logs", this);
}


int
ClusterServer::getRefreshHostAdsPeriod()
{
	return refreshHostAdsPeriod;
}


void
ClusterServer::setRefreshHostAdsPeriod( int newperiod )
{
	refreshHostAdsPeriod = newperiod;
	
	daemonCore->Cancel_Timer(refreshHostAdsTmr);
	refreshHostAdsTmr = daemonCore->Register_Timer(
				 refreshHostAdsPeriod / 2, refreshHostAdsPeriod,
				 (Eventcpp)&ClusterServer::refreshHostAds,
				 "refresh_host_ads", this);
}


StringList*
ClusterServer::paramHostList()
{
		// get list of hosts to manage found in config file
	char* hosts;
	StringList* hostList;

	hosts = param("CLUSTER_HOSTS");
	if( !hosts ) {
		dprintf(D_ALWAYS, "CLUSTER_HOSTS is not defined\n");
		return NULL;
	}
	hostList = new StringList(hosts); 
	if( !hostList ) {
		dprintf(D_ALWAYS, "Problems creating StringList\n");
		return NULL;
	}

	if( conToCanon ) delete conToCanon;
	conToCanon = new HashTable<MyString, MyString>
		(hostList->number() * 2, &MyStringHash, updateDuplicateKeys);

		// now we make the hostnames canonical
	StringList* canonicalHosts = new StringList();
	struct hostent* hostinfo;
	char* host;
	hostList->rewind();
	while( (host = hostList->next()) ) {
		hostinfo = gethostbyname(host);
		if( !hostinfo ) {
			dprintf(D_ALWAYS, "Error, could not resolve hostname `%s' in "
					"CLUSTER_HOSTS: %s\n", host,
					h_errno == HOST_NOT_FOUND ? "unknown host." :
					h_errno == NO_ADDRESS   ? "host has no IP address." :
					h_errno == NO_RECOVERY  ? "name server error." :
					h_errno == TRY_AGAIN    ? "try again." : "unknown error.");
		} else {
			canonicalHosts->insert(hostinfo->h_name);
			conToCanon->insert(MyString(host), MyString(hostinfo->h_name));
		}
	}

	delete hostList;

	return canonicalHosts;
}


char*
ClusterServer::paramHostLogFile()
{
		// find out where the host ad collection should go
	char *hostlog = param("CLUSTER_HOST_LOG");
	if( !hostlog ) {
		dprintf(D_ALWAYS, "CLUSTER_HOST_LOG is not defined\n");
		return NULL;
	}
	return hostlog;
}


char*
ClusterServer::paramTransLogFile()
{
		// find out where the transaction ad collectin should go
	char *translog = param("CLUSTER_TRANSACTION_LOG");
	if( !translog ) {
		dprintf(D_ALWAYS, "CLUSTER_TRANSACTION_LOG is not defined\n");
		return NULL;
	}
	return translog;
}


char*
ClusterServer::paramcondor_offCmd()
{
	// make sure we have access to condor_off
	char* path = param("CONDOR_OFF");
	if( !path ) {
		dprintf(D_ALWAYS, "Error, CONDOR_OFF %s not defined in the "
				"configuration file\n", param("CONDOR_ON") ?
				"is" : "and CONDOR_ON are");
		exit(1);
	} else {
		struct stat result;
		stat(path, &result);
		if( !S_ISREG(result.st_mode) ) {
			dprintf(D_ALWAYS, "Error, CONDOR_OFF has an apparently invalid "
					"setting (%s) in the configuration file\n", path);
			exit(1);
		}
	}
	return path;
}

char*
ClusterServer::paramcondor_onCmd()
{
	// make sure we have access to condor_on
	char* path = param("CONDOR_ON");
	if( !path ) {
		dprintf(D_ALWAYS, "Error, CONDOR_ON %s not defined in the "
				"configuration file\n", param("CONDOR_OFF") ?
				"is" : "and CONDOR_OFF are");
		exit(1);
	} else {
		struct stat result;
		stat(path, &result);
		if( !S_ISREG(result.st_mode) ) {
			dprintf(D_ALWAYS, "Error, CONDOR_ON has an apparently invalid "
					"setting (%s) in the configuration file\n", path);
			exit(1);
		}
	}
	return path;
}

char*
ClusterServer::paramcondor_statusCmd()
{
	// make sure we have access to condor_status
	char* path = param("CONDOR_STATUS");
	if( !path ) {
		dprintf(D_ALWAYS, "Error, CONDOR_STATUS is not defined in the "
				"configuration file\n");
		exit(1);
	} else {
		struct stat result;
		stat(path, &result);
		if( !S_ISREG(result.st_mode) ) {
			dprintf(D_ALWAYS, "Error, CONDOR_STATUS has an apparently invalid "
					"setting (%s) in the configuration file\n", path);
			exit(1);
		}
	}
	return path;
}

int
ClusterServer::paramExpireTransactionsPeriod()
{
	int result;
	char* periodstr = param("CLUSTER_EXPIRE_TRANSACTIONS_PERIOD");
	if( periodstr && sscanf(periodstr, "%d", &result) ) {
		return result;
	} else {
		return 10348;		// about every 3 hours
	}
}

int
ClusterServer::paramExpireReservationsPeriod()
{
	int result;
	char* periodstr = param("CLUSTER_EXPIRE_RESERVATIONS_PERIOD");
	if( periodstr && sscanf(periodstr, "%d", &result) ) {
		return result;
	} else {
		return 29;			// every 29 seconds
	}
}

int
ClusterServer::paramTruncateLogsPeriod()
{
	int result;
	char* periodstr = param("CLUSTER_TRUNCATE_LOGS_PERIOD");
	if( periodstr && sscanf(periodstr, "%d", &result) ) {
		return result;
	} else {
		return 3610;		// about every hour
	}
}

int
ClusterServer::paramRefreshHostAdsPeriod()
{
	int result;
	char* periodstr = param("CLUSTER_REFRESH_HOSTS_PERIOD");
	if( periodstr && sscanf(periodstr, "%d", &result) ) {
		return result;
	} else {
		return 1820;		// about every half hour
	}
}

bool
ClusterServer::obtainTransactions()
{
	if( transactions ) delete transactions;
	transactions = new ClassAdCollection;

	if( !transactions->InitializeFromLog(string(transLogFile), "", "") ) {
		dprintf(D_ALWAYS, "loading transactions from file caused error number "
				"%d: %s\n", CondorErrno, CondorErrMsg.c_str());
		return FALSE;
	} else {
		dprintf(D_ALWAYS, "transactions loaded from file\n");
	}
	return TRUE;
}


bool
ClusterServer::obtainHostAds()
{
	if( managedHosts ) delete managedHosts;
	managedHosts = new ClassAdCollection;

	/*
	  first we get any host ads that may be in our logfile
	*/

	if( !managedHosts->InitializeFromLog(string(hostLogFile), "", "") ) {
		dprintf(D_ALWAYS,
				"loading hosts from file caused error number %d: %s\n",
				CondorErrno, CondorErrMsg.c_str());
	} else {
		dprintf(D_ALWAYS, "hosts loaded from file\n");
	}
	
	/*
	  now we get rid of any ads that are currently in collection but not in
	  hostnamesToManage
	*/

	LocalCollectionQuery query;
	query.Bind(managedHosts);

	ExprTree* hostsconstraint;

	if( hostnamesToManage->number() ) {
	
		ExprTree* host1constraint;        // of the form                 
		ExprTree* host2constraint;        // Machine == "foo.cs.wisc.edu"
		Value     hostvalue;	          // for making string literals
		char*     hostchars;	          // more convenience variables

	//TEMP
	dprintf(D_ALWAYS, "oka\n");

		ExprTree* machineattrref;
		if( !cp.ParseExpression("other.Machine", machineattrref) ) {
			dprintf(D_ALWAYS, "Error! Can't parse other.Machine expression");
			dprintf(D_ALWAYS, "Aborting");
			return FALSE;
		}

	//TEMP
	dprintf(D_ALWAYS, "okc\n");

		hostnamesToManage->rewind();
				// we know at least one host was specified
		hostchars = hostnamesToManage->next();
		hostvalue.SetStringValue(string(hostchars));
		host1constraint =
			Operation::MakeOperation(Operation::NOT_EQUAL_OP, machineattrref,
									 Literal::MakeLiteral(hostvalue));
		hostsconstraint = host1constraint;
		hostchars = hostnamesToManage->next();

		if( hostchars ) {
				// ah, at least two hosts were specified, we need to make
				// the first ||
			hostvalue.SetStringValue(string(hostchars));
			host2constraint =
				 Operation::MakeOperation(Operation::NOT_EQUAL_OP,
										  machineattrref,
										  Literal::MakeLiteral(hostvalue));
			hostsconstraint =
				 Operation::MakeOperation(Operation::LOGICAL_AND_OP,
										  host1constraint,
										  host2constraint);
			hostchars = hostnamesToManage->next();
			while( hostchars ) {
						// at least three hosts were specified, so we need to
						// make as many ||'s as we need
				hostvalue.SetStringValue(string(hostchars));
				host1constraint =
					 Operation::MakeOperation(Operation::NOT_EQUAL_OP,
											  machineattrref,
											 Literal::MakeLiteral(hostvalue));
				hostsconstraint =
					 Operation::MakeOperation(Operation::LOGICAL_AND_OP,
											  hostsconstraint,
											  host1constraint);
				hostchars = hostnamesToManage->next();
			}
		}
	}

	//TEMP
	dprintf(D_ALWAYS, "okd\n");

	string tmp;
	unp.Unparse(tmp, hostsconstraint);
	dprintf(D_FULLDEBUG, "alien host descrips in log? using constraint:\n%s\n",
			tmp.c_str());


	if( !query.Query(string("root"), hostsconstraint) ) {
		dprintf(D_ALWAYS, "Error making query\n");
	}

	string currentkey;
	for( query.ToFirst(), query.Current(currentkey); !query.IsAfterLast();
		 query.Next(currentkey) ) {
		managedHosts->RemoveClassAd(currentkey);
	}

	/*
	  now make sure every host listed in hostnamesToManage has an ad
	*/

	char *host;
	bool makenewad;
	hostnamesToManage->rewind();
	while( (host = hostnamesToManage->next()) ) {
		ClassAd *adtoadd;
		ClassAd *existingad;
		existingad = managedHosts->GetClassAd(string(host));
		if( existingad ) {
				// the ad is already in the collection

				// do a couple sanity checks
			if( !existingad->Lookup("Machine") ||
				!existingad->Lookup("ReservedBy") ) {
				dprintf(D_FULLDEBUG, "Warning, ad found in logfile that is "
						"missing Machine or ReservedBy\n");
				if( !managedHosts->RemoveClassAd(string(host)) ) {
					dprintf(D_FULLDEBUG,
							"... and for some reason we can't remove it!\n");
				}
				makenewad = 1;
			} else {
				makenewad = 0;
			}
		} else {
			makenewad = 1;
		}

		if( makenewad ) {
				// the ad is not already in the collection
      
				// first get an ad from the collector, if possible
			bool gotad = FALSE;
			CondorQuery startdquery(STARTD_AD);
			ClassAdList collectorads;
			QueryResult result;
      
			char hostconstraint[128];
			snprintf(hostconstraint, 128, "Machine == \"%s\"", host);

			if( startdquery.addORConstraint(hostconstraint) != Q_OK ) {
				dprintf(D_ALWAYS, "Failed to add constraint\n");
			}

			result = startdquery.fetchAds(collectorads);

			if( result == Q_OK ) {
				collectorads.Rewind();
				if( !collectorads.Length() ) {
					// so the collector doesn't have an ad for this host.
					// make a dummy ad.
					adtoadd = new ClassAd();
					adtoadd->InsertAttr("MyType", string("Machine"));
					adtoadd->InsertAttr("Machine", string(host));
					if( !adtoadd->InsertAttr("ReservedBy",
											 string(""))) {
						dprintf(D_ALWAYS, "Error inserting empty "
								"ReservedBy attr\n");
					}
				} else {
					// ahh, the collector has an ad for this host.
					// just make a reserved by
					adtoadd = collectorads.Next();
					if( !adtoadd->InsertAttr("ReservedBy", string(""))) {
						dprintf(D_ALWAYS, "Error inserting ReservedBy "
								"attr\n");
					}
				}
				gotad = TRUE;
				// now we've made it, put it in
				if( !managedHosts->AddClassAd(host, adtoadd->Copy()) ) {
					dprintf(D_ALWAYS, "Error adding new classad to the "
							"master collection\n");
				}
			} else if( result == Q_INVALID_CATEGORY ) {
				dprintf(D_ALWAYS, "Error fetching ads from collector: "
						"Q_INVALID_CATEGORY.\n");
			} else if( result == Q_MEMORY_ERROR ) {
				dprintf(D_ALWAYS, "Error fetching ads from collector: "
						"Q_MEMORY_ERROR.\n");
			} else if( result == Q_PARSE_ERROR ) {
				dprintf(D_ALWAYS, "Error fetching ads from collector: "
						"Q_PARSE_ERROR.\n");
			} else if( result == Q_COMMUNICATION_ERROR ) {
				dprintf(D_ALWAYS, "Error fetching ads from collector: "
						"Q_COMMUNICATION_ERROR.\n");
			} else if( result == Q_INVALID_QUERY ) {
				dprintf(D_ALWAYS, "Error fetching ads from collector: "
						"Q_INVALID_QUERY.\n");
			} else if( result == Q_NO_COLLECTOR_HOST ) {
				dprintf(D_ALWAYS, "Error fetching ads from collector, "
						"unable to find collector.\n");
			} else {
				dprintf(D_ALWAYS, "Error fetching ads from collector, "
						"unknown QueryResult: %d.\n", result);
			}
		}
	}

	return TRUE;
}


bool
ClusterServer::doForgetOrRemember( Stream* s, bool forget )
{
	dprintf(D_FULLDEBUG, "We got a forget\n");

  	int protocolVersion = negotiateProtocolVersion(s);
  	if( protocolVersion < 0 ) {
  		dprintf(D_ALWAYS, "Error negotiating protocol version\n");
  		return FALSE;
  	}

#ifdef NOSECTEST
	const char* fulqualname = "wilbur";
#else
	const char* fulqualname = ((ReliSock*) s)->getFullyQualifiedUser();
#endif
	if( !fulqualname ) {
		dprintf(D_ALWAYS, "Error authenticating client, getFullyQualifiedUser"
				" returned null\n");
		return FALSE;
	}

	char* username = strdup(fulqualname);

		// ugly hack to get just the CN field out of full principle
		// this should be better
	char* userchar = username;
	if( !strncmp(userchar, "/CN=", 4) ) {
		while( strncmp(userchar, "/CN=", 4) && *userchar != 0 ) userchar++;
		if( *userchar != 0 ) username = userchar + 4;
	}
		// get part before '@', e.g., get bethenco from
		// bethenco@your.domain
	userchar = username;
	while( *userchar != '@' && *userchar != 0 ) userchar++;
	*userchar = 0;

	dprintf(D_FULLDEBUG, "client is: %s\n", username);

		// now get request ad
	ClassAd *requestad;
	if( !(requestad = receiveClassAd(s)) ) {
		dprintf(D_ALWAYS, "Error receiving request ad!\n");
		return FALSE;
	}

	string adstring;
	pp.Unparse(adstring, requestad);
	dprintf(D_FULLDEBUG, "The request ad is this: \n%s\n\n", adstring.c_str());

		// get their number of desired machines
	int count = 0;
	requestad->EvaluateAttrInt("MachinesDesired", count);

	LocalCollectionQuery query;

		// get their requirements
	query.Bind(managedHosts);
	ExprTree *requestrequirements = requestad->Lookup("Requirements");
	if( !requestrequirements ) {
		cp.ParseExpression("true", requestrequirements);
	}

		// now get their rank, if they have one
	string requestrank;
	requestad->EvaluateAttrString("Rank", requestrank);

	string requestreqstr;
	unp.Unparse(requestreqstr, requestrequirements);

	if( !managedHosts->CreateSubView(string("current_request"), "root",
									 requestreqstr, requestrank, "") ) {
		dprintf(D_ALWAYS, "Error, unable to CreateSubView, err no %d, %s\n",
				CondorErrno, CondorErrMsg.c_str());
		if( !managedHosts->DeleteView(string("current_request")) ) {
			dprintf(D_ALWAYS,"Error, can't DeleteView either! err no %d, %s\n",
					CondorErrno, CondorErrMsg.c_str());
			return FALSE;
		} else if( !managedHosts->CreateSubView(string("current_request"),
												"root", requestreqstr,
												requestrank, "") ) {
			dprintf(D_ALWAYS, "Error, still can't CreateSubView! err no %d, "
					"%s\n", CondorErrno, CondorErrMsg.c_str());
			return FALSE;
		} else {
			dprintf(D_ALWAYS, "Problem solved, deleted it and recreated it\n");
		}
	}

	if( !query.Query(string("current_request")) ) {
		dprintf(D_ALWAYS, "Error making query\n");
		return FALSE;
	}

	string currentkey;
	int matchcount = 0;
	query.ToFirst();
	if( query.Current(currentkey) ) {
		matchcount++;
		while( query.Next(currentkey) ) {
			matchcount++;
		}
	}
		// we only want up to count matches
	if( count && matchcount > count ) {
		matchcount = count;
	}

		// then forget about any reservations on the matches
	ClassAd *matchedad;
	int i = 0;
	for( query.ToFirst(), query.Current(currentkey);
		 !query.IsAfterLast() && (!count || i++ < count);
		 query.Next(currentkey) ) {

		if( !(matchedad = managedHosts->GetClassAd(currentkey)) ) {
			dprintf(D_ALWAYS,
					"Error, unable to retrieve ClassAd from collection\n");
			break;
		}

		ClassAd* updates = new ClassAd;
		if( !forget ) {
			updates->InsertAttr(string("ReservedBy"), string(username));
			Value reservevalue;
			if( !requestad->EvaluateAttr("ReserveTime", reservevalue) ) {
				dprintf(D_ALWAYS, "Error, requestad does not have a "
						"ReserveTime\n");
				delete updates;
				return FALSE;
			}

			string reservebuffer;
			unp.Unparse(reservebuffer, Literal::MakeAbsTime());
			reservebuffer += " + ";
			unp.Unparse(reservebuffer, reservevalue);

			updates->Insert("ReservedUntil",cp.ParseExpression(reservebuffer));

			string hoststring;
			if( !matchedad->EvaluateAttrString("Machine", hoststring) ) {
				dprintf(D_ALWAYS, "Error, ClassAd has no machine attribute\n");
			}
			dprintf(D_ALWAYS, "Remembered that %s actually had a "
					"reservation on %s\n", username, hoststring.c_str());
		} else {
			updates->InsertAttr(string("ReservedBy"), string(""));

			string hoststring;
			if( !matchedad->EvaluateAttrString("Machine", hoststring) ) {
				dprintf(D_ALWAYS, "Error, ClassAd has no machine attribute\n");
			}
			dprintf(D_ALWAYS, "Forgot about any reservations on %s\n",
					hoststring.c_str());
		}
    
		managedHosts->UpdateClassAd(currentkey, updates);

		adstring = "";
		pp.Unparse(adstring, matchedad);
		dprintf(D_FULLDEBUG, "Modified host ad: \n%s\n\n", adstring.c_str());
	}
	delete requestad;
	
		// this view is now useless
	if( !managedHosts->DeleteView(string("current_request")) ) {
		dprintf(D_ALWAYS, "Error, unable to delete View "
				"'current_request'\n");
	}

	return FALSE;
}


bool
ClusterServer::doReturnOrReserve( Stream *s, bool reserve )
{
	dprintf(D_FULLDEBUG, "We got a %s\n", reserve ? "reserve" : "return");

	int response = CT_RS_OK;
	string username;

		// first get the transaction ID
	int transid = receiveTransactionID(s);
	if( transid < 0 ) {
		dprintf(D_ALWAYS, "Error receiving transaction ID from client\n");
		response = CT_RS_UNKNOWN_ERROR;
		return doReturnOrReserveStage2(s,reserve, transid,username, response);
	}

#ifdef NOSECTEST
	const char* fulqualname = "wilbur";
#else
	const char* fulqualname = ((ReliSock*) s)->getFullyQualifiedUser();
#endif
	if( !fulqualname ) {
		dprintf(D_ALWAYS, "Error authenticating client, getFullyQualifiedUser"
				" returned null\n");
		response = CT_RS_CANT_AUTHENTICATE;
		return doReturnOrReserveStage2(s,reserve, transid,username, response);
	}

	char* usernamenew = strdup(fulqualname);

		// ugly hack to get just the CN field out of full principle
		// this should be better
	char* userchar = usernamenew;
	if( !strncmp(userchar, "/CN=", 4) ) {
		while( strncmp(userchar, "/CN=", 4) && *userchar != 0 ) userchar++;
		if( *userchar != 0 ) usernamenew = userchar + 4;
	}
		// get part before '@', e.g., get bethenco from
		// bethenco@your.domain
	userchar = usernamenew;
	while( *userchar != '@' && *userchar != 0 ) userchar++;
	*userchar = 0;

	dprintf(D_FULLDEBUG, "client is: %s\n", usernamenew);


		// now look up the classad of that ID and see
		// what's up
	ClassAd *currenttransaction;
	char buffer[32];
	sprintf(buffer, "%d", transid);

	if( !(currenttransaction = transactions->GetClassAd(buffer)) ) {
				// this transaction does not exist!
		dprintf(D_ALWAYS, "Error! Invalid transaction ID\n");
		response = CT_RS_UNKNOWN_ERROR;
		return doReturnOrReserveStage2(s,reserve, transid,username, response);
	}
  
		// check to make sure this person's username is the same as the one
		// with that transaction id
	if( !currenttransaction->EvaluateAttrString("Username", username) ||
		!(username == usernamenew) ) {
				// this isn't your business!
		dprintf(D_ALWAYS, "Error! Mismatched usernames\n");
		response = CT_RS_UNKNOWN_ERROR;
		return doReturnOrReserveStage2(s,reserve, transid,username, response);
	}

	ClassAd *updates = new ClassAd;
	if( !updates ) {
		dprintf(D_ALWAYS, "Error creating ClassAd\n");
		response = CT_RS_UNKNOWN_ERROR;
		return doReturnOrReserveStage2(s,reserve, transid,username, response);
	}
	if( !updates->InsertAttr("State", TR_STATE_OPENED) ) {
		dprintf(D_ALWAYS, "Error updating State\n");
		response = CT_RS_UNKNOWN_ERROR;
		return doReturnOrReserveStage2(s,reserve, transid,username, response);
	}
	if( !updates->InsertAttr("Request",
							 reserve ? TR_REQ_RESERVE : TR_REQ_RETURN)) {
		dprintf(D_ALWAYS, "Error updating Request\n");
		response = CT_RS_UNKNOWN_ERROR;
		return doReturnOrReserveStage2(s,reserve, transid,username, response);
	}
	if( !transactions->UpdateClassAd(buffer, updates) ) {
		dprintf(D_ALWAYS, "Error updating ClassAd in collection\n");
		response = CT_RS_UNKNOWN_ERROR;
		return doReturnOrReserveStage2(s,reserve, transid,username, response);
	}

	string adstring;
	pp.Unparse(adstring, currenttransaction);
	dprintf(D_FULLDEBUG, "Modified transaction: \n%s\n\n", adstring.c_str());
	
	return doReturnOrReserveStage2(s, reserve, transid,username, response);
}


bool
ClusterServer::doReturnOrReserveStage2( Stream* s, bool reserve, int transid,
										string username, int response )
{
	int matchcount;
	LocalCollectionQuery query;
	StringList* gettinghosts = new StringList;

		// now get request ad
	ClassAd *requestad;
	if( !(requestad = receiveClassAd(s)) ) {
		response = CT_RS_UNKNOWN_ERROR;
		return doReturnOrReserveStage3(s, reserve, transid, username,
									   matchcount, 0, query, response,
									   gettinghosts, requestad);
	}

	if( response != CT_RS_OK ) {
		return doReturnOrReserveStage3(s, reserve, transid, username,
									   matchcount, 0, query, response,
									   gettinghosts, requestad);
	}

	string adstring;
	pp.Unparse(adstring, requestad);
	dprintf(D_FULLDEBUG, "The request ad is this: \n%s\n\n", adstring.c_str());

		// they can only deal with hosts they are allowed to deal with
	char* hosts;
	string tmp("CLUSTER_AUTH_" + username);
	hosts = param(tmp.c_str());
	StringList *authorizedhosts;

	//TODO:
	//CANONICALIZE AUTHORIZEDHOSTS

	authorizedhosts = new StringList(hosts);
	if( reserve && (!hosts || !authorizedhosts->number()) ) {
			// this person has no hosts they can use!
		dprintf(D_ALWAYS, "Client requested a reservation but is not "
				"authorized to reserve\nany hosts. Consider defining "
				"CLUSTER_AUTH_%s in the config file.\n", username.c_str());
		response = CT_RS_NO_AUTHORIZATION;
		return doReturnOrReserveStage3(s, reserve, transid, username,
									   matchcount, 0, query, response,
									   gettinghosts, requestad);
	}

		// get their number of desired machines
	int count = 0;
	requestad->EvaluateAttrInt("MachinesDesired", count);
	if( !count ) count = INT_MAX;

		// get their requirements
	query.Bind(managedHosts);
	ExprTree *requestrequirements = requestad->Lookup("Requirements");
	if( !requestrequirements ) {
		cp.ParseExpression("true", requestrequirements);
	}

		// add some of our own requirements
	ExprTree *reservedbyref;
	if( reserve ) {
		if( !cp.ParseExpression("other.ReservedBy == \"\"",
								reservedbyref) ) {
			dprintf(D_ALWAYS,"Error! Can't parse other.ReservedBy expression");
			response = CT_RS_UNKNOWN_ERROR;
			return doReturnOrReserveStage3(s, reserve, transid, username,
										   matchcount, 0, query, response,
										   gettinghosts, requestad);
		}
	} else {
		if( !cp.ParseExpression("other.ReservedBy == \"" + username + "\"",
								reservedbyref) ) {
			dprintf(D_ALWAYS,"Error! Can't parse other.ReservedBy expression");
			response = CT_RS_UNKNOWN_ERROR;
			return doReturnOrReserveStage3(s, reserve, transid, username,
										   matchcount, 0, query, response,
										   gettinghosts, requestad);
		}
	}
		// make the right precedence
	requestrequirements = Operation::MakeOperation(Operation::PARENTHESES_OP,
												   requestrequirements);
	reservedbyref = Operation::MakeOperation(Operation::PARENTHESES_OP,
											 reservedbyref);
  
		// put 'em together
	requestrequirements = Operation::MakeOperation(Operation::LOGICAL_AND_OP,
												   requestrequirements,
												   reservedbyref);

	ExprTree *host1constraint;
	ExprTree *host2constraint;
	ExprTree *hostsconstraint;
	Value hostvalue;
	char *hostchars;
	ExprTree *machineattrref;
	if( !cp.ParseExpression("other.Machine", machineattrref) ) {
		dprintf(D_ALWAYS, "Error! Can't parse other.Machine expression");
		response = CT_RS_UNKNOWN_ERROR;
		return doReturnOrReserveStage3(s, reserve, transid, username,
									   matchcount, 0, query, response,
									   gettinghosts, requestad);
	}

	authorizedhosts->rewind();

				// we know at least one host was allowed
	hostchars = authorizedhosts->next();
	hostvalue.SetStringValue(string(hostchars));
	host1constraint =
		Operation::MakeOperation(Operation::EQUAL_OP, machineattrref,
								 Literal::MakeLiteral(hostvalue));
	hostsconstraint = host1constraint;

	hostchars = authorizedhosts->next();
	if( hostchars ) {
		// ah, at least two hosts were allowed, we need to
		// make the first ||
		hostvalue.SetStringValue(string(hostchars));
		host2constraint =
			Operation::MakeOperation(Operation::EQUAL_OP, machineattrref,
									 Literal::MakeLiteral(hostvalue));
		hostsconstraint =
			Operation::MakeOperation(Operation::LOGICAL_OR_OP,
									 host1constraint,
									 host2constraint);

		hostchars = authorizedhosts->next();
		while( hostchars ) {
			// at least three hosts were allowed, so we need to
			// make as many ||'s as we need
			hostvalue.SetStringValue(string(hostchars));
			host1constraint =
				Operation::MakeOperation(Operation::EQUAL_OP,
										 machineattrref,
										 Literal::MakeLiteral(hostvalue));
			hostsconstraint =
				Operation::MakeOperation(Operation::LOGICAL_OR_OP,
										 hostsconstraint,
										 host1constraint);
			hostchars = authorizedhosts->next();
		}
	}

	// make the right precedence
	hostsconstraint = Operation::MakeOperation(Operation::PARENTHESES_OP,
											   hostsconstraint);
	// done making the hosts constraint ExprTree, now to put it in
	requestrequirements =
		Operation::MakeOperation(Operation::LOGICAL_AND_OP,
								 requestrequirements,
								 hostsconstraint);

	delete authorizedhosts;


		// now get their rank, if they have one
	string requestrank;
	requestad->EvaluateAttrString("Rank", requestrank);

	string requestreqstr;
	unp.Unparse(requestreqstr, requestrequirements);
	dprintf(D_FULLDEBUG, "requestreqstr is %s\n", requestreqstr.c_str());

	if( !managedHosts->CreateSubView(string("current_request"), "root",
									 requestreqstr, requestrank, "") ) {
		dprintf(D_ALWAYS, "Error, unable to CreateSubView, err no %d, %s\n",
				CondorErrno, CondorErrMsg.c_str());
		if( !managedHosts->DeleteView(string("current_request")) ) {
			dprintf(D_ALWAYS,"Error, can't DeleteView either! err no %d, %s\n",
					CondorErrno, CondorErrMsg.c_str());
			response = CT_RS_UNKNOWN_ERROR;
			return doReturnOrReserveStage3(s, reserve, transid, username,
										   matchcount, 0, query, response,
										   gettinghosts, requestad);
		} else if( !managedHosts->CreateSubView(string("current_request"),
												"root", requestreqstr,
												requestrank, "") ) {
			dprintf(D_ALWAYS, "Error, still can't CreateSubView! err no %d, "
					"%s\n", CondorErrno, CondorErrMsg.c_str());
			response = CT_RS_UNKNOWN_ERROR;
			return doReturnOrReserveStage3(s, reserve, transid, username,
										   matchcount, 0, query, response,
										   gettinghosts, requestad);
		} else {
			dprintf(D_ALWAYS, "Problem solved, deleted it and recreated it\n");
		}
	}

	if( !query.Query(string("current_request")) ) {
		dprintf(D_ALWAYS, "Error making query\n");
		response = CT_RS_UNKNOWN_ERROR;
		return doReturnOrReserveStage3(s, reserve, transid, username,
									   matchcount, 0, query, response,
									   gettinghosts, requestad);
	}

	string currentkey;
	string currenthostname;
	ClassAd* matchedad;
	matchcount = 0;
	query.ToFirst();
	if( query.Current(currentkey) ) {
		matchcount++;
		if( !(matchedad = managedHosts->GetClassAd(currentkey)) ) {
			dprintf(D_ALWAYS, "Error, unable to retrieve ClassAd from "
					"collection\n");
			response = CT_RS_UNKNOWN_ERROR;
			return doReturnOrReserveStage3(s, reserve, transid, username,
										   matchcount, 0, query, response,
										   gettinghosts, requestad);
		}
		matchedad->EvaluateAttrString(ATTR_MACHINE, currenthostname);
		gettinghosts->append(currenthostname.c_str());
		while( query.Next(currentkey) && matchcount < count ) {
			matchcount++;
			if( !(matchedad = managedHosts->GetClassAd(currentkey)) ) {
				dprintf(D_ALWAYS, "Error, unable to retrieve ClassAd from "
						"collection\n");
				response = CT_RS_UNKNOWN_ERROR;
				return doReturnOrReserveStage3(s, reserve, transid, username,
											   matchcount, 0, query, response,
											   gettinghosts, requestad);
			}
			matchedad->EvaluateAttrString(ATTR_MACHINE, currenthostname);
			gettinghosts->append(currenthostname.c_str());
		}
	}

	return doReturnOrReserveStage3(s, reserve, transid, username,
								   matchcount, count, query, response,
								   gettinghosts, requestad);
}


bool
ClusterServer::doReturnOrReserveStage3( Stream* s, bool reserve, int transid,
		                         string username, int matchcount, int count,
								 LocalCollectionQuery query, int response,
								 StringList* gettinghosts, ClassAd* requestad )
{
	string adstring;
	string currentkey;

	char* gettinghosts_st = gettinghosts->print_to_string();
	s->encode();
	if( !s->code(gettinghosts_st) ) {
		dprintf(D_ALWAYS, "Error sending host string to client\n");
		if( !s->end_of_message() ) {
			dprintf(D_ALWAYS, "Error closing connection with client\n");
		}
		free(gettinghosts_st);
		delete gettinghosts;
		response = CT_RS_UNKNOWN_ERROR;
		return doReturnOrReserveStage4(s, matchcount, 0, query, response);
	}
	if( !s->end_of_message() ) {
		dprintf(D_ALWAYS, "Error closing connection with client\n");
		free(gettinghosts_st);
		delete gettinghosts;
		response = CT_RS_UNKNOWN_ERROR;
		return doReturnOrReserveStage4(s, matchcount, 0, query, response);
	}
	free(gettinghosts_st);
	delete gettinghosts;

	if( response != CT_RS_OK ) {
		return doReturnOrReserveStage4(s, matchcount, 0, query, response);
	}

		// then turn Condor on/off on the matches
	ClassAd *matchedad;
	StringList hoststodo;
	int i = 0;
	for( query.ToFirst(), query.Current(currentkey);
		 !query.IsAfterLast() && i++ < count;
		 query.Next(currentkey) ) {

		if( !(matchedad = managedHosts->GetClassAd(currentkey)) ) {
			dprintf(D_ALWAYS,
					"Error, unable to retrieve ClassAd from collection\n");
			break;
		}

		ClassAd* updates = new ClassAd;
		if( reserve ) {
			updates->InsertAttr(string("ReservedBy"), username);
			Value reservevalue;
			if( !requestad->EvaluateAttr("ReserveTime", reservevalue) ) {
				dprintf(D_ALWAYS, "Error, requestad does not have a "
						"ReserveTime\n");
				delete updates;
				response = CT_RS_UNKNOWN_ERROR;
				return doReturnOrReserveStage4(s, matchcount, count, query,
											   response);
			}

			string reservebuffer;
			unp.Unparse(reservebuffer, Literal::MakeAbsTime());
			reservebuffer += " + ";
			unp.Unparse(reservebuffer, reservevalue);

			updates->Insert("ReservedUntil",cp.ParseExpression(reservebuffer));

			string hoststring;
			if( !matchedad->EvaluateAttrString("Machine", hoststring) ) {
				dprintf(D_ALWAYS, "Error, ClassAd has no machine attribute\n");
			}
			hoststodo.append(hoststring.c_str());
		} else {
			updates->InsertAttr(string("ReservedBy"), string(""));

			string hoststring;
			if( !matchedad->EvaluateAttrString("Machine", hoststring) ) {
				dprintf(D_ALWAYS, "Error, ClassAd has no machine attribute\n");
			}
			hoststodo.append(hoststring.c_str());
		}
    
		managedHosts->UpdateClassAd(currentkey, updates);

		adstring = "";
		pp.Unparse(adstring, matchedad);
		dprintf(D_FULLDEBUG, "Modified host ad: \n%s\n\n", adstring.c_str());
	}

	enactReservation(&hoststodo, username, !reserve);

	delete requestad;

	ClassAd* updates = new ClassAd;
	char buffer[32];
	sprintf(buffer, "%d", transid);
	if( !updates->InsertAttr("State", TR_STATE_EXECUTED) ) {
		dprintf(D_ALWAYS, "Error updating State\n");
	}
	if( !transactions->UpdateClassAd(buffer, updates) ) {
		dprintf(D_ALWAYS, "Error updating ClassAd in collection\n");
	}

	return doReturnOrReserveStage4(s, matchcount, count, query, response);
}

bool
ClusterServer::doReturnOrReserveStage4(Stream* s, int matchcount, int descount,
									 LocalCollectionQuery query, int response )
{
		// now we send the response
	s->encode();
	if( !s->code(response) ) {
		dprintf(D_ALWAYS, "Error sending response to client\n");
		if( !s->end_of_message() ) {
			dprintf(D_ALWAYS, "Error closing connection with client\n");
		}
			// this view is now useless
		if( !managedHosts->DeleteView(string("current_request")) ) {
			dprintf(D_ALWAYS, "Error, unable to delete View "
					"'current_request'\n");
		}
		return FALSE;
	}
	if( response != CT_RS_OK ) {
			// this view is now useless
		if( !managedHosts->DeleteView(string("current_request")) ) {
			dprintf(D_ALWAYS, "Error, unable to delete View "
					"'current_request'\n");
		}
		if( !s->end_of_message() ) {
			dprintf(D_ALWAYS, "Error closing connection with client\n");
			return FALSE;
		}
		return FALSE;
	}

		// now we send number of matches to let client know how things went
	s->encode();
	if( !s->code(matchcount) ) {
		dprintf(D_ALWAYS, "Error sending match count to client\n");
		if( !s->end_of_message() ) {
			dprintf(D_ALWAYS, "Error closing connection with client\n");
		}
			// this view is now useless
		if( !managedHosts->DeleteView(string("current_request")) ) {
			dprintf(D_ALWAYS, "Error, unable to delete View "
					"'current_request'\n");
		}
		return FALSE;
	}
	if( !s->end_of_message() ) {
		dprintf(D_ALWAYS, "Error closing connection with client\n");
			// this view is now useless
		if( !managedHosts->DeleteView(string("current_request")) ) {
			dprintf(D_ALWAYS, "Error, unable to delete View "
					"'current_request'\n");
		}
		return FALSE;
	}

		// if there weren't matches, done
	if( matchcount ) {
			// otherwise send the classads
		int i = 0;
		string currentkey;
		ClassAd* matchedad;
		for( query.ToFirst(), query.Current(currentkey);
			 !query.IsAfterLast() && (i++ < descount);
			 query.Next(currentkey) ) {
			if( !(matchedad = managedHosts->GetClassAd(currentkey)) ) {
				dprintf(D_ALWAYS,
						"Error, can't retrieve ClassAd from collection\n");
			} else {
				if( !putClassAd(s, *matchedad) ) {
					dprintf(D_FULLDEBUG, "put returned false!\n");
				}
			}
		}
	}

		// this view is now useless
	if( !managedHosts->DeleteView(string("current_request")) ) {
		dprintf(D_ALWAYS, "Error, unable to delete View 'current_request'\n");
	}

	return FALSE;
}


int
ClusterServer::negotiateProtocolVersion( Stream* s )
{
	dprintf(D_FULLDEBUG, "Negotiating protocol version\n");

	if( !s ) return -1;

	ClassAd* clientProtos = receiveClassAd(s);
	if( !clientProtos ) return -1;

		// find newest client known protocol
	char buf[32];
	int  useProtocol = -1;
	bool canUseIt = false;
	for( int i = CT_CURRENT_PROTOCOL; i >= 0; i-- ) {
		sprintf(buf, "CanSpeakClusterProtocol%d", i);
		if(clientProtos->EvaluateAttrBool(string(buf), canUseIt) && canUseIt) {
			useProtocol = i;
			break;
		}
	}
	delete clientProtos;
	if( useProtocol < 0 ) {
		dprintf(D_ALWAYS,"Error, client does not support any protocols\n");
		return -1;
	} else {
		dprintf(D_FULLDEBUG, "Using protocol version %d\n", useProtocol);
	}

		// send response with protocol version
	ClassAd* res = new ClassAd;
	res->InsertAttr(string("ClusterProtocolVersion"), useProtocol);

	string adstring;
	pp.Unparse(adstring, res);
	dprintf(D_FULLDEBUG, "sending ad: \n%s\n\n", adstring.c_str());

	if( !putClassAd(s, *res) ) {
		dprintf(D_ALWAYS, "Error, couldn't send protocol version\n");
		return -1;
	}
	delete res;

	return useProtocol;
}


bool
ClusterServer::enactReservation( StringList* hoststodo, const string &user,
								 bool on )
{
	if( !hoststodo || !hoststodo->number() || !user.c_str() ||
		!(on?condor_onCmd:condor_offCmd) ||	!condor_statusCmd )
		return FALSE;

	char* hostliststr = hoststodo->print_to_string();

	dprintf(D_FULLDEBUG, "enactReservation(\"%s\", \"%s\", %s)\n", hostliststr,
			user.c_str(), on ? "TRUE" : "FALSE");

	char onoff[2048];

	char* nexthost;
	hoststodo->rewind();
	while( (nexthost = hoststodo->next()) ) {
		snprintf(onoff, 2048, "%s %s", on ? condor_onCmd : condor_offCmd,
				 nexthost);
		dprintf(D_ALWAYS, "system(\"%s\")\n", onoff);
		int result = system(onoff);

		if( WIFEXITED(result) ) {
			if( WEXITSTATUS(result) ) {
				dprintf(D_ALWAYS, "Error, \"%s\" returned status %d\n", onoff,
						WEXITSTATUS(result));
				return FALSE;
			} else dprintf(D_FULLDEBUG, "successfully ran %s\n", onoff);
		} else {
			dprintf(D_ALWAYS, "Error, couldn't system(\"%s\")\n", onoff);
			return FALSE;
		}
	}

		// ugly! spin wait loop to see if it worked
	char* status[hoststodo->number() + 2];
	bool done = FALSE;
	time_t startwait = time(NULL);
	while( startwait + 240 > time(NULL) && !done ) {
		char csresult[_POSIX_PATH_MAX];
		snprintf(csresult, _POSIX_PATH_MAX,
				 "/tmp/condor_cluster_server.XXXXXX");
		int tmpfd = mkstemp(csresult);
		if( tmpfd < 0 ) {
			dprintf(D_ALWAYS, "Error, couldn't make temp file %s, mkstemp "
					"returned -1\n");
			return FALSE;
		}
		int pid = fork();
		if( !pid ) {
			freopen(csresult, "w", stdout);
			freopen(csresult, "w", stderr);
			close(tmpfd);

			status[0] = condor_statusCmd;
			hoststodo->rewind();
			int i = 1;
			while( (status[i++] = hoststodo->next()) );

			dprintf(D_FULLDEBUG, "About to exec\n");
			for( i = 0; i < hoststodo->number() + 2; i++ )
				dprintf(D_FULLDEBUG, "%s\n", status[i]);

			execv(status[0], status);

			dprintf(D_ALWAYS, "Error, couldn't exec\n");
			for( i = 0; i < hoststodo->number() + 2; i++ )
				dprintf(D_ALWAYS, "%s\n", status[i]);
			exit(1);
		} else {
			close(tmpfd);
			waitpid(pid, NULL, 0);
		}

		FILE* csresultfp = fopen(csresult, "r");
		bool allspace = 1;
		if( !csresultfp ) {
			dprintf(D_ALWAYS, "Error, couldn't open %s\n");
		} else {
			char c = fgetc(csresultfp);
			while( (c = fgetc(csresultfp)) != EOF )
				if( !isspace(c) ) allspace = 0;
			fclose(csresultfp);
			if( unlink(csresult) < 0 ) {
				dprintf(D_ALWAYS, "Error, couldn't delete %s: %s\n", csresult,
						strerror(errno));
			}
		}

		if( csresultfp && (allspace && !on) || (!allspace && on) ) done = TRUE;

		if( !done ) sleep(2);
	}

	if( done ) {
		dprintf(D_ALWAYS, "Turned %s condor daemons on %s for user %s\n",
				on ? "on" : "off", hostliststr, user.c_str());
	} else {
		dprintf(D_ALWAYS, "Error, could not turn %s condor daemons on %s for "
				"user %s\n",
				on ? "on" : "off", hostliststr, user.c_str());
	}
	free(hostliststr);

	return done;
}


int
ClusterServer::receiveTransactionID( Stream *s )
{
	int transid;
	s->decode();
	if( !s->code(transid) ) {
		dprintf(D_ALWAYS, "Error receiving transaction ID\n");
		if( !s->end_of_message() ) {
			dprintf(D_ALWAYS, "Error ending message\n");
		}
		return -1;
	}
	if( !s->end_of_message() ) {
		dprintf(D_ALWAYS, "Error ending message\n");
		return -1;
	}
	return transid;
}


ClassAd*
ClusterServer::receiveClassAd( Stream *s )
{
	ClassAd *newad = new ClassAd;

	if( !getClassAd(s, *newad) ) {
		dprintf(D_ALWAYS, "Error, expected a ClassAd from the client and did "
				"not get one\n");
		delete newad;
		return NULL;
	}
	return newad;
}


bool
ClusterServer::getClassAd( Stream *sock, ClassAd& ad )
{
	ClassAdParser cp;
	if( !sock ) return FALSE;
	char *s = NULL;
	if( !sock->get(s) ) {
		return FALSE;
	}
	sock->decode();
	if( !sock->end_of_message() ) {
		free(s);
		return FALSE;
	}
	if( !cp.ParseClassAd(string(s), ad) ) {
		free(s);
		return FALSE;
	}
	free(s);
	return TRUE;
}


bool
ClusterServer::putClassAd( Stream *sock, ClassAd& ad )
{
	string buffer;
	unp.Unparse(buffer, &ad);
	if( !buffer.c_str() || !sock ) return FALSE;
	char tmp[strlen(buffer.c_str()) + 1];
	tmp[strlen(buffer.c_str())] = 0;
	strcpy(tmp, buffer.c_str());
	if( !sock->put(tmp) ) {
		return FALSE;
	}
	sock->encode();
	if( !sock->end_of_message() ) {
		return FALSE;
	}
	return TRUE;
}
