#include "daemon.h"
#include "condor_classad.h"
#include "classadList.h"
#include "condor_io.h"
#include "proc.h"
#include "sched.h"
#include "HashTable.h"
#include "string_list.h"
#include "list.h"
#include "cluster_server_enum.h"

#define CT_CURRENT_PROTOCOL 1

//#define NOSECTEST

class ClusterServer : public Service
{
  public: 

	/*
	  Create a ClusterServer.
	  Note: It will determine everything it needs to from the config file.
	*/
	ClusterServer();

	/*
	  Destroy this ClusterServer.
	*/
	~ClusterServer();

	/*
	  Register the necessary commands with daemonCore. This needs to be called
	  before any client can make a request to this ClusterServer.
	*/
	void registerCmds();

	/*
	  Register the necessary timers with daemonCore.
	*/
	void registerTmrs();

	/*
	  Callback for DaemonCore. Handle client requests about the status of their
	  previous requests.
	*/
	int handleQueryTransaction( int, Stream* );

	/*
	  Callback for DaemonCore. Handle clients initiating requests.
	*/
	int handleBeginTransaction( int, Stream* );

	/*
	  Callback for DaemonCore. Handle clients finishing requests.
	*/
	int handleEndTransaction( int, Stream* );

	/*
	  Callback for DaemonCore. Handle clients checking on things.
	*/
	int handleCheck( int, Stream* );

	/*
	  Callback for DaemonCore. Handle clients returning nodes early.
	*/
	int handleReturn( int, Stream* );

	/*
	  Callback for DaemonCore. Handle clients reserving nodes.
	*/
	int handleReserve( int, Stream* );

	/*
	  Callback for DaemonCore. Handle requests to forget all about a
	  reservation.
	*/
	int handleForget( int, Stream* );

	/*
	  Callback for DaemonCore. Handle requests to remember a reservation
	  that never happened.
	*/
	int handleRemember( int, Stream* );
#ifdef KILLTEST
	/*
	  Temporary. I (bethenco) was using this to test main_shutdown_graceful.
	*/
	int handleKill( int, Stream* );
#endif
	/*
	  Remove old (one day or older) transactions from transactions collection.
	  This is called periodically by a timer, and it can be called externally
	  at any time.
	*/
	int expireTransactions();

	/*
	  Find expired reservations and deal with them. This is called periodically
	  by a timer, and it can be called externally at any time.
	*/
	int expireReservations();

	/*
	  Truncate the transaction ads collection's and the host ads collection's
	  log files. This is called periodically by a timer, and it can be called
	  externally at any time.
	*/
	int truncateLogs();

	/*
	  Refresh the ads for the managed hosts. This will check whether
	  CLUSTER_HOSTS has changed in the config file and get or make fresh ads
	  either way.
	*/
	int refreshHostAds();

	/*
	  Give the period of the expireTransactions timer.
	  @return number of seconds between calls to expireTransactions
	*/
	int getExpireTransactionsPeriod();

	/*
	  Set the period of the expireTransactions timer.
	  @param newperiod number of seconds between calls to expireTransactions
	*/
	void setExpireTransactionsPeriod( int newperiod );

	/*
	  Give the period of the expireReservations timer.
	  @return number of seconds between calls to expireReservations
	*/
	int getExpireReservationsPeriod();

	/*
	  Set the period of the expireReservations timer.
	  @param newperiod number of seconds between calls to expireReservations
	*/
	void setExpireReservationsPeriod( int newperiod );

	/*
	  Give the period of the truncateLogs timer.
	  @return number of seconds between calls to truncateLogs
	*/
	int getTruncateLogsPeriod();

	/*
	  Set the period of the truncateLogs timer.
	  @param newperiod number of seconds between calls to truncateLogs
	*/
	void setTruncateLogsPeriod( int newperiod );

	/*
	  Give the period of the refreshHostAds timer.
	  @return number of seconds between calls to refreshHostAds
	*/
	int getRefreshHostAdsPeriod();

	/*
	  Set the period of the refreshHostAds timer.
	  @param newperiod number of seconds between calls to refreshHostAds
	*/
	void setRefreshHostAdsPeriod( int newperiod );

  private:

	int                expireTransactionsPeriod;// seconds between calls
	int                expireReservationsPeriod;// to these methods
	int                truncateLogsPeriod;
	int                refreshHostAdsPeriod;

	int                expireTransactionsTmr;	// handles returned by
	int                expireReservationsTmr;	// Register_Timer, used for
	int                truncateLogsTmr;			// changing the period later
	int                refreshHostAdsTmr;

	StringList*        hostnamesToManage;		// self explanatory
	HashTable<MyString, MyString>* conToCanon;  // maps names in the config
												// file to canonical ones

	ClassAdCollection* managedHosts;			// ad collection about hosts
	ClassAdCollection* transactions;			// information about current
												// dealings with clients

	char*			   condor_offCmd;			// path for condor_off
	char*			   condor_onCmd;			// path for condor_on
	char*			   condor_statusCmd;		// path for condor_status

	int                nextTransID;				// the next transaction id

	char*              hostLogFile;				// path for managedHosts
	char*              transLogFile;			// path for transactions

	PrettyPrint        pp;						// parsing utilities
	ClassAdUnParser    unp;
	ClassAdParser      cp;

	/*
	  Get the list of hosts to manage out of the config file, and make it into
	  a StringList. CLUSTER_HOSTS must be defined in the config file.
	  @return the StringList pointer, or null for failure
	*/
	StringList* paramHostList();

	/*
	  Get the location of the logfile for the host ads collection.
	  CLUSTER_HOST_LOG must be defined in the config file.
	  @return the filename, or null for failure
	*/
	char* paramHostLogFile();

	/*
	  Get the location of the logfile for the transaction ads collection.
	  CLUSTER_TRANSACTION_LOG must be defined in the config file.
	  @return the filename, or null for failure
	*/
	char* paramTransLogFile();
	
	/*
	  Get the location of condor_off.
	*/
	char* paramcondor_offCmd();

	/*
	  Get the location of condor_on.
	*/
	char* paramcondor_onCmd();

	/*
	  Get the location of condor_status.
	*/
	char* paramcondor_statusCmd();

	/*
	  Get the amount of time that we should allow between purgings of our
	  transaction collection.
	  @return the appropriate period in seconds
	*/
	int paramExpireTransactionsPeriod();

	/*
	  Get the amount of time that we should allow between checking for expired
	  reservations.
	  @return the appropriate period in seconds
	*/
	int paramExpireReservationsPeriod();

	/*
	  Get the amount of time that we should allow between truncating our logs.
	  @return the appropriate period in seconds
	*/
	int paramTruncateLogsPeriod();
	
	/*
	  Get the amount of time that we should allow between refreshing our host
	  ads.
	*/
	int paramRefreshHostAdsPeriod();

	/*
	  This initializes the transactions ClassAdCollection from its logfile. The
	  data member transLogFile (which is set in the constructor) must have the
	  correct value for it to work.
	  @return true for success, false for failure
	*/
	bool obtainTransactions();

	/*
	  This function initializes managedHosts, which will be used by the
	  ClusterServer from now on. Each managed host has a ClassAd in this
	  collection that includes the attribute ReservedBy and possibly
	  ReservedUntil. ReservedBy is the login of the person who has reserved
	  this node, or "nobody" if it is unreserved. ReservedUntil is the time
	  when the current reservation will expire; if the node is not reserved,
	  ReservedUntil may not be present.
	  The hosts listed in hostnamesToManage will be used to determine which
	  hosts should be managed.
	  @return true for success, false for failure
	*/
	bool obtainHostAds();

	/*
	  This handles forget and remember requests. They alter the server's
	  memory of reservations without making it try to turn condor on or off.
	*/
	bool doForgetOrRemember( Stream* s, bool forget );

	/*
	  This is the code used to begin the handling of return and reserve
	  requests. Since so much of it is the same for both return and reserve,
	  it is all in the same function with a boolean to distinguish the two.
	  It will call doReturnOrReserveStage2 to continue handling the request.
	  @param s a pointer to the Stream to communicate with
	  @param reserve should be true if this is an reserve request, else false
	  @return true for success, false for failure
	*/
	bool doReturnOrReserve( Stream* s, bool reserve );

	/*
	  This handles the return and reserve protocols from getting the request
	  ad on. It will call doReturnOrReserveStage3.
	*/
	bool doReturnOrReserveStage2( Stream* s, bool reserve, int transid,
								  string username, int response );

	/*
	  This handles the return and reserve protocols from sending the string of
	  hosts on. It will call doReturnOrReserveStage4.
	*/
	bool doReturnOrReserveStage3( Stream* s, bool reserve, int transid,
		                        string username, int matchcount, int count,
								LocalCollectionQuery query, int response,
								StringList* gettinghosts, ClassAd* requestad );

	/*
	  This handles the last bit of the return and reserve protocols.
	*/
	bool doReturnOrReserveStage4( Stream* s, int matchcount, int descount,
								  LocalCollectionQuery query, int response );
	
	/*
	  This is called at the beginning of all client-server interactions to
	  determine which protocol version we will both speak. It receives a
	  ClassAd from the client which is of the form
	  [ CanSpeakClusterProtocol1 = true; CanSpeakClusterProtocol2 = true; ]
	  and so on. The server replies with a ClassAd of the form
	  [ ClusterProtocolVersion = 2 ]. If the version is -1, that indicates that
	  the client and server do not share a common protocol version, and the
	  communication is therefore canceled.
	  @param s the stream connected to the client
	  @return the protocol version number we will be using, or -1 for failure
	*/
	int negotiateProtocolVersion( Stream* s );

	/*
	  Turn the Condor daemons off or on on a host.
	  @param host the hostname of the machine
	  @param user the login name to activate
	  @param on if it is true, daemons will be turned on, otherwise they will
			 be turned off
	  @return true for success, false for failure
	*/
	bool enactReservation( StringList* hoststodo, const string &user, bool on);

	/*
	  A small utility method that just receives a transaction ID and prints
	  out any necessary error messages.
	  @param s a pointer to the Stream we want to get the transaction ID from
	  @return the ID, or -1 if something went wrong
	*/
	int receiveTransactionID( Stream* s );

	/*
	  A small utility function that just gets a ClassAd and prints out any
	  necessary error messages.
	  @param s a pointer to the Stream we want to get the ClassAd from
	  @return the ClassAd, or null if something went wrong
	*/
	ClassAd* receiveClassAd( Stream* s );

	/*
	  Receives a ClassAd over a Sock.
	  @param s a pointer to the Sock
	  @param ad the ClassAd to fill in
	  @return true for success, false for failure
	*/
	bool getClassAd( Stream* sock, ClassAd &ad );

	/*
	  Sends a ClassAd over a Sock.
	  @param s a pointer to the Sock
	  @param ad the ClassAd to send
	  @return true for success, false for failure
	*/
	bool putClassAd ( Stream* sock, ClassAd &ad );
};
