/***************************Copyright-DO-NOT-REMOVE-THIS-LINE**
 * CONDOR Copyright Notice
 *
 * See LICENSE.TXT for additional notices and disclaimers.
 *
 * Copyright (c)1990-2001 CONDOR Team, Computer Sciences Department, 
 * University of Wisconsin-Madison, Madison, WI.  All Rights Reserved.  
 * No use of the CONDOR Software Program Source Code is authorized 
 * without the express consent of the CONDOR Team.  For more information 
 * contact: CONDOR Team, Attention: Professor Miron Livny, 
 * 7367 Computer Sciences, 1210 W. Dayton St., Madison, WI 53706-1685, 
 * (608) 262-0856 or miron@cs.wisc.edu.
 *
 * U.S. Government Rights Restrictions: Use, duplication, or disclosure 
 * by the U.S. Government is subject to restrictions as set forth in 
 * subparagraph (c)(1)(ii) of The Rights in Technical Data and Computer 
 * Software clause at DFARS 252.227-7013 or subparagraphs (c)(1) and 
 * (2) of Commercial Computer Software-Restricted Rights at 48 CFR 
 * 52.227-19, as applicable, CONDOR Team, Attention: Professor Miron 
 * Livny, 7367 Computer Sciences, 1210 W. Dayton St., Madison, 
 * WI 53706-1685, (608) 262-0856 or miron@cs.wisc.edu.
****************************Copyright-DO-NOT-REMOVE-THIS-LINE**/

#include "condor_adtypes.h"
#include "condor_classad.h"
#include "condor_common.h"
#include "condor_config.h"
#include "condor_debug.h"
#include "condor_version.h"
#include "daemon.h"
#include "query_result_type.h"
#include "safe_sock.h"
#include "classadList.h"
#include "string_list.h"
#include "cluster_server_enum.h"
#include "condor_auth.h"

Daemon          *ClusterServer;

PrettyPrint     Pp;             // global utility variables
ClassAdParser   Cp;
ClassAdUnParser Unp;

char            Verbose = 0;    // flag - whether to be verbose or not
char            *Myname;        // argv[0]

/*
  Receives a ClassAd over a Stream.
  @param s a pointer to the Stream
  @param ad the ClassAd to fill in
  @return true unless something went wrong, in which case false
*/
bool getClassAd( Stream *sock, ClassAd& ad );

/*
  Sends a ClassAd over a Stream.
  @param s a pointer to the Stream
  @param ad the ClassAd to send
  @return true unless something went wrong, in which case false
*/
bool putClassAd( Stream *sock, ClassAd& ad );

/*
  A small utility function that just gets a ClassAd and prints out any
  necessary error messages.
  @param s a pointer to the Stream we want to get the ClassAd from
  @return the ClassAd, or null if something went wrong
*/
ClassAd* receiveClassAd( Stream *s );

/*
  Print out info about the ClassAds in the list.
  @param list the ClassAds to print about
*/
void printClassAds( ClassAdList *adlist );

/*
  Print out the usage help with Myname as the name of this program (i.e.,
  argv[0]).
*/
void usage();

/*
  Print out information about this version of Condor.
*/
void version() {printf("%s\n%s\n", CondorVersion(), CondorPlatform());}

/*
  Just a subroutine to move some code out of main, which was getting quite
  large.
  @param hostname of desired pool, if null the local pool will be used
*/
void contactClusterServer( char *pool );

/*
  This is called at the beginning of all client-server interactions to
  determine which protocol version we will both speak. It sends a
  ClassAd to the server which is of the form
  [ CanSpeakClusterProtocol1 = true; CanSpeakClusterProtocol2 = true; ]
  and so on. The server replies with a ClassAd of the form
  [ ClusterProtocolVersion = 2 ]. If the version is -1, that indicates that
  the client and server do not share a common protocol version, and the
  communication is therefore canceled.
  @param s the sock connected to the server
  @return the protocol version number we will be using, or -1 for failure
*/
int negotiateProtocolVersion( Sock* s );

/*
  Construct a request ClassAd to send to the collector to obtain ClassAds.
  @param constraint an expression which will be anded into the Requirements
  of the request ClassAd; this can be null for no constraint
  @param time will go into ReserveTime (as a relative time literal)
  @param hostsRequested a StringList with the hostnames of all the hosts you
  want to consider; this can be null if you want to consider all possible
  hosts
  @return the request ClassAd or null if something went wrong
*/
ClassAd *createRequestAd( char *constraint, char *rank, float time, int count,
						  StringList *hostsRequested );

/*
  Asks the cluster server to turn Condor on or off on some nodes.
  @param requestAd pointer to the classad to send to the cluster server
  @param reserve should be true if you want to turn off Condor, false
  otherwise
*/
void returnOrReserve( ClassAd *requestAd, bool reserve );

/*
  Queries the cluster server for a list of cluster nodes matching the
  constraints; prints out the list.
  @param requestAd pointer to the classad to send to the cluster server
*/
void check( ClassAd *requestad );

/*
  Instructs the cluster management server to forget or remember reservations
  in the event of confusion.
  @param requestAd pointer to the classad to send to the cluster server
  @param forget should be true if we want to forget, false otherwise
*/
void forgetOrRemember( ClassAd *requestad, bool forget );

/*
  Callback for ClassAdList::Sort. Returns 1 if '<', 0 otherwise.
*/
int alphaSort( ClassAd *a, ClassAd *b, void* );

int
main( int argc, char *argv[] )
{
	//set_debug_flags("D_FULLDEBUG");
	//set_debug_flags("D_SECURITY");

	config();

	StringList hostsrequested;  // hosts specified on command line, if any
	char       *pool = 0;       // pool to query, if any
	char       *constraint = 0; // constraint expression, if any
	char       *rank = 0;       // rank expression, if any
	float      time = 0;        // time in hours, if any
	int        count = 0;       // machine count, if any
	
	Myname = argv[0];
	while( *argv[0] )
		if( *argv[0] == '/' )
			Myname = ++argv[0];
		else
			argv[0]++;

		// process options
	for( int i = 1; i < argc; i++ ) {
		if( !strcmp(argv[i], "-help") ) {
			usage();
			exit(0);
		} else if( !strcmp(argv[i], "-pool") ) {
			i++;
			if( !argv[i] ) {
				fprintf(stderr,
						"Invalid parameters, -pool requires an argument.\n");
				fprintf(stderr,	"e.g.,   -pool milpin.cs.kc.edu\nCheck the "
						"usage with %s -help\n\n", Myname);
				exit(1);
			}
			pool = argv[i];
		} else if( !strcmp(argv[i], "-version") ) {
			version();
			exit(0);
		} else if( !strcmp(argv[i], "-debug") ) {
			set_debug_flags("D_FULLDEBUG");
		} else if( !strcmp(argv[i], "-verbose") ) {
			Verbose = 1;
		} else if( !strcmp(argv[i], "-constraint") ) {
			i++;
			if( !argv[i] ) {
				fprintf(stderr,	"Invalid parameters, -constraint requires an "
						"argument.\n");
				fprintf(stderr, "e.g.,   -constraint 'other.OpSys == "
						"\"LINUX\"'\n");
				fprintf(stderr, "Check the usage with %s -help\n\n", Myname);
				exit(1);
			}
			constraint = argv[i];
		} else if( !strcmp(argv[i], "-rank") ) {
			i++;
			if( !argv[i] ) {
				fprintf(stderr, "Invalid parameters, -rank requires an "
						"argument.\n");
				fprintf(stderr, "e.g.,   -rank 'other.Memory'\n");
				fprintf(stderr, "Check the usage with %s -help\n\n", Myname);
				exit(1);
			}
			rank = argv[i];
		} else if( !strcmp(argv[i], "-time") ) {
			i++;
			if( !argv[i] || !sscanf(argv[i], " %f ", &time) || time <= 0 ) {
				fprintf(stderr, "Invalid parameters, -time requires a positive"
						" numerical argument.\n");
				fprintf(stderr, "e.g.,   -time 3.75\nCheck the usage with %s "
						"-help\n\n", Myname);
				exit(1);
			}
		} else if( !strcmp(argv[i], "-count") ) {
			i++;
			if( !argv[i] || !sscanf(argv[i], " %d ", &count) ) {
				fprintf(stderr, "Invalid parameters, -count requires an "
						"argument.\n");
				fprintf(stderr, "e.g.,   -count 37\nCheck the usage with %s "
						"-help\n\n", Myname);
				exit(1);
			}
		} else if( argv[i][0] == '-' ) {
			fprintf(stderr, "Invalid parameter: %s\n", argv[i]);
			fprintf(stderr, "Check the usage with %s -help\n\n", Myname);
			exit(1);
		} else if( !strcmp(argv[i], "check")   ||
				   !strcmp(argv[i], "reserve") ||
				   !strcmp(argv[i], "return")  ||
				   !strcmp(argv[i], "trans")   ||
#ifdef KILLTEST
				   !strcmp(argv[i], "kill")    ||
#endif
				   !strcmp(argv[i], "forget")  ||
				   !strcmp(argv[i], "remember") ) {
				// to be handled later in code
		} else {
			hostsrequested.append(argv[i]);
		}
	}

		// make sure they entered some command
	if( argc < 2 || (strcmp(argv[1], "check")    != 0 &&
					 strcmp(argv[1], "reserve")  != 0 &&
					 strcmp(argv[1], "return")   != 0 &&
#ifdef KILLTEST
					 strcmp(argv[1], "kill")     != 0 &&
#endif
					 strcmp(argv[1], "trans")    != 0 &&
					 strcmp(argv[1], "forget")   != 0 &&
					 strcmp(argv[1], "remember") != 0) ) {
		usage();
		exit(0);
	}

		// make sure they specified a -time with their reserve
	if( !strcmp(argv[1], "reserve") && !time ) {
		fprintf(stderr, "Invalid, reserve requires using -time parameter, "
				"e.g.\n");
		fprintf(stderr, "%s reserve -time 5.5 c05.cs.wisc.edu\n", Myname);
		exit(1);
	}

	contactClusterServer(pool);

		// make requestad
	ClassAd *requestad = createRequestAd(constraint, rank, time,
										 count, &hostsrequested);
	if( !requestad ) {
		fprintf(stderr, "Error! Had some trouble making the requestad.\n");
		exit(1);
	}

		// now execute the command, i.e. check, return, reserve, forget,
		// remember, or trans
	if( !strcmp(argv[1], "check") ) {

		check(requestad);
    
	} else if( !strcmp(argv[1], "reserve") ) {
    
		returnOrReserve(requestad, TRUE);

	} else if( !strcmp(argv[1], "return") ) {

		returnOrReserve(requestad, FALSE);
    
	} else if( !strcmp(argv[1], "trans") ) {

		int transid;
		hostsrequested.rewind();
		char *transchars;
		if( !(transchars = hostsrequested.next()) ||
			!sscanf(transchars, " %d ", &transid) ) {
			fprintf(stderr,
					"Invalid, you must specify a transaction ID, e.g.\n");
			fprintf(stderr, "%s trans 34\n", Myname);
			usage();
			exit(1);
		}
		Sock *sock = ClusterServer->startCommand(CT_QUERY_TRANSACTION,
												  Stream::reli_sock, 300);
		if( !sock ) {
			fprintf(stderr, "Error, could not contact cluster management "
					"server.\n");
			exit(1);
		}

  		int protocolVersion = negotiateProtocolVersion(sock);
  		if( protocolVersion < 0 ) {
  			fprintf(stderr, "Error, protocol version negotiation failed.\n");
  			exit(1);
  		}

		sock->encode();
		sock->code(transid);
		sock->end_of_message();
		int response;
		sock->decode();
		sock->code(response);
		sock->end_of_message();
		delete sock;    

		if( response == TR_STATE_EXECUTED || response == TR_STATE_CLOSED ) {
			printf("Transaction number %d has been completed successfully.\n",
				   transid);
		} else if( response == TR_STATE_REQUESTED ||
				   response == TR_STATE_OPENED ) {
			printf("Transaction number %d was never completed. If you still "
				   "care, you can make the\n", transid);
			printf("request again.\n");
		} else {
			printf("Transaction number %d is unknown to the cluster management"
				   " server. Either the\n", transid);
			printf("request was submitted a long time ago, or something went "
				   "wrong and the request\nwas never received.\n");
		}
#ifdef KILLTEST
	} else if( !strcmp(argv[1], "kill") ) {

		Sock *sock = ClusterServer->startCommand(CT_KILL,
												  Stream::reli_sock, 300);
		if( !sock ) {
			fprintf(stderr, "Error, could not contact cluster management "
					"server.\n");
			exit(1);
		}
		sock->end_of_message();
		delete sock;
#endif
	} else if( !strcmp(argv[1], "forget") ) {

		forgetOrRemember(requestad, TRUE);

	} else if( !strcmp(argv[1], "remember") ) {

		forgetOrRemember(requestad, FALSE);

	} else {
				// we shouldn't be able to get here, because the first arg was
				// already checked for validity
		usage();
		exit(0);

	}

	return 0;
}


/*
  Queries the cluster server for a list of cluster nodes matching the
  constraints; prints out the list.
  @param requestad pointer to the classad to send to the cluster server
*/
void
check( ClassAd *requestad )
{
	Sock *sock;
	sock = ClusterServer->startCommand(CT_CHECK, Stream::reli_sock, 300);
	if( !sock ) {
		fprintf(stderr,"Error, could not contact cluster management "
				"server.\n");
		exit(1);
	}

  	int protocolVersion = negotiateProtocolVersion(sock);
  	if( protocolVersion < 0 ) {
  		fprintf(stderr, "Error, protocol version negotiation failed\n");
  		exit(1);
  	}

		// send our classad
	if( !putClassAd(sock, *requestad) ) {
		fprintf(stderr, "Error: unable to send ClassAd\n");
	}

		// get response
	int matchcount;
	sock->decode();
	sock->code(matchcount);
	sock->end_of_message();

	printf("%d host%s that match%s your request %s found.%s\n",
		   matchcount, matchcount == 1 ? "" : "s",
		   matchcount == 1 ? "es" : "",
		   matchcount == 1 ? "was" : "were",
		   matchcount > 0 ? " Receiving ClassAds ..." : "");

		// if there were no matches, done
	if( !matchcount ) {
		return;
	}

		// otherwise get classads and put in list
	ClassAdList *matchedads = new ClassAdList;
	ClassAd *matchedad;

	sock->decode();
	for( int i = 0; i < matchcount; i++ ) {
		matchedad = new ClassAd;
		if( !getClassAd(sock, *matchedad) ) {
			fprintf(stderr,
					"Error, server did not send ClassAd number %d\n", i);
			delete sock;
			delete matchedad;
			delete matchedads;
			return;
		}
  
		string strbuf;
		if( !matchedad->EvaluateAttrString("Machine", strbuf) ) {
			fprintf(stderr, "Error, ClassAd number %d did not have a "
					"hostname\n", i);
			delete sock;
			delete matchedad;
			delete matchedads;
			return;
		}

		matchedads->Insert(matchedad);
	}

  
	printClassAds(matchedads);
  
	delete sock;
	delete matchedads;
	return;
}


/*
  Instructs the cluster management server to forget or remember reservations
  in the event of confusion.
  @param requestAd pointer to the classad to send to the cluster server
  @param forget should be true if we want to forget, false otherwise
*/
void
forgetOrRemember( ClassAd *requestad, bool forget )
{
	Sock *sock;
	sock = ClusterServer->startCommand(forget ? CT_FORGET : CT_REMEMBER,
									   Stream::reli_sock, 300);
	if( !sock ) {
		fprintf(stderr,"Error, could not contact cluster management server\n");
		exit(1);
	}

  	int protocolVersion = negotiateProtocolVersion(sock);
  	if( protocolVersion < 0 ) {
  		fprintf(stderr, "Error, protocol version negotiation failed\n");
  		exit(1);
  	}

		// send our classad
	if( !putClassAd(sock, *requestad) ) {
		fprintf(stderr, "Error: unable to send ClassAd\n");
	}

	sock->end_of_message();

		// that's all; there is no response

	delete sock;
	return;
}


/*
  Asks the cluster server to turn Condor on or off on some nodes.
  @param requestad pointer to the classad to send to the cluster server
  @param should be true if you want to turn on Condor, false otherwise
*/
void
returnOrReserve( ClassAd *requestad, bool reserve )
{
		// get our transaction id
	Sock *sock;
	sock = ClusterServer->startCommand(CT_BEGIN_TRANSACTION,
										Stream::reli_sock, 300);
	if( !sock ) {
		fprintf(stderr, "Error, could not contact cluster management "
				"server.\nRequest was not processed.\n");
		exit(1);
	}

  	int protocolVersion = negotiateProtocolVersion(sock);
  	if( protocolVersion < 0 ) {
  		fprintf(stderr, "Error, protocol version negotiation failed\n");
		fprintf(stderr, "Request was not processed.\n");
  		exit(1);
  	}

	int transid;
	sock->decode();
	if( !sock->code(transid) ) {
		fprintf(stderr, "Error, did not receive transaction ID from cluster "
				"management.\n");
		fprintf(stderr, "Request was not processed.\n");
		exit(1);
	}
	if( !sock->end_of_message() ) {
		fprintf(stderr, "Error, failed to close connection with server.\n");
		fprintf(stderr, "Request was not processed.\n");
		exit(1);
	}
  
	printf("Your transaction ID is %d\n", transid);
	
	delete sock;

		// now use that id to do an reserve command
	sock = ClusterServer->startCommand(reserve ? CT_RESERVE : CT_RETURN,
										Stream::reli_sock, 300);
	if( !sock ) {
		fprintf(stderr, "Error, could not contact cluster management "
				"server.\nRequest was not processed.\n");
		exit(1);
	}

	sock->encode();
	if( !sock->code(transid) ) {
		fprintf(stderr, "Error, failed to send transaction ID\n");
		fprintf(stderr, "Request was not processed.\n");
		return;
	}
	if( !sock->end_of_message() ) {
		fprintf(stderr, "Error, failed to close connection with server\n");
		fprintf(stderr, "Request was not processed.\n");
		return;
	}

		// send our classad
	if( !putClassAd(sock, *requestad) ) {
		fprintf(stderr, "Error, unable to send request ad\n");
		fprintf(stderr, "Request was not processed.\n");
	}

		// get a string which is just the hostnames of the machines the
		// server is trying to reserve (it could be null)
		// this is just to give the user something to look at while they wait
	char* gettinghosts = NULL;
	sock->decode();
	if( !sock->code(gettinghosts) ) {
		fprintf(stderr, "Error, failed to receive list of hosts\n");
		fprintf(stderr, "Request was not processed.\n");
		return;
	}
	if( !sock->end_of_message() ) {
		fprintf(stderr, "Error, failed to close connection with server\n");
		fprintf(stderr, "Request was not processed.\n");
		return;
	}
	if( gettinghosts && *gettinghosts ) {
		printf("Server is working on %s %s.\n"
			   "Please be patient, this may take several minutes ...\n",
			   reserve ? "turning off condor on" : "turning on condor on",
			   gettinghosts);
		free(gettinghosts);
	}

		// get response
	int response;
	if( !sock->code(response) ) {
		fprintf(stderr, "Error, failed to receive server message.\n");
		fprintf(stderr, "It is uncertain whether the server executed your "
				"request.\n");
		fprintf(stderr, "You can check with the command\n");
		fprintf(stderr, "%s trans %d\n", Myname, transid);
		return;
	}

	switch( response ) {
	  case CT_RS_OK:
		break;
	  case CT_RS_CANT_AUTHENTICATE:
		fprintf(stderr, "Error, the server could not authenticate you.\n");
		fprintf(stderr, "Request was not processed.\n");
		return;
	  case CT_RS_NO_AUTHORIZATION:
		fprintf(stderr, "Error, you do not have authorization to reserve any "
				"cluster nodes.\n");
		fprintf(stderr, "Request was not processed.\n");
		return;
	  case CT_RS_UNKNOWN_ERROR:
		fprintf(stderr, "An unknown error has occured.\n");
		fprintf(stderr, "It is uncertain whether the server executed your "
				"request.\n");
		fprintf(stderr, "You can check with the command\n");
		fprintf(stderr, "%s trans %d\n", Myname, transid);
		return;
	  default:
		fprintf(stderr, "Error, received undefined response from server.\n");
		fprintf(stderr, "It is uncertain whether the server executed your "
				"request.\n");
		fprintf(stderr, "You can check with the command\n");
		fprintf(stderr, "%s trans %d\n", Myname, transid);
		return;
	}

	int matchcount;
	sock->decode();
	if( !sock->code(matchcount) ) {
		fprintf(stderr, "Error, failed to receive match count.\n");
		fprintf(stderr, "It is uncertain whether the server executed your "
				"request.\n");
		fprintf(stderr, "You can check with the command\n");
		fprintf(stderr, "%s trans %d\n", Myname, transid);
		return;
	}
	if( !sock->end_of_message() ) {
		fprintf(stderr, "Error, failed to close connection with server.\n");
		fprintf(stderr, "It is uncertain whether the server executed your "
				"request.\n");
		fprintf(stderr, "You can check with the command\n");
		fprintf(stderr, "%s trans %d\n", Myname, transid);
		return;
	}

	printf("%d matching host%s %s %s.%s%s\n",
		   matchcount,
		   matchcount == 1 ? "" : "s",
		   matchcount == 1 ? "was" : "were",
		   matchcount  > 0 ? reserve ? "reserved" : "returned" : "found",
		   matchcount  > 0 ? " Receiving ClassAd" : "",
		   matchcount  > 0 ? matchcount > 1 ? "s ..." : " ..." : "");

		// if there were no matches, just end the transaction
	if( matchcount < 1 ) {
		delete sock;
		sock = ClusterServer->startCommand(CT_END_TRANSACTION,
											Stream::reli_sock, 300);
		if( !sock ) {
			fprintf(stderr, "Error, could not contact cluster management "
					"server.\nRequest was processed, but there were no "
					"matching ClassAds anyway.\n");
			exit(1);
		}
		sock->encode();
		if( !sock->code(transid) ) {
			fprintf(stderr, "Error, failed to send transaction ID to "
					"server\n");
		}
		if( !sock->end_of_message() ) {
			fprintf(stderr, "Error, failed to close connection with server\n");
		}
		delete sock;
		return;
	}

		// otherwise get classads and put in list
	ClassAdList *matchedads = new ClassAdList;
	ClassAd *matchedad;

	for( int i = 0; i < matchcount; i++ ) {
		matchedad = new ClassAd;

		if( !getClassAd(sock, *matchedad) ) {
			fprintf(stderr, "Error, server did not send ClassAd number "
					"%d\n", i);
			delete sock;
			delete matchedad;
			delete matchedads;
			return;
		}
    
		string strbuf;
		if( !matchedad->EvaluateAttrString("Machine", strbuf) ) {
			fprintf(stderr, "Error, ClassAd number %d did not have a "
					"hostname\n", i);
			delete sock;
			delete matchedad;
			delete matchedads;
			return;
		}

		matchedads->Insert(matchedad);
	}

	delete sock;

		// now end the transaction
	sock = ClusterServer->startCommand(CT_END_TRANSACTION,
										Stream::reli_sock, 300);
	if( !sock ) {
		fprintf(stderr, "Error, could not contact cluster management "
				"server.\nIt is uncertain whether the server executed your "
				"request.\nYou can check with the command\n%s trans %d\n",
				Myname, transid);
		exit(1);
	}
	sock->encode();
	if( !sock->code(transid) ) {
		fprintf(stderr, "Error, failed to send transaction ID to server\n");
		return;
	}

	if( !sock->end_of_message() ) {
		fprintf(stderr, "Error, failed to close connection with server\n");
		return;
	}

		// now let's print out info on what we got
	printClassAds(matchedads);
  
	delete sock;
	delete matchedads;
	return;
}


/*
  Construct a request ClassAd to send to the collector to obtain ClassAds.
  @param constraint an expression which will be anded into the Requirements
  of the request ClassAd; this can be null for no constraint
  @param time will go into ReserveTime (as a relative time literal)
  @param hostsrequested a StringList with the hostnames of all the hosts you
  want to consider; this can be null if you want to consider all possible
  hosts
  @return the request ClassAd or null if something went wrong
*/
ClassAd*
createRequestAd( char *constraint, char *rank, float time, int count,
				 StringList *hostsrequested )
{
	ClassAd *requestad = new ClassAd;

	if( !requestad->InsertAttr("TargetType", string("Machine")) ) {
		fprintf(stderr, "Error, cannot insert an attribute into requestad\n");
		return NULL;
	}
	if( time ) {
		Value timeval;
		timeval.SetRelativeTimeValue((long int) (3600 * time));
		if( !requestad->Insert("ReserveTime", Literal::MakeLiteral(timeval))) {
			fprintf(stderr,
					"Error, cannot insert an attribute into requestad\n");
			return NULL;
		}
	}
	if( count && !requestad->InsertAttr("MachinesDesired", count) ) {
		fprintf(stderr, "Error, cannot insert an attribute into requestad\n");
		return NULL;
	}

		// make the hostnames canonical
	StringList* canonicalHosts = new StringList();
	struct hostent* hostinfo;
	char* host;
	hostsrequested->rewind();
	while( (host = hostsrequested->next()) ) {
		hostinfo = gethostbyname(host);
		if( !hostinfo ) {
			fprintf(stderr, "Error, could not resolve hostname `%s': %s\n"
					"Omitting that host.\n", host,
					h_errno == HOST_NOT_FOUND ? "unknown host." :
					h_errno == NO_ADDRESS   ? "host has no IP address." :
					h_errno == NO_RECOVERY  ? "name server error." :
					h_errno == TRY_AGAIN    ? "try again." : "unknown error.");
		} else {
			canonicalHosts->insert(strdup(hostinfo->h_name));
		}
	}

	hostsrequested = canonicalHosts;

	/*
	  If the user specifies hosts on the command line, this code adds an
	  expression form (Machine=="foo1" || Machine=="foo2") to the Requirements
	  of the request ad. If the request ad already has requirements, those are
	  anded on to the new stuff.
	*/
	if( hostsrequested->number() ) {
		ExprTree *host1constraint;    // of the form                 
		ExprTree *host2constraint;    // Machine == "foo.cs.wisc.edu"
		ExprTree *hostsconstraint;    // of various forms
		Value hostvalue;	          // for making string litera

		char *hostchars;              // more convenience variables
		ExprTree *machineattrref;

		if( !Cp.ParseExpression("other.Machine", machineattrref) ) {
			fprintf(stderr, "Error, can't parse other.Machine expression");
			return NULL;
		}

		hostsrequested->rewind();

				// we know at least one host was specified
		hostchars = hostsrequested->next();
		hostvalue.SetStringValue(string(hostchars));
		host1constraint =
			Operation::MakeOperation(Operation::EQUAL_OP, machineattrref,
									 Literal::MakeLiteral(hostvalue));
		hostsconstraint = host1constraint;

		hostchars = hostsrequested->next();
		if( hostchars ) {
				// ah, at least two hosts were specified, we need to make
				// the first ||
			hostvalue.SetStringValue(string(hostchars));
			host2constraint =
				 Operation::MakeOperation(Operation::EQUAL_OP,
										  machineattrref,
										  Literal::MakeLiteral(hostvalue));
			hostsconstraint =
				 Operation::MakeOperation(Operation::LOGICAL_OR_OP,
										  host1constraint,
										  host2constraint);

			hostchars = hostsrequested->next();
			while( hostchars ) {
						// at least three hosts were specified, so we need to
						// make as many ||'s as we need
				hostvalue.SetStringValue(string(hostchars));
				host1constraint =
					 Operation::MakeOperation(Operation::EQUAL_OP,
											  machineattrref,
											  Literal::MakeLiteral(hostvalue));
				hostsconstraint =
					 Operation::MakeOperation(Operation::LOGICAL_OR_OP,
											  hostsconstraint,
											  host1constraint);
				hostchars = hostsrequested->next();
			}
		}

				// make the right precedence
		hostsconstraint = Operation::MakeOperation(Operation::PARENTHESES_OP,
												   hostsconstraint);
				// done making the hosts constraint ExprTree, now to put it in
		if( !requestad->Insert("Requirements", hostsconstraint) ) {
			fprintf(stderr,
					"Error, cannot insert an attribute into requestad\n");
			return NULL;
		}
	}

		// insert the constraint, if there is one  
	if( constraint ) {
		ExprTree *constraintexpr;
		if( !Cp.ParseExpression(string(constraint), constraintexpr, TRUE) ) {
			fprintf(stderr,
					"Error, cannot parse constraint: %s\n", constraint);
			return NULL;
		}

				// put it in
				// first see if there is already something there
		ExprTree *oldrequirements;
		oldrequirements = requestad->Remove("Requirements");

		if( oldrequirements ) {
			if( !requestad->Insert("Requirements",
						 Operation::MakeOperation(Operation::LOGICAL_AND_OP,
												  oldrequirements,
												  constraintexpr)) ) {
				fprintf(stderr, "Error, cannot insert an attribute into "
						"requestad\n");
				return NULL;
			}
		} else {
			if( !requestad->Insert("Requirements", constraintexpr) ) {
				fprintf(stderr, "Error, cannot insert an attribute into "
						"requestad\n");
				return NULL;
			}
		}
	}

		// insert the rank, if there is one
	if( rank && !requestad->InsertAttr("Rank", string(rank)) ) {
		fprintf(stderr, "Error, cannot insert an attribute into requestad\n");
		return NULL;
	}

		// If there was no constraint specified and no hosts were specified,
		// make a blank requirements so the server does not complain.
	ExprTree *reqtree;
	reqtree = requestad->Lookup("Requirements");
	if( !reqtree ) {
		ExprTree *trueexpr;
		Cp.ParseExpression("true", trueexpr);
		if( !requestad->InsertAttr("Requirements", trueexpr) ) {
			fprintf(stderr, "Error, cannot insert an attribute into "
					"requestad\n");
		}
	}

	return requestad;
}

/*
  This is called at the beginning of all client-server interactions to
  determine which protocol version we will both speak. It sends a
  ClassAd to the server which is of the form
  [ CanSpeakClusterProtocol1 = true; CanSpeakClusterProtocol2 = true; ]
  and so on. The server replies with a ClassAd of the form
  [ ClusterProtocolVersion = 2 ]. If the version is -1, that indicates that
  the client and server do not share a common protocol version and the
  communication is therefore canceled.
  @param s the sock connected to the server
  @return the protocol version number we will be using, or -1 for failure
*/
int
negotiateProtocolVersion( Sock* s )
{
	ClassAd* myProtos = new ClassAd;
	myProtos->InsertAttr(string("CanSpeakClusterProtocol1"), true);
	if( !putClassAd(s, *myProtos) ) {
		fprintf(stderr, "Error, couldn't send protocol version negotiation ad "
				"to server\n");
		delete myProtos;
		return -1;
	}
	delete myProtos;

	ClassAd* useProtos = receiveClassAd(s);
	if( !useProtos ) return -1;

	int version;
	if(!useProtos->EvaluateAttrInt(string("ClusterProtocolVersion"), version)){
		fprintf(stderr, "Error, protocol version negotiation response invalid"
				"\n");
		delete useProtos;
		return -1;
	}
	delete useProtos;

	if( version < 0 ) {
		fprintf(stderr, "Error, no common protocol versions between client "
				"and server\n");
		return -1;
	}

	return version;
}

/*
  Just a subroutine to move some code out of main, which was getting quite
  large.
  @param hostname of desired pool, if null the local pool will be used
*/
void
contactClusterServer( char *pool )
{
	char* clusthost = param("CLUSTER_SERVER_HOST");
	char* clustportst = param("CLUSTER_SERVER_PORT");

	if( !clusthost || !clustportst ) {
		fprintf(stderr, "Error, CLUSTER_SERVER_HOST and CLUSTER_SERVER_PORT "
				"must be defined the config\nfile.\n");
		exit(1);
	}

	int clustport = atoi(clustportst);

	//ClusterServer = new Daemon(DT_CLUSTER, 0, pool);
	ClusterServer = new Daemon(clusthost, clustport);

	// printf("Locating cluster server\n");
	if( !ClusterServer->locate() ) {
		fprintf(stderr, "Error, couldn't locate cluster_server: %s\n",
				ClusterServer->error());
		exit(1);
	}

	// printf("cluster is %s %s\n", ClusterServer->hostname(),
	//        ClusterServer->addr());

	free(clusthost);
	free(clustportst);
}

/*
  Print out the usage help with Myname as the name of this program (i.e.,
  argv[0]).
*/
void
usage()
{
	printf(
"Usage: %s reserve -time HOURS [OPTION]... [HOST]...\n"
"   or: %s return [OPTION]... [HOST]...\n"
"   or: %s check [OPTION]... [HOST]...\n"
"   or: %s trans [OPTION]... TRANSID\n"
"\n"
"%s reserve -time HOURS [OPTION]... [HOST]...\n"
"\n"
"Request to have Condor turned off and your login activated on the specified\n"
"cluster nodes. If no hostnames are given on the command line and no\n"
"constraints are specified, then you are requesting to reserve all the hosts\n"
"you control. If -count is specified, you are requesting at most that many\n"
"nodes (if more are available, they are chosen by rank, if specified).\n"
"\n"
"%s return [OPTION]... [HOST]...\n"
"\n"
"Same as above, except you are requesting to return nodes you have already\n"
"reserved.\n"
"\n"
"%s check [OPTION]... [HOST]...\n"
"\n"
"Don't reserve or return any nodes, just print out the status of the nodes\n"
"listed on the command line (if any) and matching your constraint (if any).\n"
"\n"
"%s trans [OPTION]... TRANSID\n"
"\n"
"Check on the status of the transaction begun with transaction ID TRANSID.\n"
"This is only useful under unusual circumstances, such as the cluster\n"
"management server crashing in the middle of a client request.\n"
"\n"
"%s forget [OPTION]... [HOST]...\n"
"\n"
"Make the cluster management server forget about a reservation without\n"
"trying to turn condor back on. This is only useful under unusual\n"
"circumstances, such as someone manually turning condor back on on a\n"
"reserved host.\n"
"\n"
"%s remember [OPTION]... [HOST]...\n"
"\n"
"Make the cluster management server remember a reservation that never\n"
"happened. This is only useful under unusual circumstances, such as someone\n"
"manually turning off condor on an unreserved host.\n"
"\n"
"  -help                Display this usage description\n"
"  -version             Display Condor version\n"
"  -debug               Show extra debugging info\n"
"  -verbose             Print out an entire ClassAd for each result\n"
"  -pool HOSTNAME       Use this central manager\n"
"  -constraint CONST    Use this expression as the requirements; only useful\n"
"                       with -count\n"
"  -rank RANK           Use this expression as the rank; not used by trans\n"
"  -time HOURS          Number of hours to reserve nodes; used only with\n"
"                       reserve\n"
"  -count X             Return or reserve up to X machines; not used by\n"
"                       check or trans\n"
"\n"
"Examples (assuming typical shell quoting):\n"
"%s reserve -time 5.5 c05.cs.wisc.edu c10.cs.wisc.edu\n"
"%s return -count 5 -constraint 'other.OpSys == \"LINUX\"'\n"
"%s check -pool fargo.upd.nc.edu\n",
           Myname, Myname, Myname, Myname, Myname, Myname, Myname, Myname,
           Myname, Myname, Myname, Myname, Myname);
}

/*
  Callback for ClassAdList::Sort. Returns 1 if '<', 0 otherwise.
*/
int
alphaSort( ClassAd *a, ClassAd *b, void* )
{
	string machinea, machineb;
	a->EvaluateAttrString("Machine", machinea);
	b->EvaluateAttrString("Machine", machineb);

	const char* ap = machinea.c_str();
	const char* bp = machineb.c_str();

	while( *ap == *bp && *ap && *bp ) {
		ap++;
		bp++;
	}

	return *ap < *bp ? 1 : 0;
}

/*
  Print out info about the ClassAds in the list.
  @param list the ClassAds to print about
*/
void
printClassAds( ClassAdList *adlist )
{
	string strbuf, strbuf2, strbuf3;
	ClassAd *matchedad;
	adlist->Sort((SortFunctionType) alphaSort);
	adlist->Rewind();
	while( (matchedad = adlist->Next()) ) {
		if( Verbose ) {
			    // print out whole ClassAds
			strbuf = "";
			Pp.Unparse(strbuf, matchedad);
			printf("%s\n\n", strbuf.c_str());
		} else {
			    // print out one line summaries
			if( matchedad->EvaluateAttrString("ReservedBy", strbuf) &&
				strbuf.c_str() && *strbuf.c_str() ) {
				    // reserved
				matchedad->EvaluateAttrString("Machine", strbuf2);
				Value *reservetime = new Value;
				matchedad->EvaluateAttr("ReservedUntil", *reservetime);
				Unp.Unparse(strbuf3, *reservetime);
				printf("%s is reserved by %s until %s\n", strbuf2.c_str(),
					   strbuf.c_str(), strbuf3.c_str());
				delete reservetime;
			} else {
				    // unreserved
				matchedad->EvaluateAttrString("Machine", strbuf2);
				printf("%s is unreserved\n", strbuf2.c_str());
			}
		}
		strbuf = ""; strbuf2 = ""; strbuf3 = "";
	}
}

/*
  A small utility function that just gets a ClassAd and prints out any
  necessary error messages.
  @param s a pointer to the Stream we want to get the ClassAd from
  @return the ClassAd, or null if something went wrong
*/
ClassAd*
receiveClassAd( Stream *s )
{
	ClassAd *newad = new ClassAd;

	if( !getClassAd(s, *newad) ) {
		fprintf(stderr, "Error, expected a ClassAd and did not get one\n");
		delete newad;
		return NULL;
	}
	return newad;
}

/*
  Receives a ClassAd over a Stream.
  @param s a pointer to the Stream
  @param ad the ClassAd to fill in
  @return true unless something went wrong, in which case false
*/
bool
getClassAd( Stream *sock, ClassAd& ad )
{
	ClassAdParser cp;
	if( !sock ) return FALSE;
	char *s = NULL;
	if( !sock->get(s) ) {
		return FALSE;
	}
	sock->decode();
	if( !sock->end_of_message() ) {
		free(s);
		return FALSE;
	}
	if( !cp.ParseClassAd(string(s), ad) ) {
		free(s);
		return FALSE;
	}
	free(s);
	return TRUE;
}

/*
  Sends a ClassAd over a Stream.
  @param s a pointer to the Stream
  @param ad the ClassAd to send
  @return true unless something went wrong, in which case false
*/
bool
putClassAd( Stream *sock, ClassAd& ad )
{
	string buffer;
	Unp.Unparse(buffer, &ad);
	if( !buffer.c_str() || !sock ) return FALSE;
	char tmp[strlen(buffer.c_str()) + 1];
	tmp[strlen(buffer.c_str())] = 0;
	strcpy(tmp, buffer.c_str());
	if( !sock->put(tmp) ) {
		return FALSE;
	}
	sock->encode();
	if( !sock->end_of_message() ) {
		return FALSE;
	}

	return TRUE;
}
