/* ============================================================================
 * DaP_Server.C - Data Positioning Scheduler Server
 * Reliable File Transfer through SRB-FTP-GsiFTP-NeST
 * by Tevfik Kosar <kosart@cs.wisc.edu>
 * University of Wisconsin - Madison
 * February 2002 - ...
 * ==========================================================================*/

#include "dap_server.h"
#include "dap_server_interface.h"
#include "dap_classad_reader.h"
#include "dap_logger.h"
#include "dap_utility.h"
#include "dap_error.h"
#include "dap_scheduler.h"

#define MAX_NUM_JOBS 10
#define MAX_RETRY 100
#define MAXDELAY_INMINUTES 120 //maximum time a job is allowed to finish

extern char WORK_DIR[MAXSTR];
extern char *logfilename;
extern char *xmllogfilename;
extern char *userlogfilename;

char historyfilename[MAXSTR];

int  daemon_std[3];
int  submit_thread_reaper_id, status_thread_reaper_id;
int  transfer_dap_reaper_id, reserve_dap_reaper_id, release_dap_reaper_id;

unsigned long last_dap = 0;  //changed only on new request

ClassAdCollection      *dapcollection;
Scheduler dap_queue;

int listenfd_submit;
int listenfd_status;
int listenfd_remove;

/* ==========================================================================
 * close daemon core file pointers
 * ==========================================================================*/
int return_max_number_of_jobs(char *protocol)
{
  if (!strcmp(protocol, "gsiftp-gsiftp"))
    return 10;
  else if (!strcmp(protocol, "diskrouter-diskrouter"))
    return 1;
     
}

/* ==========================================================================
 * open daemon core file pointers
 * ==========================================================================*/
void open_daemon_core_file_pointers()
{
  char f1[MAXSTR], f2[MAXSTR], f3[MAXSTR];
  
  snprintf(f1,MAXSTR,"%s/dap_dameon.stdin","/tmp");  
  snprintf(f2,MAXSTR,"%s/dap_daemon.stdout","/tmp"); 
  snprintf(f3,MAXSTR,"%s/dap_daemon.stderr","/tmp");

  //set stdin, stdout and stderr file pointers
  daemon_std[0] = open(f1, O_CREAT | O_RDWR);  //stdin
  daemon_std[1] = open(f2, O_CREAT | O_RDWR);  //stdout
  daemon_std[2] = open(f3, O_CREAT | O_RDWR);  //stderr
}

/* ==========================================================================
 * close daemon core file pointers
 * ==========================================================================*/
void close_daemon_core_file_pointers()
{
  close(daemon_std[0]);
  close(daemon_std[1]);
  close(daemon_std[2]);
}
/* ============================================================================
 * check whether the status of a dap job is already logged or not 
 * ==========================================================================*/
int already_logged(char *dap_id, char *status)
{
  ClassAd                *job_ad;
  char                    current_status[MAXSTR];
  string                  key;

  key = "key = ";
  key += dap_id;

  job_ad = dapcollection->GetClassAd(key);
  if (!job_ad) {
    return 0; //no such job!
  }
  else{
    getValue(job_ad, "status", current_status);
    strncpy(current_status, strip_str(current_status), MAXSTR);

    if (!strcmp(current_status, status)){
      return 1; //already logged
    }
    else {
      return 0; //not logged
    }
  }//else
}

/* ============================================================================
 * get the last dap_id from the dap-jobs-to-process
 * ==========================================================================*/
void get_last_dapid()
{
  LocalCollectionQuery    query;
  ClassAd                *job_ad;
  unsigned long           max_dapid = 0;
  char                    dap_id[MAXSTR];
  string                  key;

  query.Bind(dapcollection);  
  query.Query("root", NULL);
  query.ToFirst();

  if ( query.Current(key) ){
    do{
      job_ad = dapcollection->GetClassAd(key);
      if (!job_ad) {
	break; //no matching classad
      }
      else{
	getValue(job_ad, "dap_id", dap_id);
	strncpy(dap_id, strip_str(dap_id), MAXSTR);
	
	if (atol(dap_id) > max_dapid) {
	  max_dapid = atol(dap_id);
	}
      }
    }while (query.Next(key));
  }
  
  //check also the history log file for max dapid
  ClassAd_Reader adreader(historyfilename);
  while(1) {
    if ( !adreader.readAd()) {
      break;
    }
    adreader.getValue("dap_id", dap_id);
    strncpy(dap_id, strip_str(dap_id), MAXSTR);

    if (atol(dap_id) > max_dapid) {
      max_dapid = atol(dap_id);
    }
  }
  //end of while
  
  last_dap = max_dapid;
}

/* ============================================================================
 * dap transfer function
 * ==========================================================================*/
int transfer_dap(char *dap_id, char *src_url, char *dest_url, char *arguments)
{

  char src_protocol[MAXSTR], src_host[MAXSTR], src_file[MAXSTR];
  char dest_protocol[MAXSTR], dest_host[MAXSTR], dest_file[MAXSTR];
  char command[MAXSTR], commandbody[MAXSTR], argument_str[MAXSTR];
  int pid;

  parse_url(src_url, src_protocol, src_host, src_file);
  parse_url(dest_url, dest_protocol, dest_host, dest_file);

  strncpy(src_url, strip_str(src_url), MAXSTR);  
  strncpy(dest_url, strip_str(dest_url), MAXSTR);  
  strncpy(arguments, strip_str(arguments), MAXSTR);  

  snprintf(commandbody, MAXSTR, "DaP.transfer.%s-%s", src_protocol, 
	   dest_protocol);

  //create a new process to transfer the files
  snprintf(command, MAXSTR, "%s/%s/%s", WORK_DIR, DAP_CATALOG,commandbody);
  snprintf(argument_str ,MAXSTR, "%s %s %s %s", 
	   commandbody, src_url, dest_url, arguments);
  
  //  printf("***Argument_str : %s\n", argument_str);
  //  printf("***arguements:*%s*\n", arguments);
  

  pid = daemonCore->Create_Process(
				   command,	// exec
				   argument_str,  // argv
				   PRIV_ROOT,
				   transfer_dap_reaper_id,          //reaper 
				   FALSE,NULL,WORK_DIR,FALSE,NULL,
				   daemon_std             //std.file ids 
				   );

  if (pid > 0) {
    dap_queue.insert(dap_id, pid);
    return DAP_SUCCESS;
  }
  else{
    transfer_dap_reaper(NULL, 0 ,111); //executable not found!
    return DAP_ERROR;                  //--> Find a better soln!
  }
  
}

/* ============================================================================
 * dap reserve function
 * ==========================================================================*/
void reserve_dap(char *dap_id, char *reserve_id, char *reserve_size, char *duration, char *dest_host)
{

  char command[MAXSTR], argument_str[MAXSTR];
  int pid;
  
  //dynamically create a shell script to execute
  snprintf(command, MAXSTR, "%s/DaP.reserve", WORK_DIR);
  
  //create a new process to transfer the files
  snprintf(argument_str ,MAXSTR, "DaP.reserve %s %s %s", 
	   dest_host, reserve_size, duration);

  pid = daemonCore->Create_Process(
				   command,	// exec
				   argument_str,  // argv
				   PRIV_ROOT,
				   reserve_dap_reaper_id,          //reaper 
				   FALSE,NULL,WORK_DIR,FALSE,NULL,
				   daemon_std             //std.file ids 
				   );
  

  dap_queue.insert(dap_id, pid);
}

/* ============================================================================
 * dap release function
 * ==========================================================================*/
void release_dap(char *dap_id, char *reserve_id, char *dest_host)
{

  char command[MAXSTR], argument_str[MAXSTR];
  int pid;

  //get the corresponding lot_id
  LocalCollectionQuery    query;
  ClassAd                *job_ad;
  char                    lot_id[MAXSTR] = "", current_reserve_id[MAXSTR];
  char                    reserve_dap_id[MAXSTR];
  string                  key, constraint;
  ClassAdParser           parser;

  constraint = "other.dap_type == \"reserve\"";
  ExprTree *constraint_tree = parser.ParseExpression( constraint );
  if (!constraint_tree) {
    dprintf(D_ALWAYS, 
	    "Error in parsing constraint: %s\n", constraint.c_str());
  }

  query.Bind(dapcollection);  
  query.Query("root", constraint_tree);
  query.ToFirst();

  if ( query.Current(key) ){
    do{
      job_ad = dapcollection->GetClassAd(key);
      if (!job_ad) {
	dprintf(D_ALWAYS,"No matching add!\n");
	return; //no matching classad
      }
      else{
	getValue(job_ad, "reserve_id", current_reserve_id);
	strncpy(current_reserve_id, strip_str(current_reserve_id), MAXSTR);
	
	if (!strcmp(current_reserve_id, reserve_id)) {
	  getValue(job_ad, "lot_id", lot_id);
	  strncpy(lot_id, strip_str(lot_id), MAXSTR);
	}
      }
    }while (query.Next(key));
  }
  
  //dynamically create a shell script to execute
  snprintf(command, MAXSTR, "%s/DaP.release", WORK_DIR);
  
  //create a new process to transfer the files
  snprintf(argument_str ,MAXSTR, "DaP.release %s %s", 
	   dest_host, lot_id);

  pid = daemonCore->Create_Process(
				   command,	// exec
				   argument_str,  // argv
				   PRIV_ROOT,
				   release_dap_reaper_id,       //reaper 
				   FALSE,NULL,WORK_DIR,FALSE,NULL,
				   daemon_std             //std.file ids 
				   );

  dap_queue.insert(dap_id, pid);
}

/* ============================================================================
 * process the request last read..
 * - if the request is already logged as being in progress, don't log it again
 * ==========================================================================*/
void process_request(ClassAd *currentAd)
{
  char dap_id[MAXSTR], dap_type[MAXSTR];
  
  //get the dap_id of the request
  getValue(currentAd, "dap_id", dap_id); 
  strncpy(dap_id, strip_str(dap_id), MAXSTR);

  //log the new status of the request
  write_collection_log(dapcollection, dap_id, 
		       "status = \"processing_request\"");
  
  char lognotes[MAXSTR];
  getValue(currentAd, "LogNotes", lognotes);
  
  write_xml_user_log(userlogfilename, "MyType", "\"ExecuteEvent\"", 
		     "EventTypeNumber", "1", 
		     "Cluster", dap_id,
		     //		     "Proc", "-1",
		     //		     "Subproc", "-1",
		     "LogNotes", lognotes);

  
  //write_xml_log(xmllogfilename, currentAd, "\"processing_request\"");
  

  //get the dap_type
  getValue(currentAd, "dap_type", dap_type); 
  strncpy(dap_type, strip_str(dap_type), MAXSTR);

  //if dap_type == transfer
  if (!strcmp(dap_type, "transfer")){
    
    char src_url[MAXSTR], dest_url[MAXSTR], arguments[MAXSTR];
    char use_protocol[MAXSTR], alt_protocols[MAXSTR];
    char src_alt_protocol[MAXSTR], dest_alt_protocol[MAXSTR];
    char next_protocol[MAXSTR];
    int protocol_num;

    getValue(currentAd, "src_url", src_url);
    getValue(currentAd, "dest_url", dest_url);
    getValue(currentAd, "arguments", arguments);

    //select which protocol to use
    getValue(currentAd, "use_protocol", use_protocol);
    protocol_num = atoi(use_protocol);
    
    if (!strcmp(use_protocol, "0")){
      printf("use_protocol: %s\n", use_protocol);      
      transfer_dap(dap_id, src_url, dest_url, arguments);
    }
    else{
      getValue(currentAd, "alt_protocols", alt_protocols);
      printf("alt. protocols = %s\n", alt_protocols);
      
      if (strcmp(alt_protocols,"")) {
	strcpy(next_protocol, strtok(alt_protocols, ",") );   

	for(int i=1;i<protocol_num;i++){
	  strcpy(next_protocol, strtok(NULL, ",") );   
	}

	printf("next protocol = %s\n", next_protocol);
	
	if (strcmp(next_protocol,"")) {
	  strcpy(src_alt_protocol, strtok(next_protocol, "-") );   
	  strcpy(dest_alt_protocol, strtok(NULL, "") );   
	  
	  printf("src protocol = %s\n", strip_str(src_alt_protocol));
	  printf("dest protocol = %s\n", strip_str(dest_alt_protocol));

	}
	
      }

      char src_alt_url[MAXSTR], dest_alt_url[MAXSTR];
      char src_protocol[MAXSTR], src_host[MAXSTR], src_file[MAXSTR];
      char dest_protocol[MAXSTR], dest_host[MAXSTR], dest_file[MAXSTR];
      
      parse_url(src_url, src_protocol, src_host, src_file);
      parse_url(dest_url, dest_protocol, dest_host, dest_file);

      if (!strcmp(strip_str(src_alt_protocol), "file"))
	snprintf(src_alt_url, MAXSTR, "file:%s", src_file);  
      else
	snprintf(src_alt_url, MAXSTR, "%s://%s/%s", 
		 strip_str(src_alt_protocol), src_host, src_file);

      if (!strcmp(strip_str(dest_alt_url), "file"))
	snprintf(dest_alt_url, MAXSTR, "file:%s", dest_file);  
      else
	snprintf(dest_alt_url, MAXSTR, "%s://%s/%s", 
		 strip_str(dest_alt_protocol), dest_host, dest_file);


      //--> These "arguments" may need to be removed or chaged !!
      transfer_dap(dap_id, src_alt_url, dest_alt_url, arguments);
      
    }
    
      

    
    

  }

  //if dap_type == reserve 
  else if(!strcmp(dap_type, "reserve")){

    char dest_host[MAXSTR], reserve_size[MAXSTR];
    char duration[MAXSTR], reserve_id[MAXSTR];

    getValue(currentAd, "dest_host", dest_host);
    strncpy(dest_host, strip_str(dest_host), MAXSTR);
    
    getValue(currentAd, "reserve_size", reserve_size); 
    strncpy(reserve_size, strip_str(reserve_size), MAXSTR);
    
    getValue(currentAd, "duration", duration); 
    strncpy(duration, strip_str(duration), MAXSTR);

    getValue(currentAd, "reserve_id", reserve_id); 
    strncpy(reserve_id, strip_str(reserve_id), MAXSTR);

    reserve_dap(dap_id, reserve_id, reserve_size, duration, dest_host);
     
  }

  //if dap_type == release
  else if(!strcmp(dap_type, "release")){
  
    char dest_host[MAXSTR], reserve_id[MAXSTR], lot_id[MAXSTR];

    getValue(currentAd, "dest_host", dest_host); 
    strncpy(dest_host, strip_str(dest_host), MAXSTR);
    
    getValue(currentAd, "reserve_id", reserve_id); 
    strncpy(reserve_id, strip_str(reserve_id), MAXSTR);

    release_dap(dap_id, reserve_id, dest_host);
  }
}

/* ============================================================================
 * regularly check for requests which are in the state of being processed but 
 * not completed yet
 * ==========================================================================*/
void regular_check_for_requests_in_process()
{
  LocalCollectionQuery    query;
  ClassAd                *job_ad;
  ClassAdParser           parser;
  string                  key, constraint;
  char   dap_id[MAXSTR], timestamp[MAXSTR];
  double timediff_inseconds, maxdelay_inseconds;
  int num_attempts = 0; 
		
  //  printf("regular check for requests in process..\n");
  
  maxdelay_inseconds = MAXDELAY_INMINUTES * 60;
  
  //set the constraint for the query
  constraint = "other.status == \"processing_request\"";
  //constraint = "other.status == \"request_completed\"";
  ExprTree *constraint_tree = parser.ParseExpression( constraint );
  if (!constraint_tree) {
    dprintf(D_ALWAYS, 
	    "Error in parsing constraint: %s\n", constraint.c_str());
  }

  query.Bind(dapcollection);  
  query.Query("root", constraint_tree);
  query.ToFirst();

  if ( query.Current(key) ){
    do{
      job_ad = dapcollection->GetClassAd(key);
      if (!job_ad) { //no matching classad
	break;
      }
      else{
	//create a new classad for time comparison
	ClassAd *classad = new ClassAd;
	ClassAdParser parser;
	ExprTree *expr = NULL;

	classad->Insert("currenttime",Literal::MakeAbsTime());

	getValue(job_ad, "timestamp", timestamp); 
	if ( !parser.ParseExpression(timestamp, expr) ) {
	  dprintf(D_ALWAYS,"Parse error\n");
	}
	classad->Insert("timeofrequest", expr);

	if( !parser.ParseExpression("InSeconds(currenttime - timeofrequest)", 
				    expr) ) {
	  dprintf(D_ALWAYS,"Parse error\n");
	}
	classad->Insert("diffseconds", expr);

	if (!classad->EvaluateAttrNumber("diffseconds", timediff_inseconds)) {
	  dprintf(D_ALWAYS, "Error in parsing the attribute: diffseconds\n");
	}
	
	//if the elapsed time is greater than the maximum delay allowed,
	//then reprocess the request..

    	if (timediff_inseconds > maxdelay_inseconds){
	  getValue(job_ad, "dap_id", dap_id); 
	  dprintf(D_ALWAYS, 
		  "dap_id: %s, not completed in %d minutes. Reprocessing...\n", 
		  dap_id, (int)timediff_inseconds/60);
	  
	  
	  
	    unsigned int pid;
	    if (dap_queue.get_pid(pid, dap_id) == DAP_SUCCESS){

	      if (dap_queue.remove(dap_id) == DAP_SUCCESS){
		dprintf(D_ALWAYS, "Killing process %d, and removing dap %s\n", 
			pid, dap_id);
		kill(pid, SIGKILL);

		char attempts[MAXSTR], tempstr[MAXSTR];
		char current_status[MAXSTR];
		string modify_s = "";

		//get num_attempts
		if (getValue(job_ad, "num_attempts", attempts) != DAP_SUCCESS)
		  num_attempts = 0;
		else
		  num_attempts = atoi(attempts);
		
		printf("dap id: %s, attempt # %d\n", dap_id, num_attempts+2);
		
		//		string sstatus = "";
		
		if (num_attempts + 1 < MAX_RETRY){
		  //		  sstatus = "\"request_rescheduled\"";
		  modify_s += "status = \"request_rescheduled\";";
		}
		 
		else{
		  //		  sstatus = "\"request_failed\"";
		  modify_s += "status = \"request_failed\";";
		}
		 
		snprintf(tempstr, MAXSTR, 
			  "num_attempts = %d;",num_attempts + 1);
		 
		 modify_s +=  tempstr;
		 modify_s += "error_code = \"\"";
		 
		 write_collection_log(dapcollection, dap_id, modify_s.c_str());
		 
		 //when request is failed for good, remove it from the collection 
		 //and log it to the history log file..

		 char lognotes[MAXSTR] ;
		 getValue(job_ad, "LogNotes", lognotes);
		   
		 if (num_attempts + 1 >=  MAX_RETRY){

		   dapcollection->RemoveClassAd(key);
		   write_dap_log(historyfilename, "request_failed", 
				 "dap_id", dap_id, "error_code", "\"ABORTED!\"");
		   
		   char exit_status_str[MAXSTR];
		   snprintf(exit_status_str, MAXSTR, "%d", 9);
		   
		   write_xml_user_log(userlogfilename, "MyType", "\"JobCompletedEvent\"", 
				      "EventTypeNumber", "5", 
				      "TerminatedNormally", "1",
				      "ReturnValue", exit_status_str,
				      "Cluster", dap_id,
				      "LogNotes", lognotes);
		   
		   printf("Request aborted!\n");		   
		 }
		 
	      }
	      else{
		dprintf(D_ALWAYS, "Error in Removing process %d\n", pid);
	      }
	      
	    }
	    
	    char use_protocol[MAXSTR];

	    getValue(job_ad, "use_protocol", use_protocol);

	    if (dap_queue.get_numjobs() < MAX_NUM_JOBS){
		
		if (num_attempts + 2 < MAX_RETRY){
		  process_request(job_ad);
		}
		
	      }
	    
	}
	
	if (classad != NULL) delete classad;	
      }//else
    }while (query.Next(key));
  }//if
}

/* ============================================================================
 * check at the startup for requests which are in the state of being processed  * but not completed yet
 * ==========================================================================*/
void startup_check_for_requests_in_process()
{
  LocalCollectionQuery    query;
  ClassAd                *job_ad;
  ClassAdParser           parser;
  string                  key, constraint;
  
  //set the constraint for the query
  constraint = "other.status == \"processing_request\"";
  ExprTree *constraint_tree = parser.ParseExpression( constraint );
  if (!constraint_tree) {
    dprintf(D_ALWAYS,"Error in parsing constraint!\n");
  }

  query.Bind(dapcollection);  
  query.Query("root", constraint_tree);
  query.ToFirst();

  if ( query.Current(key) ){
    do{
      job_ad = dapcollection->GetClassAd(key);
      if (!job_ad) { //no matching classad
	break;
      }
      else{
	if (dap_queue.get_numjobs() < MAX_NUM_JOBS)
	  process_request(job_ad);
      }
    }while (query.Next(key));
  }
}
/* ============================================================================
 * check for requests which are rescheduled for execution
 * ==========================================================================*/
void regular_check_for_rescheduled_requests()
{
  LocalCollectionQuery    query;
  ClassAd                *job_ad;
  ClassAdParser           parser;
  string                  key, constraint;
  
  //set the constraint for the query
  constraint = "other.status == \"request_rescheduled\"";
  ExprTree *constraint_tree = parser.ParseExpression( constraint );
  if (!constraint_tree) {
    dprintf(D_ALWAYS,"Error in parsing constraint!\n");
  }

  query.Bind(dapcollection);  
  query.Query("root", constraint_tree);
  query.ToFirst();

  if ( query.Current(key) ){
    do{
      job_ad = dapcollection->GetClassAd(key);
      if (!job_ad) { //no matching classad
	break;
      }
      else{
	if (dap_queue.get_numjobs() < MAX_NUM_JOBS)
	  process_request(job_ad);
      }
    }while (query.Next(key));
  }
}

/* ============================================================================
 * initialize dapcollection 
 * ==========================================================================*/
void initialize_dapcollection()
{
  dapcollection = new ClassAdCollection();

  if (dapcollection == NULL ){
    dprintf(D_ALWAYS,"Error: Failed to construct dap collection!\n");
  }

  if (!dapcollection->InitializeFromLog(logfilename)) {
    dprintf(D_ALWAYS,"Couldn't recover ClassAdCollection from log: %s!\n",
	   logfilename);
  }
}


/* ============================================================================
 * make some initializations before calling main()
 * ==========================================================================*/
void initializations()
{

  //initialize and open daemoncore file pointers
  open_daemon_core_file_pointers();

  //initialize dapcollection 
  initialize_dapcollection();

  //init history file name 
  snprintf(historyfilename, MAXSTR, "%s.history", logfilename);

  //init xml log file name 
  //  sprintf(xmllogfilename, "%s.xml", logfilename);<-- change this

  if (!userlogfilename){
    userlogfilename = (char*) malloc(MAXSTR * sizeof(char));
    snprintf(userlogfilename, MAXSTR, "%s.user_log", logfilename);
  }
  
  //get the last dap_id from dap-jobs-to-process
  get_last_dapid();  

  //check the logfile for noncompleted requests at startup
  startup_check_for_requests_in_process();

  //listen to incoming requests from clients
  listen_to_submit_requests();
  listen_to_remove_requests();
  listen_to_status_requests();
}

/* ============================================================================
 * main body of the condor_srb_reqex
 * ==========================================================================*/
int call_main()
{

  LocalCollectionQuery    query;
  ClassAd                *job_ad;
  ClassAdParser           parser;
  string                  key, constraint;
  
  //setup constraints for the query
  constraint = "other.status == \"request_received\"";
  ExprTree *constraint_tree = parser.ParseExpression( constraint );
  if (!constraint_tree) {
    dprintf(D_ALWAYS, "Error in parsing constraint!\n");
  }

  query.Bind(dapcollection);  
  query.Query("root", constraint_tree);
  query.ToFirst();

  if ( query.Current(key) ){
    do{
      job_ad = dapcollection->GetClassAd(key);
      if (!job_ad) { //no matching classad
	break;
      }
      else{
	if (dap_queue.get_numjobs() < MAX_NUM_JOBS)
	  process_request(job_ad);
      }
    }while (query.Next(key));
  }

  accept_submit_requests();
  accept_status_requests();
  accept_remove_requests();

  //  if (constraint_tree != NULL) delete constraint_tree;
  return TRUE;
}

/* ============================================================================
 * write incoming dap requests to file: <dap-jobs-to-process>
 * ==========================================================================*/
void write_requests_to_file(int sockfd)
{
  ssize_t n;
  char line[MAX_TCP_BUF];

  for ( ; ; ){
    if ((n = recv(sockfd, line, MAX_TCP_BUF, 0)) == -1){
      dprintf(D_ALWAYS, "Server: recv error!\n");
      exit(1);
    }
    else if (n == 0) break;
    
    //convert request into a classad
    ClassAdParser parser;
    ClassAd *requestAd = NULL;
    PrettyPrint unparser;
    string adbuffer = "";
    char last_dapstr[MAXSTR];
    ExprTree *expr = NULL;

    string linestr = line;

    //check the validity of the request
    requestAd = parser.ParseClassAd(linestr);
    if (requestAd == NULL) {
      dprintf(D_ALWAYS, "Invalid input format! Not a valid classad!\n");
      return;
    }
    
    //add the dap_id to the request
    snprintf(last_dapstr, MAXSTR, "%d", last_dap + 1);
    if ( !parser.ParseExpression(last_dapstr, expr) ) {
      dprintf(D_ALWAYS,"Parse error\n");
    }
    requestAd->Insert("dap_id", expr);

    //add the status to the request
    if ( !parser.ParseExpression("\"request_received\"", expr) ) {
      dprintf(D_ALWAYS,"Parse error\n");
    }
    requestAd->Insert("status", expr);
    requestAd->Insert("timestamp",Literal::MakeAbsTime());

    //use the default protocol first
    if ( !parser.ParseExpression("0", expr) ) {
      dprintf(D_ALWAYS,"Parse error\n");
    }
    requestAd->Insert("use_protocol", expr);

    //insert the request to the dap collection
    char keystr[MAXSTR];
    snprintf(keystr, MAXSTR, "key = %d", last_dap + 1);

    //add the rquest classad to dap collection
    dapcollection->AddClassAd(keystr, requestAd);

    //increment last dap_id
    last_dap ++;

    unparser.Unparse(adbuffer, requestAd);
    dprintf(D_ALWAYS, "New Request => %s\n",adbuffer.c_str());

    
    //---> XML Stuff
    //ClassAdXMLUnParser  xmlunparser;
    //string xmlstring;
    //xmlunparser.SetCompactSpacing(false);
    //xmlunparser.Unparse(xmlstring, requestAd);
    //printf("new Request in XML format=> %s\n",xmlstring.c_str());
    //write_xml_log(xmllogfilename, requestAd, "\"request_received\"");
    
    
    //write XML user logs
    char lognotes[MAXSTR];
    getValue(requestAd, "LogNotes", lognotes);

    write_xml_user_log(userlogfilename, "MyType", "\"SubmitEvent\"", 
			"EventTypeNumber", "0", 
		       "Cluster", last_dapstr,
		       // "Proc", "-1",
		       // "Subproc", "-1",
		       "LogNotes", lognotes);
    

    
    //send response to the client
    char response[MAXSTR];
    
    snprintf(response, MAXSTR, "%s", last_dapstr);
    if ( send (sockfd, response, strlen(response),0) == -1){
      dprintf(D_ALWAYS, "Server: send error!\n");
      exit(1);
    }

    //    if (expr != NULL) delete expr;
  }

}

/* ============================================================================
 * listen to the incoming dap_submit requests from clients
 * ==========================================================================*/
void listen_to_submit_requests()
{
  struct sockaddr_in serv_addr;

  if ( (listenfd_submit = socket(AF_INET, SOCK_STREAM, 0)) < 0){ 
    perror("dap_server:cannot open socket!\n");	
    exit(1);			     		
  }				     		
  
  bzero(&serv_addr, sizeof(serv_addr)); 	
  serv_addr.sin_family      = AF_INET;   	
  serv_addr.sin_addr.s_addr = htonl(INADDR_ANY);
  serv_addr.sin_port        = htons(SERV_SUBMIT_TCP_PORT);

  if(bind(listenfd_submit,(struct sockaddr*) &serv_addr, sizeof(serv_addr)) < 0){
    perror("dap_server:cannot bind to local port..\n");
    exit(1);
  }		

  if (listen(listenfd_submit, LISTENQ) < 0){
    perror("dap_server:error in listen!\n");
    exit(1);  
  }	
      
  dprintf(D_ALWAYS, "Listening to dap_submit requests from clients..\n");
}

void accept_submit_requests()
{
  int connfd;
  pid_t childpid;
  socklen_t clilen;
  struct sockaddr_in cli_addr;
  
  for ( ; ; ){
    int flags;
    
    flags = fcntl(listenfd_submit, F_GETFL, 0);
    fcntl(listenfd_submit, F_SETFL, flags | O_NONBLOCK); 
    
    clilen = sizeof(cli_addr);
    if ((connfd = accept(listenfd_submit,
			 (struct sockaddr*) &cli_addr, &clilen)) <0){
      return;
    }	

    write_requests_to_file(connfd);
    close(connfd);
  }
}

/* ============================================================================
 * process dap_status requests from clients and send the result back
 * ==========================================================================*/
void send_dap_status_to_client(int sockfd)
{
  ssize_t n;
  char dap_id[MAXSTR], nextdap[MAXSTR];
  char status[MAXSTR], timestamp[MAXSTR], error_code[MAXSTR];
  char tempstr[MAXSTR];
  string adbuffer = "";
  ClassAdUnParser unparser;

  for ( ; ; ){
    if ((n = recv(sockfd, dap_id, MAX_TCP_BUF, 0)) == -1){
      dprintf(D_ALWAYS, "Server: recv error!\n");
      exit(1);
    }
    else if (n == 0) break;
    
    dap_id[n] = '\0';

    ClassAd                *job_ad;
    char                    current_status[MAXSTR];
    string                  key;

    //search the collection for the classad
    key = "key = ";
    key += dap_id;

    job_ad = dapcollection->GetClassAd(key);
    if (job_ad) {
      getValue(job_ad, "status", status);
      strncpy(status, strip_str(status), MAXSTR);

      getValue(job_ad, "timestamp", timestamp); 

      if ( !strcmp(status, "request_failed") ){
	getValue(job_ad, "error_code", error_code); 
	snprintf(tempstr, MAXSTR, "status = %s;\nerrorcode = %s;\ntimestamp = %s;\n\n", 
		status, error_code, timestamp);
      }
      else{
	snprintf(tempstr, MAXSTR, "status = %s;\ntimestamp = %s;\n\n", 
		status, timestamp);
      }
      adbuffer += tempstr;
    }//if
    
    //if the classad is not in the collection, search the history log file
    if (adbuffer == ""){
      ClassAd_Reader adreader(historyfilename);
      
      while(1) {
	if ( !adreader.readAd()) {
	  break;
	}
	adreader.getValue("dap_id", nextdap); 
	strncpy(nextdap, strip_str(nextdap), MAXSTR);
      
	if ( !strcmp(dap_id, nextdap) ){
	  adreader.getValue("status", status); 
	  adreader.getValue("timestamp", timestamp); 
	  
	  if ( !strcmp(status, "request_failed") ){
	    adreader.getValue("error_code", error_code); 
	    snprintf(tempstr, MAXSTR, 
		    "status = %s;\nerrorcode = %s;\ntimestamp = %s;\n\n",
		    status, error_code, timestamp);
	  }
	  else{
	    snprintf(tempstr, MAXSTR, 
		    "status = %s;\ntimestamp = %s;\n\n", status, timestamp);
	  }
	  
	  adbuffer += tempstr;
	}//if
      }//while
    }//if
    
    //if the history for dap_id still not found:
    if (adbuffer == ""){
      snprintf(tempstr, MAXSTR, "There is no log history for this job...\n");
      adbuffer = tempstr;
    }
    
    dprintf(D_ALWAYS, "=============================================\n");
    dprintf(D_ALWAYS, "status report for the job with dap_id: %s ==>\n\n%s",dap_id, adbuffer.c_str());
    dprintf(D_ALWAYS, "=============================================\n");	

    //send response to the client
    if ( send (sockfd, adbuffer.c_str(), strlen(adbuffer.c_str()),0) == -1){
      dprintf(D_ALWAYS, "Server: send error!\n");
      exit(1);
    }
  }
}

/* ============================================================================
 * listen to the incoming dap_status requests from clients
 * ==========================================================================*/
void listen_to_status_requests()
{
  struct sockaddr_in serv_addr;

  if ( (listenfd_status = socket(AF_INET, SOCK_STREAM, 0)) < 0){ 
    perror("dap_server:cannot open socket!\n");	
    exit(1);			     		
  }				     		
  
  bzero(&serv_addr, sizeof(serv_addr)); 	
  serv_addr.sin_family      = AF_INET;   	
  serv_addr.sin_addr.s_addr = htonl(INADDR_ANY);
  serv_addr.sin_port        = htons(SERV_STATUS_TCP_PORT);

  if(bind(listenfd_status,(struct sockaddr*) &serv_addr, sizeof(serv_addr)) < 0){
    perror("dap_server:cannot bind to local port..\n");
    exit(1);
  }		

  if (listen(listenfd_status, LISTENQ) < 0){
    perror("dap_server:error in listen!\n");
    exit(1);  
  }	
      
  dprintf(D_ALWAYS, "Listening to dap_status requests from clients..\n");
}

void accept_status_requests()
{
  int connfd;
  pid_t childpid;
  socklen_t clilen;
  struct sockaddr_in cli_addr;
  
  for ( ; ; ){
    int flags;
    
    flags = fcntl(listenfd_status, F_GETFL, 0);
    fcntl(listenfd_status, F_SETFL, flags | O_NONBLOCK); 
    
    clilen = sizeof(cli_addr);
    if ((connfd = accept(listenfd_status,
			 (struct sockaddr*) &cli_addr, &clilen)) <0){
      return;
    }	

    send_dap_status_to_client(connfd);
    close(connfd);
  }
}

/* ============================================================================
 * remove requests from queue ..
 * ==========================================================================*/
void remove_requests_from_queue(int sockfd)
{
  ssize_t n;
  char dap_id[MAXSTR], nextdap[MAXSTR];
  char status[MAXSTR], timestamp[MAXSTR], error_code[MAXSTR];
  char tempstr[MAXSTR];
  string adbuffer = "";
  ClassAdUnParser unparser;

  
  for ( ; ; ){
    if ((n = recv(sockfd, dap_id, MAX_TCP_BUF, 0)) == -1){
      dprintf(D_ALWAYS, "Server: recv error!\n");
      exit(1);
    }
    else if (n == 0) break;
    
    dap_id[n] = '\0';

    ClassAd                *job_ad;
    char                    current_status[MAXSTR];
    string                  key;

    //search the collection for the classad
    key = "key = ";
    key += dap_id;

    job_ad = dapcollection->GetClassAd(key);
    if (job_ad) {

      unsigned int pid;

      if (dap_queue.get_pid(pid, dap_id) == DAP_SUCCESS){
	dprintf(D_ALWAYS, "Killing process: %d for DaP job: %s\n", pid, dap_id);
	kill(pid, SIGKILL);
      }
      else{
	dprintf(D_ALWAYS, "Corresponding process for DaP job: %s not found\n", dap_id);
      }
      
      int ret;
      
      if (dap_queue.remove(dap_id) == DAP_SUCCESS){
	dprintf(D_ALWAYS, "Removed DaP job: %s\n", dap_id);
      }
      else{
	dprintf(D_ALWAYS, "Error in removing DaP job: %s from dap_queue\n", dap_id);
	dprintf(D_ALWAYS, "Still deleting DaP job: %s from collection & logs\n", dap_id);
      }
      
      string modify_s = "status = \"request_removed\"";
      write_collection_log(dapcollection, dap_id, modify_s.c_str());
	  
      char lognotes[MAXSTR] ;
      getValue(job_ad, "LogNotes", lognotes);
      
      dapcollection->RemoveClassAd(key);

      write_dap_log(historyfilename, "request_removed", 
		    "dap_id", dap_id, "error_code", "\"REMOVED!\"");
      
      write_xml_user_log(userlogfilename, "MyType", "\"JobAbortedEvent\"", 
			 "EventTypeNumber", "9", 
			 "TerminatedNormally", "0",
			 "Cluster", dap_id,
			 "LogNotes", lognotes);
      
      
      adbuffer += "DaP job ";
      adbuffer += dap_id;
      adbuffer += " is removed from queue..\n";      
    }//if
    else {
      dprintf(D_ALWAYS, "Couldn't find/remove DaP job: %s\n", dap_id);
      adbuffer += "Couldn't find/remove DaP job: ";
      adbuffer += dap_id;
      adbuffer += "\n";
    }
    
    //send response to the client
    if ( send (sockfd, adbuffer.c_str(), strlen(adbuffer.c_str()),0) == -1){
      dprintf(D_ALWAYS, "Server: send error!\n");
      exit(1);
    }
  }
}

/* ============================================================================
 * listen to the incoming dap_rm requests from clients
 * ==========================================================================*/
void listen_to_remove_requests()
{
  struct sockaddr_in serv_addr;

  if ( (listenfd_remove = socket(AF_INET, SOCK_STREAM, 0)) < 0){ 
    perror("dap_server:cannot open socket!\n");	
    exit(1);			     		
  }				     		
  
  bzero(&serv_addr, sizeof(serv_addr)); 	
  serv_addr.sin_family      = AF_INET;   	
  serv_addr.sin_addr.s_addr = htonl(INADDR_ANY);
  serv_addr.sin_port        = htons(SERV_REMOVE_TCP_PORT);

  if(bind(listenfd_remove,(struct sockaddr*) &serv_addr, sizeof(serv_addr)) < 0){
    perror("dap_server:cannot bind to local port..\n");
    exit(1);
  }		

  if (listen(listenfd_remove, LISTENQ) < 0){
    perror("dap_server:error in listen!\n");
    exit(1);  
  }	
      
  dprintf(D_ALWAYS, "Listening to dap_status requests from clients..\n");
}

void accept_remove_requests()
{
  int connfd;
  pid_t childpid;
  socklen_t clilen;
  struct sockaddr_in cli_addr;
  
  for ( ; ; ){
    int flags;
    
    flags = fcntl(listenfd_remove, F_GETFL, 0);
    fcntl(listenfd_remove, F_SETFL, flags | O_NONBLOCK); 
    
    clilen = sizeof(cli_addr);
    if ((connfd = accept(listenfd_remove,
			 (struct sockaddr*) &cli_addr, &clilen)) <0){
      return;
    }	

    remove_requests_from_queue(connfd);
    close(connfd);
  }
}

/* ============================================================================
 * reaper functions:
 * ==========================================================================*/
int submit_thread_reaper(Service *,int pid,int exit_status)
{
  return TRUE;
}

int status_thread_reaper(Service *,int pid,int exit_status)
{
  return TRUE;
}

/* ============================================================================
 * shared dap reaper function
 * ==========================================================================*/
int dap_reaper(string modify_s, int pid,int exit_status)
{
  char dap_id[MAXSTR], dap_type[MAXSTR], reserve_id[MAXSTR];
  char fname[MAXSTR], linebuf[MAXSTR]="", linebufstr[MAXSTR] = "\"";
  FILE *f;
  ClassAd *job_ad;
  string key = "key = ";
  
  dprintf(D_ALWAYS, "Process %d terminated with exit status %d \n", 
	 pid, exit_status);
  snprintf(fname, MAXSTR, "%s/out.%d", WORK_DIR, pid);

  if (dap_queue.get_dapid(pid, dap_id) != DAP_SUCCESS){
    printf("Process %d not in queue!\n", pid);
    //dap_queue.remove("0"); //just decrease num_jobs
    return DAP_ERROR;
  }
  
  dap_queue.remove(dap_id);

  key += dap_id;
  job_ad = dapcollection->GetClassAd(key);
  if (!job_ad) {
    return 0; //no such job!
  }

  getValue(job_ad, "dap_type", dap_type);
  strncpy(dap_type, strip_str(dap_type), MAXSTR);

  char lognotes[MAXSTR] ;
  getValue(job_ad, "LogNotes", lognotes);

  //---- if process completed succesfully ----
  if (exit_status == 0){
    if (!strcmp(dap_type, "reserve")){
      f = fopen(fname, "r");
	fgets( linebuf, MAXSTR, f);

	modify_s += "lot_id = ";
	modify_s += linebuf;
	modify_s += ";";
    }
    
    modify_s += "status = \"request_completed\"";
    write_collection_log(dapcollection, dap_id, modify_s.c_str());

    //-- remove the corresponding reserve request from collection
    if (!strcmp(dap_type, "release")){

      getValue(job_ad, "reserve_id", reserve_id);
      strncpy(reserve_id, strip_str(reserve_id), MAXSTR);

      LocalCollectionQuery    query;
      ClassAd                *job_ad2;
      char                    current_reserve_id[MAXSTR];
      char                    reserve_dap_id[MAXSTR];
      string                  key2, constraint;
      ClassAdParser           parser;

      constraint = "other.dap_type == \"reserve\"";
      ExprTree *constraint_tree = parser.ParseExpression( constraint );
      if (!constraint_tree) {
	dprintf(D_ALWAYS, "Error in parsing constraint: %s\n", constraint.c_str());
      }

      query.Bind(dapcollection);  
      query.Query("root", constraint_tree);
      query.ToFirst();

      if ( query.Current(key2) ){
	do{
	  job_ad2 = dapcollection->GetClassAd(key2);
	  if (!job_ad2) {
	    dprintf(D_ALWAYS, "No matching add!\n");
	    break; //no matching classad
	  }
	  else{
	    getValue(job_ad2, "reserve_id", current_reserve_id);
	    strncpy(current_reserve_id, strip_str(current_reserve_id), MAXSTR);
	    
	    if (!strcmp(current_reserve_id, reserve_id)) {
	      getValue(job_ad2, "dap_id", reserve_dap_id);
	      strncpy(reserve_dap_id, strip_str(reserve_dap_id), MAXSTR);
	    }
	  }
	}while (query.Next(key2));
      }
      
      key2 = "key = ";
      key2 += reserve_dap_id;
      dprintf(D_ALWAYS, "Reserve key: %s\n", reserve_dap_id);

      dapcollection->RemoveClassAd(key2);
    }//--

    //when request is comleted, remove it from the collection and 
    //log it to the history log file..
    if (!strcmp(dap_type, "reserve")){
      fclose(f);
      unlink(fname);
    }
    else{
      dapcollection->RemoveClassAd(key);
    }
    write_dap_log(historyfilename, "request_completed", "dap_id", dap_id);
    //    write_xml_log(xmllogfilename, job_ad, "\"request_completed\"");
    
    write_xml_user_log(userlogfilename, "MyType", "\"JobCompleteEvent\"", 
		       "EventTypeNumber", "5", 
		       "TerminatedNormally", "1",
		       "ReturnValue", "0",
		       "Cluster", dap_id,
		       "LogNotes", lognotes);
    
  }

  //---- if process failed (exit_status != 0) ----
  else{
    int num_attempts; 
    char attempts[MAXSTR], tempstr[MAXSTR];
    char                    current_status[MAXSTR];
    char use_protocol[MAXSTR], alt_protocols[MAXSTR];
    int protocol_num, max_protocols = 0;

    //get num_attempts
    if (getValue(job_ad, "num_attempts", attempts) != DAP_SUCCESS)
      num_attempts = 0;
    else
      num_attempts = atoi(attempts);

    printf("number of attempts = %d\n", num_attempts);
    

    if ( (f = fopen(fname, "r")) == NULL ){
      dprintf(D_ALWAYS, "No error details for pid: %d\n", pid);
    }
    else{
      fgets( linebuf, MAXSTR, f);
    }

    //    string sstatus = "";
    
    if (num_attempts + 1 < MAX_RETRY){
      //sstatus = "\"request_rescheduled\"";
      modify_s += "status = \"request_rescheduled\";";
    }
    
    else{
      //      sstatus = "\"request_failed\"";
      modify_s += "status = \"request_failed\";";
    }
    
    snprintf(tempstr, MAXSTR, 
	     "num_attempts = %d;",num_attempts + 1);
    
    modify_s +=  tempstr;
    modify_s += "error_code = \"";
    modify_s += linebuf;
    modify_s += "\";";

    //get which protocol to use
    getValue(job_ad, "use_protocol", use_protocol);
    protocol_num = atoi(use_protocol);
    
    getValue(job_ad, "alt_protocols", alt_protocols);
    
    if (strcmp(alt_protocols,"")) {
      strtok(alt_protocols, ",");   
      max_protocols ++;
    }
    
    while( strtok(NULL, ",") != NULL){
      max_protocols ++;
      printf("max protocols = %d\n", max_protocols);
      
    }
    
    //set which protocol to use
    if (protocol_num < max_protocols){
      snprintf(tempstr, MAXSTR, "use_protocol = %d;",protocol_num + 1);
      modify_s +=  tempstr;
    }
    else{
      snprintf(tempstr, MAXSTR, "use_protocol = %d;",0);
      modify_s +=  tempstr;
    }
    
    write_collection_log(dapcollection, dap_id, modify_s.c_str());
    //    write_xml_log(xmllogfilename, job_ad, sstatus.c_str());
      
    //when request is failed for good, remove it from the collection and 
    //log it to the history log file..
    if (num_attempts + 1 >=  MAX_RETRY){
      strncat(linebufstr, linebuf, MAXSTR-2);
      strcat(linebufstr, "\"");

      dapcollection->RemoveClassAd(key);

      write_dap_log(historyfilename, "request_failed", 
		      "dap_id", dap_id, "error_code", linebufstr);
            
      char exit_status_str[MAXSTR];
      snprintf(exit_status_str, MAXSTR, "%d", exit_status);
      
      write_xml_user_log(userlogfilename, "MyType", "\"JobCompletedEvent\"", 
			 "EventTypeNumber", "5", 
			 "TerminatedNormally", "1",
			 "ReturnValue", exit_status_str,
			 "Cluster", dap_id,
			 "LogNotes", lognotes);

      
    }
      
    if (f != NULL){
      fclose(f);
      unlink(fname);
    }
    
  }//else
  return DAP_SUCCESS;
}

/* ============================================================================
 * transfer dap reaper function
 * ==========================================================================*/
int transfer_dap_reaper(Service *,int pid,int exit_status)
{
  string modify_s = "";
  return dap_reaper(modify_s, pid, exit_status);
}

/* ============================================================================
 * reserve dap reaper function
 * ==========================================================================*/
int reserve_dap_reaper(Service *,int pid,int exit_status)
{
  string modify_s = "";
  return dap_reaper(modify_s, pid, exit_status);
  
}
/* ============================================================================
 * release dap reaper function
 * ==========================================================================*/
int release_dap_reaper(Service *,int pid,int exit_status)
{
  string modify_s = "";
  return dap_reaper(modify_s, pid, exit_status);
}














