// demo.C (2000-Nov-30)
//
// Peter Couvares <pfc@cs.wisc.edu>
// University of Wisconsin-Madison
//
// draws from example code in the ORBacus manual
//
// currently uses an ugly mix of C and C++ i/o methods...

#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <time.h>
#include <OB/CORBA.h>
#include <OB/CosNaming.h>

#include "hrm.h"
#include "hrm_call_back_impl.h"
#include "demo.h"
#include "util.h"

// file with stringified object reference for Fermi's CORBA name service
const char* NSRefFile = "fermi_ns.ref";

// file with the filenames to stage/transfer/etc.
const char* fsetFile = "fset";

// global vars needed by callback
HRM::HRMServant_var hrmServant;		// remote HRM server object
const char *userId = "nader2000";	// local user id

int
main( int argc, char* argv[] )
{
	int status = EXIT_SUCCESS;
	CORBA::ORB_var orb;

	// initialize & run the ORB
	try {
		if( DEBUG )
			fprintf( stdout, "Initializing the ORB...\n" );
		orb = CORBA::ORB_init( argc, argv );
		if( DEBUG )
			fprintf( stdout, "Running the ORB...\n" );
		status = run( orb );
	}
	catch( const CORBA::Exception& e ) {
		fprintf( stderr, "ERROR: caught CORBA exception, will exit...\n" );
		cerr << "===\n" << e << "\n===\n";
		status = EXIT_FAILURE;
	}

	// if ORB was successfully created, destroy it to release resources
	if( !CORBA::is_nil( orb ) ) {
		try {
			if( DEBUG )
				fprintf( stdout, "Destroying the ORB...\n" );
			orb -> destroy();
		}
		catch( const CORBA::Exception& e ) {
			fprintf( stderr, "ERROR: caught CORBA exception, will exit...\n" );
			cerr << "===\n" << e << "\n===\n";
			status = EXIT_FAILURE;
		}
	}
	return status;
}

int
run( CORBA::ORB_ptr orb )
{
	unsigned i;
	CORBA::Object_var hrmObj;

	//// set up the callback ////

	// obtain a reference to the Root POA, and then "narrow" it to the
	// right object type
	CORBA::Object_var poaObj = orb -> resolve_initial_references( "RootPOA" );
	PortableServer::POA_var rootPoa = PortableServer::POA::_narrow( poaObj );

	// using the Root POA, obtain a reference to its POA Manager
	PortableServer::POAManager_var manager = rootPoa -> the_POAManager();

	// create a servant for the callback
	client_call_back_impl* clientCallBackImpl = new client_call_back_impl();
	PortableServer::ServantBase_var servant = clientCallBackImpl;
	
	// incarnate a CORBA object for the callback
	HRM::client_call_back_var clientCallBack = clientCallBackImpl -> _this();

	//// connect to the remote HRM's ORB ////

	// get stringified object reference for the remote CORBA name service
	char* nsRefString = getNSRefString();

	if( nsRefString == NULL ) {
        fprintf( stderr,
				 "ERROR: can't get remote name service info; exiting...\n" );
		delete clientCallBackImpl;
        return EXIT_FAILURE;
	}

	// convert the stringified object reference to an actual object
	CORBA::Object_var nsObj = orb -> string_to_object( nsRefString );
	CosNaming::NamingContext_var nc =
		CosNaming::NamingContext::_narrow( nsObj.in() );
	
	// build the name we want to to resolve
	CosNaming::Name name;
	name.length(2);
	name[0].id = CORBA::string_dup( "ppdg" );
	name[0].kind = CORBA::string_dup( "" );
	name[1].id = CORBA::string_dup( "sam_hrm" );
	name[1].kind = CORBA::string_dup( "" );

	if( DEBUG )
		fprintf( stdout, "Contacting remote CORBA name service to resolve sam_hrm...\n" );

	// resolve the name to obtain an actual HRM object
	try {
		hrmObj = nc -> resolve( name );
	}
    catch( const CosNaming::NamingContext::NotFound& e ) {
		cerr << ": Got a `NotFound' exception (";
		switch( e.why ) {
		case CosNaming::NamingContext::missing_node:
			cerr << "missing node";
			break;

		case CosNaming::NamingContext::not_context:
			cerr << "not context";
			break;

		case CosNaming::NamingContext::not_object:
			cerr << "not object";
			break;
		}
		cerr << ")" << endl;
		return EXIT_FAILURE;
	}
    catch( const CosNaming::NamingContext::CannotProceed& ) {
		cerr << ": Got a `CannotProceed' exception" << endl;
		return EXIT_FAILURE;
	}
    catch( const CosNaming::NamingContext::InvalidName& ) {
		cerr << ": Got an `InvalidName' exception" << endl;
		return EXIT_FAILURE;
	}
    catch( const CosNaming::NamingContext::AlreadyBound& ) {
		cerr << ": Got an `AlreadyBound' exception" << endl;
		return EXIT_FAILURE;
	}
	catch(const CosNaming::NamingContext::NotEmpty&) {
		cerr << ": Got a `NotEmpty' exception" << endl;
		return EXIT_FAILURE;
	}

	hrmServant = HRM::HRMServant::_narrow( hrmObj );

	// ask what file transfer protocols the HRM supports
	if( DEBUG )
		fprintf( stdout, "Getting supported protocols...\n" );
	HRM::PROTOCOL_SET_T* protocolSet = hrmServant -> getSupportedProtocols();
	for( i=0; i < protocolSet -> length(); i++ ) {
		cout << "\t" << (*protocolSet)[i] << endl;
	}

	// get the file set to stage
	HRM::FIDSET_T* fset = buildFileset( fsetFile );

	char requestIdStr[1024];
	sprintf( requestIdStr, "%ld", time(NULL) );

    fprintf( stdout, "Submitting stage request \"%s\" for %lu file%s:\n",
			 requestIdStr, fset->length(), (fset->length() > 1 ? "s" : "") );
    for( i=0; i < fset->length(); i++ ) {
        cout << "\t" << (*fset)[i] << endl;
	}

	try {
		HRM::RETURN_STATUS_SET_T* statusSet = 
			hrmServant -> stage( userId, requestIdStr, clientCallBack, *fset );
	}
	catch( const CORBA::Exception& e) {
		fprintf( stderr, "ERROR: caught CORBA exception, will exit...\n" );
		cerr << "===\n" << e << "\n===\n";
		return 0;
	}

	if( DEBUG )
		fprintf( stdout, "Waiting for callbacks...\n" );

	// activate the POA Manager to allow the Root POA to start
	// processing requests
	manager -> activate();

	// inform the ORB that we are ready to accept requests
	orb -> run();

	return EXIT_SUCCESS;
}

HRM::FIDSET_T*
buildFileset( const char* file )
{
	HRM::FIDSET_T* fset = new HRM::FIDSET_T;
	char* s = new char[MAX_FNAME];
	int i = 0;

	FILE* fp = fopen( file, "r" );
	if( fp == NULL ) {
        fprintf( stderr, "ERROR: can't open file %s (%s)\n", file,
				 strerror( errno ) );
		return fset;
	}

	while( fgets( s, MAX_FNAME, fp ) ) {
		chomp( s );
		fset->length( i + 1 );		
		(*fset)[i] = s;
		i++;
		s = new char[MAX_FNAME];
	}
	
	free( s );
	fclose( fp );
	return fset;
}

char*
getNSRefString( void )
{
	// magic chars required at start of CORBA stringified object reference
	const char* objRefMagic = "IOR:";

	FILE* fp = fopen( NSRefFile, "r" );
	if( fp == NULL ) {
        fprintf( stderr, "ERROR: can't open file %s (%s)\n", NSRefFile,
				 strerror( errno ) );
		return NULL;
	}

	char* s = new char[MAX_STR_OBJ_REF];
	if( s == NULL ) {
		fprintf( stderr, "ERROR: out of memory\n" );
		return NULL;
	}

	fgets( s, MAX_STR_OBJ_REF, fp );
	chomp( s );
    fclose( fp );

	// check if it looks like a valid object reference
	if( strncmp( s, objRefMagic, strlen( objRefMagic ) ) ) {
		fprintf( stderr, "ERROR: invalid object reference in %s\n",
				 NSRefFile );
		return NULL;
	}

	return s;
}
