/***************************Copyright-DO-NOT-REMOVE-THIS-LINE**
 * CONDOR Copyright Notice
 *
 * See LICENSE.TXT for additional notices and disclaimers.
 *
 * Copyright (c)1990-1998 CONDOR Team, Computer Sciences Department, 
 * University of Wisconsin-Madison, Madison, WI.  All Rights Reserved.  
 * No use of the CONDOR Software Program Source Code is authorized 
 * without the express consent of the CONDOR Team.  For more information 
 * contact: CONDOR Team, Attention: Professor Miron Livny, 
 * 7367 Computer Sciences, 1210 W. Dayton St., Madison, WI 53706-1685, 
 * (608) 262-0856 or miron@cs.wisc.edu.
 *
 * U.S. Government Rights Restrictions: Use, duplication, or disclosure 
 * by the U.S. Government is subject to restrictions as set forth in 
 * subparagraph (c)(1)(ii) of The Rights in Technical Data and Computer 
 * Software clause at DFARS 252.227-7013 or subparagraphs (c)(1) and 
 * (2) of Commercial Computer Software-Restricted Rights at 48 CFR 
 * 52.227-19, as applicable, CONDOR Team, Attention: Professor Miron 
 * Livny, 7367 Computer Sciences, 1210 W. Dayton St., Madison, 
 * WI 53706-1685, (608) 262-0856 or miron@cs.wisc.edu.
****************************Copyright-DO-NOT-REMOVE-THIS-LINE**/

#include "condor_common.h"
#include "condor_debug.h"
#include "netbandwidthalloc.h"

int
NetworkBandwidthAllocator::
RequestBandwidth(double minMbps, double &maxMbps, const char source[],
				 const char destination[], bool canDeny)
{
	unsigned int src = ntohl(inet_addr(source));
	unsigned int dest = ntohl(inet_addr(destination));
	unsigned int current_addr;

	if (src == dest) return 1;	// loopback
	if (!capacityinfo) return 1; // no capacity set: allow anything
	if (!router) return -1;		// error if we can't route!

		// first, see how much we can allocate
	int rval = 1;
	for (current_addr = src;
		 rval > 0;
		 rval = router->NextHop(current_addr, src, dest)) {
		double capacity = 0.0;
		if (capacityinfo->GetCapacity(current_addr, capacity) < 0) continue;
		if (canDeny && capacity <= 0.0) {
			dprintf(D_FULLDEBUG, "netman: request for %.0f-%.0f Mbps (%s --> "
					"%s) refused because subnet %s has zero capacity\n",
					minMbps, maxMbps, source, destination,
					inet_itoa(current_addr));
			return 0;			// nothing allowed on this network!
		}
		double current = 0.0;
		network.lookup(current_addr, current);
		if (canDeny && current + minMbps > capacity) {
			return 0;			// network already overallocated
		} else if (current + maxMbps > capacity) {
			maxMbps = capacity - current;
		}
		if (!canDeny && maxMbps < minMbps) {
			maxMbps = minMbps;
		}
	}

	if (rval < 0) return -1;

		// now, allocate it
	rval = 1;
	for (current_addr = src;
		 rval > 0;
		 rval = router->NextHop(current_addr, src, dest)) {
		double capacity = 0.0;
		if (capacityinfo->GetCapacity(current_addr, capacity) < 0) continue;
		double current = 0.0;
		network.lookup(current_addr, current);
		current += maxMbps;
		network.remove(current_addr);
		network.insert(current_addr, current);
	}

	return 1;
}

int
NetworkBandwidthAllocator::
ReleaseBandwidth(double allocMbps, const char source[],
				 const char destination[])
{
	unsigned int src = ntohl(inet_addr(source));
	unsigned int dest = ntohl(inet_addr(destination));

	if (src == dest) return 0;	// loopback
	if (!capacityinfo) return 0; // no capacity set: allow anything
	if (!router) return -1;		// error if we can't route!

	int rval = 1;
	for (unsigned int current_addr = src;
		 rval > 0;
		 rval = router->NextHop(current_addr, src, dest)) {
		double capacity = 0.0;
		if (capacityinfo->GetCapacity(current_addr, capacity) < 0) continue;
		double current = 0.0;
		network.lookup(current_addr, current);
		current -= allocMbps;
		network.remove(current_addr);
		network.insert(current_addr, current);
	}

	if (rval < 0) return -1;

	return 0;
}

int
NetworkBandwidthAllocator::
Bandwidth(double &availMbps, double &totMbps, const char source[],
		  const char destination[])
{
	unsigned int src = ntohl(inet_addr(source));
	unsigned int dest = ntohl(inet_addr(destination));

	if (src == dest) return -1;	// loopback
	if (!capacityinfo) return -1; // no capacity set
	if (!router) return -1;		// error if we can't route!

		// first, see how much we can allocate
	availMbps = totMbps = DBL_MAX;
	int rval = 1;
	for (unsigned int current_addr = src;
		 rval > 0;
		 rval = router->NextHop(current_addr, src, dest)) {
		double capacity = 0.0;
		if (capacityinfo->GetCapacity(current_addr, capacity) < 0) continue;
		if (capacity <= 0.0) {
			availMbps = totMbps = 0.0;
			return 0;			// nothing allowed on this network!
		}
		if (totMbps > capacity) {
			totMbps = capacity;
		}
		if (availMbps > 0.0) {
			double current = 0.0;
			network.lookup(current_addr, current);
			double avail = capacity - current;
			if (avail <= 0.0) {
				availMbps = 0.0;	// network already overallocated
			} else if (availMbps > avail) {
				availMbps = avail;
			}
		}
	}

	if (rval < 0) return -1;

	return 0;
}

