/***************************Copyright-DO-NOT-REMOVE-THIS-LINE**
 * CONDOR Copyright Notice
 *
 * See LICENSE.TXT for additional notices and disclaimers.
 *
 * Copyright (c)1990-1998 CONDOR Team, Computer Sciences Department, 
 * University of Wisconsin-Madison, Madison, WI.  All Rights Reserved.  
 * No use of the CONDOR Software Program Source Code is authorized 
 * without the express consent of the CONDOR Team.  For more information 
 * contact: CONDOR Team, Attention: Professor Miron Livny, 
 * 7367 Computer Sciences, 1210 W. Dayton St., Madison, WI 53706-1685, 
 * (608) 262-0856 or miron@cs.wisc.edu.
 *
 * U.S. Government Rights Restrictions: Use, duplication, or disclosure 
 * by the U.S. Government is subject to restrictions as set forth in 
 * subparagraph (c)(1)(ii) of The Rights in Technical Data and Computer 
 * Software clause at DFARS 252.227-7013 or subparagraphs (c)(1) and 
 * (2) of Commercial Computer Software-Restricted Rights at 48 CFR 
 * 52.227-19, as applicable, CONDOR Team, Attention: Professor Miron 
 * Livny, 7367 Computer Sciences, 1210 W. Dayton St., Madison, 
 * WI 53706-1685, (608) 262-0856 or miron@cs.wisc.edu.
****************************Copyright-DO-NOT-REMOVE-THIS-LINE**/

#include "condor_common.h"
#include "condor_debug.h"
#include "netcapacityalloc.h"

NetworkCapacityAllocator::~NetworkCapacityAllocator()
{
	NetworkRecord *current;
	network.startIterations();
	while (network.iterate(current)) {
		delete current;
	}
	network.clear();
}

int
NetworkCapacityAllocator::
RequestCapacity(double bytes, const char source[],
				const char destination[], bool canDeny,
				bool willReturn, bool goodputEligible)
{
	unsigned int src = ntohl(inet_addr(source));
	unsigned int dest = ntohl(inet_addr(destination));

	if (src == dest) return 1;	// loopback
	if (!capacityinfo) return 1; // no capacity set: allow anything
	if (!router) return -1;		// error if we can't route!

	time_t current_time_t = time(0);
	double current_time = (double)current_time_t;
	int rval = 1;
	for (unsigned int current_addr = src;
		 rval > 0;
		 rval = router->NextHop(current_addr, src, dest)) {
		double capacity = 0.0;
		if (capacityinfo->GetCapacity(current_addr, capacity) < 0) continue;
		if (capacity <= 0.0) {
			dprintf(D_FULLDEBUG, "netman: request for %.0f bytes (%s --> %s) "
					"refused because subnet %s has zero capacity\n",
					bytes, source, destination, inet_itoa(current_addr));
			Abort();
			return 0; // nothing allowed on this network!
		}
		NetworkRecord *current = NULL;
		if (network.lookup(current_addr, current) < 0) {
			current = new NetworkRecord(current_addr);
			if (network.insert(current_addr, current) < 0) {
				dprintf(D_ALWAYS, "NetworkCapacityAllocator::RequestCapacity "
						"hashtable insert failed!\n");
				return -1;
			}
		}

			// deny request if this network is already allocated
		if (canDeny &&
			((current->allocated/131072.0 >= capacity*horizon) ||
			 (current->reserved >= current_time + horizon))) {
			if (goodputEligible && goodput_reservation > 0.0 &&
				(current->goodput_reserved < current_time + horizon)) {
				current->uses_res_goodput_cap = true;
			} else {
				dprintf(D_FULLDEBUG, "netman: request for %.0f bytes "
						"(%s --> %s) refused because subnet %s has reached "
						"capacity\n",
						bytes, source, destination, inet_itoa(current->addr));
				current->denied = true;
				Abort();
				return 0;
			}
		}

			// otherwise, add pending reservation
		if (current->pending == 0.0) {
			if (transaction_len == max_transaction_len) {
				max_transaction_len += 100;
				transaction = (NetworkRecord **)
					realloc(transaction,
							sizeof(NetworkRecord *)*max_transaction_len);
			}
			transaction[transaction_len++] = current;
		}
		current->pending += bytes;
		if (willReturn) current->pending_standing += bytes;
	}

	if (rval < 0) return -1;

	return 1;
}

int
NetworkCapacityAllocator::Commit(const char user[],
								 double *bytes, int *num_hops)
{
	double current_time = (double)time(0);

	if (bytes) *bytes = 0.0;
	if (num_hops) *num_hops = 0;

	for (int i=0; i < transaction_len; i++) {
		NetworkRecord *current = transaction[i];
		if (current->pending > 0.0) {
			double capacity = 0.0;
			if (capacityinfo->GetCapacity(current->addr, capacity) < 0 ||
				capacity <= 0.0) continue;
			double usage = current->pending/131072.0/capacity;
			if (current->reserved < current_time) {
				current->reserved = current_time;
			}
			current->reserved += usage;
			if (current->uses_res_goodput_cap && goodput_reservation > 0.0) {
					// scale usage according to goodput reservation
				double goodput_usage = usage / goodput_reservation;
				if (current->goodput_reserved < current_time) {
					current->goodput_reserved = current_time;
				}
				current->goodput_reserved += goodput_usage;
			}
			dprintf(D_BANDWIDTH,
					"netman: alloc %.0f (%.0f) KB on %s\n",
					current->pending/1024, current->pending_standing/1024,
					inet_itoa(current->addr));
			if (bytes) {
				if (*bytes < current->pending) {
					*bytes = current->pending;
				}
			}
			if (num_hops) (*num_hops)++;
			current->allocated += current->pending_standing;
			current->pending = 0.0;
			current->pending_standing = 0.0;
			if (netusage) {
				netusage->AddUsage(user, current->addr, usage);
			}
		}
		current->denied = false;
		current->uses_res_goodput_cap = false;
	}
	transaction_len = 0;

	return 0;
}

int
NetworkCapacityAllocator::Abort()
{
	for (int i=0; i < transaction_len; i++) {
		transaction[i]->pending = 0.0;
		transaction[i]->pending_standing = 0.0;
		transaction[i]->uses_res_goodput_cap = false;
	}
	transaction_len = 0;

	return 0;
}

int
NetworkCapacityAllocator::GetGoodputInfo(double *usage, double *bytes,
										 int *num_hops)
{
	double current_time = (double)time(0);

	bool uses_res_goodput_cap = false;
	if (usage) *usage = 0.0;
	if (bytes) *bytes = 0.0;
	if (num_hops) *num_hops = 0;

	for (int i=0; i < transaction_len; i++) {
		NetworkRecord *current = transaction[i];
		if (current->pending > 0.0 &&
			current->uses_res_goodput_cap && goodput_reservation > 0.0) {
			uses_res_goodput_cap = true;
			double capacity = 0.0;
			if (capacityinfo->GetCapacity(current->addr, capacity) < 0 ||
				capacity <= 0.0) continue;
			double this_usage =
				current->pending/131072.0/capacity/goodput_reservation/horizon;
			if (usage && this_usage > *usage) {
				*usage = this_usage;
			}
			if (bytes && *bytes < current->pending) {
				*bytes = current->pending;
			}
			if (num_hops) (*num_hops)++;
		}
	}

	return uses_res_goodput_cap;
}

int
NetworkCapacityAllocator::ReturnCapacity(double bytes, const char source[],
										 const char destination[])
{
	unsigned int src = ntohl(inet_addr(source));
	unsigned int dest = ntohl(inet_addr(destination));

	if (src == dest) return 1;	// loopback

	int rval = 1;
	for (unsigned int current_addr = src;
		 rval > 0;
		 rval = router->NextHop(current_addr, src, dest)) {
		NetworkRecord *current = NULL;
		if (network.lookup(current_addr, current) < 0) continue;
		if (current->allocated < bytes) {
			dprintf(D_BANDWIDTH, "netman: return %.0f KB on %s\n",
					current->allocated/1024, inet_itoa(current->addr));
			current->allocated = 0.0;
		} else {
			dprintf(D_BANDWIDTH, "netman: return %.0f KB on %s\n",
					bytes/1024, inet_itoa(current->addr));
			current->allocated -= bytes;
		}
	}

	return rval;
}

int
NetworkCapacityAllocator::ShowDeniedRequests(int debug_mask)
{
	NetworkRecord *current;
	network.startIterations();
	while (network.iterate(current)) {
		if (current->denied) {
			dprintf(debug_mask, "netman: capacity denied on %s\n",
					inet_itoa(current->addr));
		}
		current->denied = false;
	}
	
	return 0;
}

