/***************************Copyright-DO-NOT-REMOVE-THIS-LINE**
 * CONDOR Copyright Notice
 *
 * See LICENSE.TXT for additional notices and disclaimers.
 *
 * Copyright (c)1990-1998 CONDOR Team, Computer Sciences Department, 
 * University of Wisconsin-Madison, Madison, WI.  All Rights Reserved.  
 * No use of the CONDOR Software Program Source Code is authorized 
 * without the express consent of the CONDOR Team.  For more information 
 * contact: CONDOR Team, Attention: Professor Miron Livny, 
 * 7367 Computer Sciences, 1210 W. Dayton St., Madison, WI 53706-1685, 
 * (608) 262-0856 or miron@cs.wisc.edu.
 *
 * U.S. Government Rights Restrictions: Use, duplication, or disclosure 
 * by the U.S. Government is subject to restrictions as set forth in 
 * subparagraph (c)(1)(ii) of The Rights in Technical Data and Computer 
 * Software clause at DFARS 252.227-7013 or subparagraphs (c)(1) and 
 * (2) of Commercial Computer Software-Restricted Rights at 48 CFR 
 * 52.227-19, as applicable, CONDOR Team, Attention: Professor Miron 
 * Livny, 7367 Computer Sciences, 1210 W. Dayton St., Madison, 
 * WI 53706-1685, (608) 262-0856 or miron@cs.wisc.edu.
****************************Copyright-DO-NOT-REMOVE-THIS-LINE**/
#ifndef __NETWORK_CAPACITY_ALLOC_H__
#define __NETWORK_CAPACITY_ALLOC_H__

#include "HashTable.h"
#include "router.h"
#include "netcapacity.h"
#include "netusage.h"

class NetworkCapacityAllocator {
public:
	NetworkCapacityAllocator() : horizon(0.0), goodput_reservation(0.0),
		network(255, addrhash),
		transaction(NULL), transaction_len(0), max_transaction_len(0),
		router(NULL), capacityinfo(NULL), netusage(NULL) {}
	~NetworkCapacityAllocator();

	void BindRouter(Router *r) { router = r; }
	void BindCapacity(NetworkCapacity *c) { capacityinfo = c; } 
	void BindUsage(NetworkUsage *n) { netusage = n; }

	// allocate future bandwidth up to horizon and deny allocations
	// once horizon is reached
	int SetHorizon(int seconds)
		{ int old = (int)horizon; horizon = (double)seconds; return old; }

	int GetHorizon() { return (int)horizon; }

	// reserve specified percentage (between 0.0 and 1.0) of capacity
	// on each subnet for "goodput" requests; for now this parameter
	// is global across all network segments, but at some point it
	// should probably be configurable per network; returns previous
	// value
	double ReserveForGoodput(double r) {
		double old = goodput_reservation; goodput_reservation = r; return old;
	}

	// request permission to transfer specified number of bytes from
	// the source to the destination network endpoints (IP addresses
	// in dotted notation); if canDeny is false, the capacity will
	// always be reserved, even if it requires overallocating the
	// network; if willReturn is true, the capacity will be reserved
	// until it is returned by a call to ReturnCapacity(); returns 1
	// if request is granted, 0 if denied, -1 if error occurred;
	// requests are granted in transactions, which should be either
	// committed or aborted using the Commit() or Abort() methods; if
	// the request is denied or an error occurs, the current
	// transaction is aborted
	int RequestCapacity(double bytes, const char source[],
						const char destination[], bool canDeny=true,
						bool willReturn=false, bool goodputEligible=false);

	// commit current transaction (and begin a new one); network usage
	// is expected to begin immediately after commit (i.e., the
	// timestamp is set at commit time); returns 0 on success, -1 on
	// error; bytes and num_hops are value-result parameters which
	// specify the maximum number of bytes sent over one network link
	// and the number of separate network subnets involved in the
	// allocation respectively; if this object is bound to a NetworkUsage
	// object, then the committed usage is reported to that object for
	// the specified user
	int Commit(const char user[]="", double *bytes=NULL, int *num_hops=NULL);

	// abort current transaction (and begin a new one); returns 0 on
	// success, -1 on error
	int Abort();

	// function to help the caller determine if a goodput allocation
	// should be committed; returns 1 if the current transaction uses
	// reserved goodput capacity on any subnets, 0 if not, and -1 on
	// error; if goodput capacity is used, the maximum percentage of
	// goodput capacity used across the subnets is returned in usage,
	// the maximum bytes sent is returned in bytes, and the number of
	// network subnets where goodput capacity was allocated is
	// returned in num_hops
	int GetGoodputInfo(double *usage=NULL, double *bytes=NULL,
					   int *num_hops=NULL);

	// deallocate the specified number of bytes from a previous
	// capacity request so new allocations can be made; since
	// allocated network capacity is not guaranteed, we require
	// feedback from the caller to know when the allocated capacity
	// has been used; this way, we don't keep allocating more capacity
	// on a network that's delivering less capacity than we are
	// configured to expect; returns 0 on success, -1 on error
	int ReturnCapacity(double bytes, const char source[],
					   const char destination[]);

	// dprintf() a record for each network where requests were denied
	// since the last call to this method or to Commit(), using the
	// debug_mask given; returns 0 on success, -1 on error
	int ShowDeniedRequests(int debug_mask);

private:
	struct NetworkRecord {
		NetworkRecord() : addr(0), pending(0.0),
			pending_standing(0.0), reserved(0.0), allocated(0.0), 
			denied(false), active(true), uses_res_goodput_cap(false)  {}
		NetworkRecord(unsigned int a) : addr(a), pending(0.0),
			pending_standing(0.0), reserved(0.0), allocated(0.0),
			denied(false), active(true), uses_res_goodput_cap(false) {}
		unsigned int addr;
		double pending;			// pending allocation (in bytes)
		double pending_standing; // pending standing allocation (in bytes)
		double reserved;		// time_t when original allocs expire
		double goodput_reserved; // time_t when goodput allocs expire
		double allocated;		// standing allocated capacity (in bytes)
		bool denied;			// any requests for this network denied?
		bool active;			// flag to remove old recs after reconfig
		bool uses_res_goodput_cap;
	};

	double horizon;
	double goodput_reservation;
	HashTable<unsigned int, NetworkRecord *> network;
	NetworkRecord **transaction;
	int transaction_len, max_transaction_len;

	Router *router;
	NetworkCapacity *capacityinfo;
	NetworkUsage *netusage;
};

#endif
