/***************************Copyright-DO-NOT-REMOVE-THIS-LINE**
 * CONDOR Copyright Notice
 *
 * See LICENSE.TXT for additional notices and disclaimers.
 *
 * Copyright (c)1990-1998 CONDOR Team, Computer Sciences Department, 
 * University of Wisconsin-Madison, Madison, WI.  All Rights Reserved.  
 * No use of the CONDOR Software Program Source Code is authorized 
 * without the express consent of the CONDOR Team.  For more information 
 * contact: CONDOR Team, Attention: Professor Miron Livny, 
 * 7367 Computer Sciences, 1210 W. Dayton St., Madison, WI 53706-1685, 
 * (608) 262-0856 or miron@cs.wisc.edu.
 *
 * U.S. Government Rights Restrictions: Use, duplication, or disclosure 
 * by the U.S. Government is subject to restrictions as set forth in 
 * subparagraph (c)(1)(ii) of The Rights in Technical Data and Computer 
 * Software clause at DFARS 252.227-7013 or subparagraphs (c)(1) and 
 * (2) of Commercial Computer Software-Restricted Rights at 48 CFR 
 * 52.227-19, as applicable, CONDOR Team, Attention: Professor Miron 
 * Livny, 7367 Computer Sciences, 1210 W. Dayton St., Madison, 
 * WI 53706-1685, (608) 262-0856 or miron@cs.wisc.edu.
****************************Copyright-DO-NOT-REMOVE-THIS-LINE**/

#include "condor_common.h"
#include "condor_config.h"
#include "condor_debug.h"
#include "condor_attributes.h"
#include "netman.h"

static int
NetAllocHash(const struct NetworkManager::NetAllocKey& key, int numBuckets) 
{
	int hashval = 0;
	const unsigned char *byte = (const unsigned char *)&key.startdIP;
	for (unsigned int i=0; i < sizeof(unsigned int); i++) {
		hashval += *byte;
		byte++;
	}
	hashval += key.vm;
	hashval *= key.type;
	return ((unsigned int)hashval % numBuckets);
}

NetworkManager::NetworkManager() : active(false),
								   NetAllocations(501, NetAllocHash),
								   NetAllocationKeys(201, hashFunction)
{
	netcapalloc.BindRouter(&router);
	netcapalloc.BindCapacity(&capinfo);
	netcapalloc.BindUsage(&netusage);
	netusagealloc.BindRouter(&router);
	netusagealloc.BindCapacity(&capinfo);
	netusagealloc.BindUsage(&netusage);
}

NetworkManager::~NetworkManager()
{
}

void
NetworkManager::Config()
{
	char *tmp = NULL;

	active = true;
	tmp = param("NETWORK_ROUTING_INFO");
	if (tmp) {
		if (router.LoadRoutingTable(tmp) < 0) {
			dprintf(D_ALWAYS, "NetMan: Failed to load routing table!\n");
			free(tmp);
			active = false;
			return;
		}
		free(tmp);
	} else {
		active = false;
		return;
	}

	tmp = param("NETWORK_CAPACITY_INFO");
	if (tmp) {
		if (capinfo.LoadNetworkCapacity(tmp) < 0) {
			dprintf(D_ALWAYS,
					"NetMan: Failed to load network capacity info!\n");
			free(tmp);
			active = false;
			return;
		}
		free(tmp);
	} else {
		active = false;
		return;
	}

	tmp = param("NETWORK_HORIZON");
	if (tmp) {
		horizon = atoi(tmp);
		free(tmp);
		dprintf(D_ALWAYS, "NETWORK_HORIZON = %d\n", horizon);
	} else {
		horizon = 300;
	}
	netcapalloc.SetHorizon(horizon);

	tmp = param("NETWORK_CAPACITY_RESERVED_FOR_GOODPUT");
	if (tmp) {
		goodput_reservation = atof(tmp);
		free(tmp);
		dprintf(D_ALWAYS, "NETWORK_CAPACITY_RESERVED_FOR_GOODPUT = %.2f\n",
				goodput_reservation);
	} else {
		goodput_reservation = 0.0;
	}
	netcapalloc.ReserveForGoodput(goodput_reservation);

	tmp = param("MAX_GOODPUT_NETWORK_CAPACITY_PER_JOB");
	if (tmp) {
		goodput_job_limit = atof(tmp);
		free(tmp);
		dprintf(D_ALWAYS, "MAX_GOODPUT_NETWORK_CAPACITY_PER_JOB = %.2f\n",
				goodput_job_limit);
	} else {
		goodput_job_limit = 0.0;
	}

	tmp = param("NETWORK_USAGE_HORIZON");
	if (tmp) {
		usage_horizon = atoi(tmp);
		free(tmp);
		dprintf(D_ALWAYS, "NETWORK_USAGE_HORIZON = %d\n", usage_horizon);
	} else {
		usage_horizon = 3600;
	}
	netusagealloc.SetHorizon(usage_horizon);

	if (active) {
		dprintf(D_ALWAYS, "Network Manager configured and active.\n");
	}
}

int
NetworkManager::HandleNetworkRequest(Stream *stream)
{
	if (!active) return -1;

	ClassAd request;
	int bytes, total_bytes = 0;
	int bytes_sent = 0, bytes_recvd = 0;
	int prev_send_est = 0, prev_recv_est = 0;
	char source[100], destination[100], xfer_type[100], user[100], rhost[100];

	int num_requests, force, rval = 1;
	int cluster=0, proc=0;

	if (!stream->code(num_requests)) return -1;

	// Abort any old transactions.
	netcapalloc.Abort();

	for (int i=0; rval == 1 && i < num_requests; i++) {
        if (!getOldClassAd(stream, request)) {
			dprintf(D_ALWAYS, "Failed to read ClassAd %d in "
					"REQUEST_NETWORK_commandHandler!\n", i);
			return -1;
		}
		// get ATTR_TRANSFER_TYPE and ATTR_USER from first ClassAd only
		// other attributes apply to each request

        // Should convert to use string instead
		if ((i == 0 &&
			 (!request.EvaluateAttrString(ATTR_TRANSFER_TYPE, xfer_type, 100) ||
			  !request.EvaluateAttrString(ATTR_USER, user, 100))) ||
			(!request.EvaluateAttrInt(ATTR_FORCE, force) ||
			 !request.EvaluateAttrInt(ATTR_REQUESTED_CAPACITY, bytes) ||
			 !request.EvaluateAttrString(ATTR_SOURCE, source, 100) ||
			 !request.EvaluateAttrString(ATTR_DESTINATION, destination, 100))) {
			dprintf(D_ALWAYS,
					"Malformed ClassAd in REQUEST_NETWORK_commandHandler!\n");
            request.Puke();
			return -1;
		}
		total_bytes += bytes;
		// ATTR_REMOTE_HOST is optionally specified in first ClassAd
		if (i == 0) {
			if (!request.EvaluateAttrString(ATTR_REMOTE_HOST, rhost, 100)) {
				strcpy(rhost, "<0.0.0.0:0>");
			}
			request.EvaluateAttrInt(ATTR_CLUSTER_ID, cluster);	
			request.EvaluateAttrInt(ATTR_PROC_ID, proc);
				// Log usage information for RemoteSyscalls
			if (strcmp(xfer_type, "RemoteSyscalls") == 0 && 
				request.EvaluateAttrInt(ATTR_BYTES_SENT, bytes_sent) &&
				request.EvaluateAttrInt(ATTR_BYTES_RECVD, bytes_recvd) &&
				request.EvaluateAttrInt(ATTR_PREV_SEND_ESTIMATE,
									  prev_send_est) &&
				request.EvaluateAttrInt(ATTR_PREV_RECV_ESTIMATE,
									  prev_recv_est)) {
				dprintf(D_BANDWIDTH,
						"RSC: %d+%d (%d+%d) %s <%s:%d> %s\n",
						bytes_sent/1024, bytes_recvd/1024,
						prev_send_est/1024, prev_recv_est/1024,
						user, ((Sock *)stream)->endpoint_ip_str(),
						((Sock *)stream)->endpoint_port(), rhost);
			}
		}
		bool willReturn = false;
		if (strcmp(xfer_type, "RemoteSyscalls") != 0) {
				// If this is a bulk transfer, then we should check to
				// see if we've already got an allocation record for
				// it.  If we do, we need to update the record and
				// allocate the diff.  Otherwise, we should create a
				// new record.
			willReturn = true;	// require notification for bulk xfer compltn
			NetAllocKey key;
			key.vm = 0;
			request.EvaluateAttrInt(ATTR_VIRTUAL_MACHINE_ID, key.vm);
			char xferkeystr[64];
			xferkeystr[0] = '\0';
			request.EvaluateAttrString(ATTR_TRANSFER_KEY, xferkeystr, 64);
			HashKey xferkey(xferkeystr);
			int endpoint;
			if (strcmp(xfer_type, "InitialCheckpoint") == 0) {
				key.startdIP = inet_addr(destination);
				endpoint = inet_addr(source);
				key.type = NET_ALLOC_ICKPT;
			} else if (strcmp(xfer_type, "CheckpointRestart") == 0) {
				key.startdIP = inet_addr(destination);
				endpoint = inet_addr(source);
				key.type = NET_ALLOC_RESTART;
			} else if (strcmp(xfer_type, "PeriodicCheckpoint") == 0) {
				key.startdIP = inet_addr(source);
				endpoint = inet_addr(destination);
				key.type = NET_ALLOC_PCKPT;
			} else if (strcmp(xfer_type, "VacateCheckpoint") == 0) {
				key.startdIP = inet_addr(source);
				endpoint = inet_addr(destination);
				key.type = NET_ALLOC_VCKPT;
			} else if (strcmp(xfer_type, "CkptServerUpdate") == 0) {
					// need to use xferkey to find real NetAllocKey!
				if (!xferkeystr[0]) {
					dprintf(D_ALWAYS,
							"TransferKey undefined in CkptServerUpdate!\n");
					return -1;
				}
				if (NetAllocationKeys.lookup(xferkey, key) < 0) {
					dprintf(D_ALWAYS, "Unknown TransferKey %s in "
							"REQUEST_NETWORK_commandHandler!\n", xferkeystr);
					return -1;
				}
					// endpoint is the checkpoint server; check source
					// and destination and use the one that's not equal to
					// the startd IP address
				endpoint = inet_addr(source);
				if (endpoint == key.startdIP) {
					endpoint = inet_addr(destination);
				}
			} else {
				dprintf(D_ALWAYS, "Unknown transfer type %s in "
						"REQUEST_NETWORK_commandHandler!\n", xfer_type);
				request.fPrint(stdout);
				printf("\n");
				return -1;
			}
			NetAlloc *rec = NULL;
			NetAllocations.lookup(key, rec);
			if (rec) {
				struct in_addr sin;
				memcpy(&sin, &key.startdIP, sizeof(struct in_addr));
				dprintf(D_FULLDEBUG,
						"netalloc: key lookup (vm%d@%s %d) succeeded!\n",
						key.vm, inet_ntoa(sin), key.type);
					// make sure we have a stashed xferkey
				if (xferkeystr[0]) {
					NetAllocKey tmpkey;
					if (NetAllocationKeys.lookup(xferkey, tmpkey) < 0) {
						if (NetAllocationKeys.insert(xferkey, key) < 0) {
							dprintf(D_ALWAYS,
									"netalloc: hashtable insert failed!\n");
						}
					}
				}
			} else {
				struct in_addr sin;
				memcpy(&sin, &key.startdIP, sizeof(struct in_addr));
				dprintf(D_FULLDEBUG,
						"netalloc: key lookup (vm%d@%s %d) failed!\n",
						key.vm, inet_ntoa(sin), key.type);
			}
			if (rec && rec->endpoint == endpoint) { // found one
				int diff = bytes - rec->bytes;
				rec->bytes = bytes;
				bytes = diff;
				if (rec->bytes <= 0) {
					dprintf(D_FULLDEBUG,
							"netalloc: removing allocation record!\n");
					if (NetAllocations.remove(key) < 0) {
						dprintf(D_ALWAYS,
								"netalloc: hashtable remove failed!\n");
					}
					delete rec;
						// remove any stashed xferkey
					NetAllocationKeys.remove(xferkey);
				} else {
					rec->timestamp = time(0);
					rec->expiration_time = rec->timestamp+horizon;
				}
			} else if (bytes > 0) {
				if (rec) {
					rec->expiration_time = time(0);
					rec = NULL;
					CleanupOldNetAllocations();
				}
				rec = new NetAlloc();
				dprintf(D_FULLDEBUG,
						"netalloc: inserting new allocation record!\n");
				if (NetAllocations.insert(key, rec) < 0 || 
					NetAllocationKeys.insert(xferkey, key) < 0) {
					dprintf(D_ALWAYS, "netalloc: hashtable insert failed!\n");
				}
				rec->bytes = bytes;
				rec->endpoint = endpoint;
				rec->timestamp = time(0);
				rec->expiration_time = rec->timestamp+horizon;
			}
		}

		if (bytes > 0) {
			rval = netcapalloc.RequestCapacity(bytes, source, destination,
											   (force == 0), willReturn);
		} else if (bytes < 0) {
			rval = netcapalloc.ReturnCapacity(-bytes, source, destination);
			if (rval == 0) {
				rval = 1;		// indicate success with rval==1
			}
		}
	}
	stream->end_of_message();

	if (rval == 1) {
		dprintf(D_MATCH, "      %s %d.%d %s %d bytes <%s:%d> %s\n",
				xfer_type, cluster, proc, user, total_bytes,
				((Sock *)stream)->endpoint_ip_str(),
				((Sock *)stream)->endpoint_port(), rhost);
		rval = netcapalloc.Commit(user);
		if (rval == 0) rval = 1;
	}

	if (stream->type() == Stream::reli_sock) {
		stream->encode();
		stream->code(rval);
		stream->end_of_message();
	}

	return 0;
}

int
NetworkManager::
RequestPlacement(const char user[], double share,
				 const char scheddIP[], int executableSize,
				 const char lastCkptServerIP[], int ckptSize,
				 ClassAd &request, ClassAd &offer,
				 ClassAd *preemptJob)
{
	if (!active) return 2;		// allow anything if not configured

	netcapalloc.Abort();		// initialize pending allocation

	int rval = 0;

	// find startdIP
	char startdIPbuf[128];
	if (!offer.EvaluateAttrString(ATTR_STARTD_IP_ADDR, startdIPbuf, 128)) {
		dprintf(D_ALWAYS, "      Could not lookup %s\n", ATTR_STARTD_IP_ADDR);
		return -1;
	}
	// remove port part from IP addresses
	char *colon = strchr(startdIPbuf, ':');
	if (!colon) {
		dprintf(D_ALWAYS, "      Invalid %s: %s\n", ATTR_STARTD_IP_ADDR,
				startdIPbuf);
		return -1;
	}
	*colon = '\0';
	char *startdIP = startdIPbuf+1; // skip the leading '<'

		// only use reserved "goodput" capacity if we're not preempting
		// a job, since we're reserving the capacity to reduce idle time
		// in the pool, and preempting a job won't help with that goal
	bool goodputEligible =
		(preemptJob == NULL && !offer.Lookup(ATTR_REMOTE_USER));

	// allocate bandwidth to transfer the executable
	if (executableSize > 0) {
		rval = RequestCapacity(user, share, double(executableSize)*1024,
							   scheddIP, startdIP, true, true,
							   goodputEligible);
		if (rval < 2) return rval;
	}

	// allocate bandwidth for RSC
	int serverTime = 0;
	double bytesSent = 0.0, bytesRcvd = 0.0;
	request.EvaluateAttrInt(ATTR_SERVER_TIME, serverTime);
	request.EvaluateAttrReal(ATTR_RSC_BYTES_SENT, bytesSent);
	request.EvaluateAttrReal(ATTR_RSC_BYTES_RECVD, bytesRcvd);
	if (serverTime > 0) {
		if (bytesSent > 0.0) {
			rval = RequestCapacity(user, share, bytesSent/serverTime, scheddIP,
								   startdIP, goodputEligible);
			if (rval < 2) return rval;
		}
		if (bytesRcvd > 0.0) {
			rval = RequestCapacity(user, share, bytesRcvd/serverTime, startdIP,
								   scheddIP, goodputEligible);
			if (rval < 2) return rval;
		}
	}

	// allocate bandwidth for the restart checkpoint transfer
	if (ckptSize > 0 && lastCkptServerIP[0]) {
		rval = RequestCapacity(user, share, double(ckptSize)*1024,
							   lastCkptServerIP, startdIP, true, true,
							   goodputEligible);
		if (rval < 2) return rval;
	}

	// allocate (estimated) bandwidth for preemption checkpoint transfer
	bool didPreemptAlloc = false;
	int universe = CONDOR_UNIVERSE_STANDARD;
	int vckptSize = 0;
	char ckptServer[128], ckptServerIP[16];
	ckptServerIP[0] = '\0';
	if (!preemptJob || !preemptJob->EvaluateAttrInt(ATTR_JOB_UNIVERSE,
                                                        universe)) {
		offer.EvaluateAttrInt(ATTR_JOB_UNIVERSE, universe);
	}
	if (universe == CONDOR_UNIVERSE_STANDARD) {
		int executableSize = 0;
		if (!preemptJob ||
			!preemptJob->EvaluateAttrInt(ATTR_IMAGE_SIZE, vckptSize)) {
			offer.EvaluateAttrInt(ATTR_IMAGE_SIZE, vckptSize);
		}
		if (!preemptJob ||
			!preemptJob->EvaluateAttrInt(ATTR_EXECUTABLE_SIZE, executableSize)) {
			offer.EvaluateAttrInt(ATTR_EXECUTABLE_SIZE, executableSize);
		}
		vckptSize -= executableSize;	// imagesize = ckptsize + executablesz
		if (vckptSize > 0) {
			ckptServer[0] = '\0';
            // Not quite sure about this conversion to new classad. Hao
            // original code is preemptJob->EvalString(...)
			if (preemptJob && !preemptJob->EvaluateAttrString(ATTR_CKPT_SERVER,
                                                              ckptServer, 128)) {
				preemptJob->EvaluateAttrString(ATTR_MACHINE, ckptServer, 128);
			} else if (!offer.EvaluateAttrString(ATTR_CKPT_SERVER, ckptServer,128)) {
				offer.EvaluateAttrString(ATTR_CLIENT_MACHINE, ckptServer, 128);
			}
			if (ckptServer[0]) {
				struct hostent *hp = gethostbyname(ckptServer);
				if (!hp) {
					dprintf(D_ALWAYS, "      DNS lookup for %s %s failed!\n",
							ATTR_CKPT_SERVER, ckptServer);
					return -1;
				}
				strcpy(ckptServerIP,
					   inet_ntoa(*((struct in_addr *)hp->h_addr)));
				rval = RequestCapacity(user, share, double(vckptSize)*1024,
									   startdIP, ckptServerIP, true, true);
				if (rval < 2) return rval;
				didPreemptAlloc = true;
			}
		}
	}

		// Now check to see if we used the reserved "goodput" capacity, and
		// if so, if this requests meets the remaining conditions for 
		// using that extra capacity.
	if (goodputEligible) {
		double usage = 0.0;
		if (netcapalloc.GetGoodputInfo(&usage) == 1 &&
			goodput_job_limit > 0.0 && usage > goodput_job_limit) {
			return 0;
		}
	}

	// setup our allocation records
	lastNetAllocKey.vm = 1;
	offer.EvaluateAttrInt(ATTR_VIRTUAL_MACHINE_ID,lastNetAllocKey.vm);
	lastNetAllocKey.startdIP = inet_addr(startdIP);
	lastICkptAlloc.bytes = executableSize*1024;
	lastICkptAlloc.endpoint = inet_addr(scheddIP);
	lastRestartAlloc.bytes = ckptSize*1024;
	lastRestartAlloc.endpoint = (lastCkptServerIP[0]) ?
		inet_addr(lastCkptServerIP) : 0;
	if (didPreemptAlloc) {
		lastVCkptAlloc.bytes = vckptSize*1024;
		lastVCkptAlloc.endpoint = inet_addr(ckptServerIP);
	} else {
		lastVCkptAlloc.bytes = 0;
	}

	return -1;
}

int
NetworkManager::RequestCapacity(const char user[], double share, double bytes,
								const char source[], const char destination[],
								bool canDeny, bool willReturn,
								bool goodputEligible)
{
	if (!active) return 2;		// allow anything if not configured

	int rval;
	rval = netcapalloc.RequestCapacity(bytes, source, destination,
									   canDeny, willReturn, goodputEligible);
	if (rval <= 0) return rval;
	rval = netusagealloc.FairShareReceived(user, share,
										   source, destination);
	if (rval == 1 || rval == -1) return rval;
	return 2;
}

int
NetworkManager::CommitPlacement(const char user[])
{
	if (!active) return 0;		// nothing to do if not active

	// match was successful; commit our network capacity allocation
	// (this will generate D_BANDWIDTH debug messages)
	netcapalloc.Commit(user);

	// commit our network allocation records (with usage)
	time_t current_time = time(0);
	struct in_addr sin;
	memcpy(&sin, &lastNetAllocKey.startdIP, sizeof(struct in_addr));
	if (lastICkptAlloc.bytes > 0) {
		lastICkptAlloc.timestamp = current_time;
		lastICkptAlloc.expiration_time = current_time+horizon;
		lastNetAllocKey.type = NET_ALLOC_ICKPT;
		NetAlloc *rec = new NetAlloc(lastICkptAlloc);
		dprintf(D_FULLDEBUG, "netman: inserting with key vm%d@%s %d)\n",
				lastNetAllocKey.vm, inet_ntoa(sin), lastNetAllocKey.type);
		if (NetAllocations.insert(lastNetAllocKey, rec) < 0) {
			dprintf(D_ALWAYS, "netman: hashtable insert failed!\n");
		}
	}
	if (lastRestartAlloc.bytes > 0) {
		lastRestartAlloc.timestamp = current_time;
		lastRestartAlloc.expiration_time = current_time+horizon;
		lastNetAllocKey.type = NET_ALLOC_RESTART;
		NetAlloc *rec = new NetAlloc(lastRestartAlloc);
		dprintf(D_FULLDEBUG, "netman: inserting with key vm%d@%s %d)\n",
				lastNetAllocKey.vm, inet_ntoa(sin), lastNetAllocKey.type);
		if (NetAllocations.insert(lastNetAllocKey, rec) < 0) {
			dprintf(D_ALWAYS, "netman: hashtable insert failed!\n");
		}
	}
	if (lastVCkptAlloc.bytes > 0) {
		lastVCkptAlloc.timestamp = current_time;
		lastVCkptAlloc.expiration_time = current_time+horizon;
		lastNetAllocKey.type = NET_ALLOC_VCKPT;
		NetAlloc *rec = new NetAlloc(lastVCkptAlloc);
		dprintf(D_FULLDEBUG, "netman: inserting with key vm%d@%s %d)\n",
				lastNetAllocKey.vm, inet_ntoa(sin), lastNetAllocKey.type);
		if (NetAllocations.insert(lastNetAllocKey, rec) < 0) {
			dprintf(D_ALWAYS, "netman: hashtable insert failed!\n");
		}
	}
	return 0;
}

/* 
   As the name implies, this method cleans out old network allocation
   records.  The records should be removed when the transfer
   completes, but the UDP message telling us that the transfer
   completed may be lost or the transfer agent may have crashed.  We
   require all transfer agents to refresh their allocation at least
   every horizon seconds, so records for long duration transfers
   should not be timed out.  In any case, if we timeout a record for
   an active reservation, we will simply instantiate a new one when we
   get our next update.
*/
int
NetworkManager::CleanupOldNetAllocations()
{
	if (!active) return 0;		// nothing to do if not active

	time_t current_time = time(0);

	NetAlloc *rec;
	NetAllocKey key;
	NetAllocations.startIterations();
	while (NetAllocations.iterate(key, rec)) {
		if (rec->expiration_time <= current_time) {
			struct in_addr sin;
			char src[16], dst[16];
			if (key.type == NET_ALLOC_ICKPT ||
				key.type == NET_ALLOC_RESTART) {
				memcpy(&sin, &rec->endpoint,
					   sizeof(struct in_addr));
				strcpy(src, inet_ntoa(sin));
				memcpy(&sin, &key.startdIP, sizeof(struct in_addr));
				strcpy(dst, inet_ntoa(sin));
			} else {
				memcpy(&sin, &key.startdIP, sizeof(struct in_addr));
				strcpy(src, inet_ntoa(sin));
				memcpy(&sin, &rec->endpoint,
					   sizeof(struct in_addr));
				strcpy(dst, inet_ntoa(sin));
			}
			dprintf(D_MATCH, "      Network allocation timed out: %d bytes "
					"%s --> %s\n", rec->bytes, src, dst);
			netcapalloc.ReturnCapacity(rec->bytes, src, dst);
			if (NetAllocations.remove(key) < 0) {
				dprintf(D_ALWAYS, "netalloc: hashtable remove failed!\n");
			}
			delete rec;
				// remove associated HashKey
			NetAllocationKeys.startIterations();
			HashKey k;
			NetAllocKey tmp;
			while (NetAllocationKeys.iterate(k, tmp)) {
				if (key == tmp) {
					NetAllocationKeys.remove(k);
				}
			}
		}
	}

	return 0;
}
