/***************************Copyright-DO-NOT-REMOVE-THIS-LINE**
 * CONDOR Copyright Notice
 *
 * See LICENSE.TXT for additional notices and disclaimers.
 *
 * Copyright (c)1990-1998 CONDOR Team, Computer Sciences Department, 
 * University of Wisconsin-Madison, Madison, WI.  All Rights Reserved.  
 * No use of the CONDOR Software Program Source Code is authorized 
 * without the express consent of the CONDOR Team.  For more information 
 * contact: CONDOR Team, Attention: Professor Miron Livny, 
 * 7367 Computer Sciences, 1210 W. Dayton St., Madison, WI 53706-1685, 
 * (608) 262-0856 or miron@cs.wisc.edu.
 *
 * U.S. Government Rights Restrictions: Use, duplication, or disclosure 
 * by the U.S. Government is subject to restrictions as set forth in 
 * subparagraph (c)(1)(ii) of The Rights in Technical Data and Computer 
 * Software clause at DFARS 252.227-7013 or subparagraphs (c)(1) and 
 * (2) of Commercial Computer Software-Restricted Rights at 48 CFR 
 * 52.227-19, as applicable, CONDOR Team, Attention: Professor Miron 
 * Livny, 7367 Computer Sciences, 1210 W. Dayton St., Madison, 
 * WI 53706-1685, (608) 262-0856 or miron@cs.wisc.edu.
****************************Copyright-DO-NOT-REMOVE-THIS-LINE**/

#include "condor_common.h"
#include "condor_debug.h"
#include "netres.h"

NetworkReservations::~NetworkReservations()
{
	NetworkRecord *current;
	network.startIterations();
	while (network.iterate(current)) {
		NetworkRecord::Reservation *r = NULL;
		current->reslist.Rewind();
		while ((r = current->reslist.Next()) != NULL) {
			current->reslist.DeleteCurrent();
			delete r;
		}
		delete current;
	}
	network.clear();
}

int
NetworkReservations::ReserveCapacity(double bytes, const char source[],
									 const char destination[],
									 bool search_forward,
									 time_t *start_time, time_t *end_time,
									 int id, const char user[], double *u_rate)
{
	time_t current_time = time(0);
	if (!start_time || !end_time) return -1;
	if (*start_time < current_time) {
		*start_time = current_time;
	}
	if (*end_time <= *start_time) {
		if (!search_forward) {
			dprintf(D_FULLDEBUG, "NetworkReservations::ReserveCapacity "
					"called with unspecified end_time and !search_forward!\n");
			return -1;
		}
		*end_time = 0;
	}

	unsigned int src = ntohl(inet_addr(source));
	unsigned int dest = ntohl(inet_addr(destination));
	if (src == dest) {
		if (*end_time == 0) *end_time = *start_time;
		if (u_rate) *u_rate = 0.0;
		return 1;	// loopback
	}

	// compute the duration of the transfer, based on the current capacity
	// of each link in the path
	double rate = 0.0;
	int duration = TransferDuration(bytes, src, dest, &rate);
	if (duration <= 0) return -1;
	if (u_rate) *u_rate = rate;

	// Now we need to search for the best feasible allocation, if one exists.
	// *start_time and *end_time will change while we search, so we need
	// to keep a copy of the original limits for the feasibility test.
	time_t start_time_limit = *start_time;
	time_t end_time_limit = *end_time;
	// We recompute our rate based on our 1 second granularity of
	// scheduling to decrease round-off error.  This is particularly
	// important for small transfers, where it is not necessary to
	// reserve the entire capacity of the network for the whole
	// second.
	rate = bytes/131072.0/(double)duration;
	// Most of the search code is the same for both directions, so we
	// use conditionals for the few statements where the search
	// direction matters.  This is the first example.
	if (search_forward) {
		*end_time = *start_time + duration;
	} else {
		*start_time = *end_time - duration;
	}
	// The following do loop finds a new feasible allocation if one
	// exists.  We check feasibility on each network hop.  If the
	// current *start_time and *end_time are not feasible for a given
	// hop, then we search forward until a we find a feasible pair for
	// that hop.  So, we know we've found a feasible allocation on
	// each hop if we get through a loop iteration without a change in
	// *start_time and *end_time.
	time_t prev_start_time, prev_end_time;
	bool first_pass = true;
	do {
		if (*start_time < start_time_limit ||
			(end_time_limit && *end_time > end_time_limit)) {
			goto failure;
		}
		prev_start_time = *start_time;
		prev_end_time = *end_time;
		// This for loop checks the feasibility of the allocation
		// on each network hop.
		int rval = 1;
		for (unsigned int current_addr = src;
			 rval > 0;
			 rval = router->NextHop(current_addr, src, dest)) {
			double capacity = 0.0;
			if (capacityinfo->GetCapacity(current_addr, capacity) < 0)
				continue;	// capacity not managed for this hop
			ASSERT(capacity > 0.0);	// TransferDuration() checked it
			NetworkRecord *current = GetNetworkRecord(current_addr);
			if (!current) return -1;
			if (first_pass) { // first pass only
				current->RemoveOldReservations(current_time);
				if (search_forward) {
					current->resiter.ToBeforeFirst();
				} else {
					current->resiter.ToAfterLast();
				}
				current->bw_in_use = 0.0;
			}
			// The following loop searches through the list of
			// reservations looking for an opening for the current
			// request.  We break out when we've searched through all
			// reservations, we reach a time limit, or we find an open
			// slot.  At that point, if (feasible == true), then
			// *start_time and *end_time point to a feasible
			// allocation.
			NetworkRecord::Reservation *r = NULL;
			bool feasible = true; // feasible until some conflict found
			while (true) {
				if (search_forward) {
					if ((r = current->resiter.Next()) == NULL) break;
					current->bw_in_use += r->bw;
				} else {
					if ((r = current->resiter.Prev()) == NULL) break;
					current->bw_in_use -= r->bw;
				}
				double bw_avail = capacity - current->bw_in_use;
				if (!feasible) {
					if ((search_forward && end_time_limit &&
						 r->ts+duration > end_time_limit) ||
						(!search_forward &&
						 r->ts-duration < start_time_limit)) {
						break;	// give up
					}
					if (bw_avail >= rate) {
						if (search_forward && r->ts > *start_time) {
							*start_time = r->ts;
							*end_time = *start_time + duration;
						} else if (!search_forward && r->ts < *end_time) {
							*end_time = r->ts;
							*start_time = *end_time - duration;
						}
						if ((!end_time_limit || *end_time <= end_time_limit) &&
							(*start_time >= start_time_limit)) {
							feasible = true;
						}
					}
				} else /* if (feasible) */ {
					if ((search_forward && r->ts >= *end_time) ||
						(!search_forward && r->ts <= *start_time)) {
						break; // found one!!!
					}
					if (bw_avail < rate) {
						feasible = false;
					}
				}
			}
			if (!feasible) { // couldn't find anything
				goto failure;
			}
			// Start from where we left off on the next pass
			// by pushing the iterator back one place.
			if (search_forward) {
				current->resiter.Prev();
				if (r) current->bw_in_use -= r->bw;
			} else {
				current->resiter.Next();
				if (r) current->bw_in_use += r->bw;
			}
		}
		if (rval < 0) return -1; // routing error
		first_pass = false;
	} while (prev_start_time != *start_time && prev_end_time != *end_time);
	
	AddReservation(src, dest, *start_time, *end_time, rate, id, user);
	
	return 1;

failure:
	dprintf(D_FULLDEBUG, "netman: request for %.0f bytes (%s --> %s) can't "
			"be scheduled before deadline\n", bytes, source, destination);
	return 0;
}

int
NetworkReservations::ReserveBandwidth(double bytes,
									  double min_rate, double *max_rate,
									  const char source[],
									  const char destination[],
									  bool search_forward,
									  reservation_preference pref,
									  time_t *start_time, time_t *end_time,
									  int id, const char user[])
{
	unsigned int src = ntohl(inet_addr(source));
	unsigned int dest = ntohl(inet_addr(destination));
	if (src == dest) {
		if (*end_time == 0) *end_time = *start_time;
		return 1;	// loopback
	}

	time_t current_time = time(0);
	if (!start_time || !end_time) return -1;
	if (*start_time < current_time) {
		*start_time = current_time;
	}
	if (*end_time <= *start_time) {
		if (!search_forward) {
			dprintf(D_FULLDEBUG, "NetworkReservations::ReserveBandwidth "
					"called with unspecified end_time and !search_forward!\n");
			return -1;
		}
		*end_time = 0;
	}
	if (min_rate < 0.0) min_rate = 0.0;
	if (*max_rate == 0.0) {
		*max_rate = Bandwidth(src, dest);
	}
	if (*max_rate < min_rate) {
		dprintf(D_FULLDEBUG, "NetworkReservations::ReserveBandwidth: "
				"*max_rate < min_rate (%.0f < %.0f).\n",
				*max_rate, min_rate);
		return -1;
	}
	if (search_forward && pref == pref_latest_start) {
		dprintf(D_FULLDEBUG, "NetworkReservations::ReserveBandwidth: "
				"search_forward and pref_latest_start are incompatible.\n");
		return -1;
	}
	if (!search_forward && pref == pref_earliest_completion) {
		dprintf(D_FULLDEBUG, "NetworkReservations::ReserveBandwidth: "
				"!search_forward and pref_earliest_completion are "
				"incompatible.\n");
		return -1;
	}

	// Now we need to search for the best feasible allocation, if one exists.
	// *start_time and *end_time will change while we search, so we need
	// to keep a copy of the original limits for the feasibility test.
	time_t start_time_limit = *start_time;
	time_t end_time_limit = *end_time;
	// Allocate max_rate if no conflicts arise.
	double rate = *max_rate;
	int duration = (int)(bytes/131072.0/(rate))+1;
	// Most of the search code is the same for both directions, so we
	// use conditionals for the few statements where the search
	// direction matters.  This is the first example.
	if (search_forward) {
		*end_time = *start_time + duration;
	} else {
		*start_time = *end_time - duration;
	}
	// These variables hold the best feasible allocation found (so
	// far).  If they are still 0 when we exit the loop, then we
	// couldn't find any feasible allocations.
	time_t best_start_time = 0, best_end_time = 0;
	int best_duration = 0;
	// We keep looping until we either know we've found the best
	// allocation (found_best == true), we've checked every feasible
	// allocation (give_up == true), or we've searched past our time
	// limits.  The check on the time limits is probably redundant,
	// since give_up becomes true when we hit the time limits, but we
	// keep it just to be safe.
	bool found_best = false, give_up = false, full_pass = true;
	time_t backtrack_time = 0;	// where should we start backtracking from?
	while (!found_best && !give_up && (*end_time > start_time_limit) &&
		   (!end_time_limit || *start_time < end_time_limit)) {
		// The following do loop finds a new feasible allocation if
		// (another) one exists.  We check feasibility on each network
		// hop.  If the current *start_time and *end_time are not
		// feasible for a given hop, then we search forward until a we
		// find a feasible pair for that hop.  So, we know we've found
		// a feasible allocation on each hop if we get through a loop
		// iteration without a change in *start_time and *end_time.
		time_t prev_start_time, prev_end_time;
		do {
			prev_start_time = *start_time;
			prev_end_time = *end_time;
			found_best = false;	// initialize for each allocation
			// This for loop checks the feasibility of the allocation
			// on each network hop.
			int rval = 1;
			for (unsigned int current_addr = src;
				 !give_up && rval > 0;
				 rval = router->NextHop(current_addr, src, dest)) {
				double capacity = 0.0;
				if (capacityinfo->GetCapacity(current_addr, capacity) < 0)
					continue;	// capacity not managed for this hop
				if (capacity < min_rate) {
					dprintf(D_FULLDEBUG,
							"netman: request for %.0f bytes (%s --> %s) "
							"refused because subnet %s capacity %.0f is "
							"less than min rate %.0f\n",
							bytes, source, destination,
							inet_itoa(current_addr), capacity, min_rate);
					return 0;	// no chance of success
				}
				int min_duration; // min possible duration for this network
				if (capacity < *max_rate) {
					min_duration = (int)(bytes/131072.0/(capacity))+1;
				} else {
					min_duration = (int)(bytes/131072.0/(*max_rate))+1;
				}
				NetworkRecord *current = GetNetworkRecord(current_addr);
				if (!current) return -1;
				if (full_pass) { // first pass or backtracking
					current->RemoveOldReservations(current_time);
					if (search_forward) {
						current->resiter.ToBeforeFirst();
					} else {
						current->resiter.ToAfterLast();
					}
					current->bw_in_use = 0.0;
				}
				// The following loop searches through the list of
				// reservations looking for an opening for the current
				// request.  We break out when we've searched through
				// all reservations, we reach a time limit, or we find
				// an open slot.  At that point, if (feasible ==
				// true), then *start_time and *end_time point to a
				// feasible allocation.  When we find a candidate
				// *start_time, we search ahead until *end_time.  If
				// it turns out that bandwidth isn't available all the
				// way to *end_time, we need to backtrack and search
				// from the candidate *start_time for a new candidate,
				// since we might have passed by other possibile
				// candidates when testing this one.
				NetworkRecord::Reservation *r = NULL;
				bool feasible = false;
				if ((!best_duration || duration < best_duration) &&
					(!end_time_limit || *end_time <= end_time_limit) &&
					(*start_time >= start_time_limit)) {
					feasible = true;
				}				
				while (true) {
					if (search_forward) {
						if ((r = current->resiter.Next()) == NULL) break;
						current->bw_in_use += r->bw;
					} else {
						if ((r = current->resiter.Prev()) == NULL) break;
						current->bw_in_use -= r->bw;
					}
					double bw_avail = capacity - current->bw_in_use;
					if (!feasible) {
						if ((search_forward && end_time_limit &&
							 r->ts+min_duration > end_time_limit) ||
							(!search_forward &&
							 r->ts-min_duration < start_time_limit)) {
							break;	// give up
						}
						if (bw_avail >= min_rate) {
							rate = (bw_avail < *max_rate) ?
								bw_avail : *max_rate;
							duration = (int)(bytes/131072.0/(rate))+1;
							if (search_forward) {
								if (r->ts > *start_time) *start_time = r->ts;
								*end_time = *start_time + duration;
							} else {
								if (r->ts < *end_time) *end_time = r->ts;
								*start_time = *end_time - duration;
							}
							if ((!best_duration || duration < best_duration) &&
								(!end_time_limit ||
								 *end_time <= end_time_limit) &&
								(*start_time >= start_time_limit)) {
								feasible = true;
							}
						}
					} else /* if (feasible) */ {
						if ((search_forward && r->ts >= *end_time) ||
							(!search_forward && r->ts <= *start_time)) {
							break; // found one!!!
						}
						if (bw_avail < min_rate) {
							feasible = false;
						} else if (bw_avail < rate ||
								   (search_forward && r->ts < *start_time) ||
								   (!search_forward && r->ts > *end_time)) {
							// update our rate if
							// - the available bandwidth in our current
							//   feasible allocation has decreased or
							// - we haven't reached the start of our
							//   current feasible allocation, so we need to
							//   keep the rate up-to-date for when we reach
							//   the start
							rate = (bw_avail < *max_rate) ?
								bw_avail : *max_rate;
							duration = (int)(bytes/131072.0/(rate))+1;
							if (search_forward) {
								*end_time = *start_time + duration;
							} else {
								*start_time = *end_time - duration;
							}
							if ((best_duration && duration > best_duration) ||
								(end_time_limit &&
								 *end_time > end_time_limit)) {
								feasible = false;
							}
						} else if (bw_avail > rate) {
							// We've found a pocket of greater bandwidth,
							// so we'll need to backtrack to see if we can
							// get a better allocation inside this pocket.
							if (!backtrack_time ||
								(search_forward && backtrack_time > r->ts) ||
								(!search_forward && backtrack_time < r->ts)) {
								backtrack_time = r->ts;
							}
						}
					}
				}
				if (!feasible) {
					give_up = true; // couldn't find anything else
					break;
				}
				if (duration == min_duration || pref == pref_first) {
					found_best = true; // won't find anything better than this
				}
				// Start from where we left off on the next pass
				// by pushing the iterator back one place.
				if (search_forward) {
					current->resiter.Prev();
					if (r) current->bw_in_use -= r->bw;
				} else {
					current->resiter.Next();
					if (r) current->bw_in_use += r->bw;
				}
			}
			if (rval < 0) return -1; // routing error
			full_pass = false;
		} while (!give_up && prev_start_time != *start_time &&
				 prev_end_time != *end_time);
		if (!give_up) {		// found one
			if (!best_start_time || found_best || 		// if first or best
				(pref == pref_earliest_completion &&	// or better than
				 (*end_time < best_end_time)) || 		// previous best
				(pref == pref_latest_start && 			// then we've got
				 (*start_time > best_start_time))) {	// a new best
				min_rate = rate;
				best_start_time = *start_time;
				best_end_time = *end_time;
				best_duration = duration;
				if (pref == pref_earliest_completion) {	  // set limits to only
					end_time_limit = best_end_time-1;     // look for better
				} else if (pref == pref_latest_start) {	  // reservations
					start_time_limit = best_start_time+1;
				}
			}
		}
		if (!found_best && backtrack_time) {
			if (search_forward) {
				*start_time = backtrack_time;
				*end_time = *start_time + best_duration;
			} else {
				*end_time = backtrack_time;
				*start_time = *end_time - best_duration;
			}
			backtrack_time = 0;
			full_pass = true;	// reset iterators for backtracking
		}
	}
	if (best_start_time) {		// if found something
		*start_time = best_start_time;	
		*end_time = best_end_time;
		*max_rate = min_rate;
	} else {
		dprintf(D_FULLDEBUG, "netman: request for %.0f bytes "
				"(%s --> %s) can't be scheduled before "
				"deadline\n", bytes, source, destination);
		return 0;
	}

	AddReservation(src, dest, *start_time, *end_time, *max_rate, id, user);

	return 1;
}

int
NetworkReservations::ReserveBandwidth(double min_rate, double *max_rate,
									  const char source[],
									  const char destination[],
									  time_t start_time, time_t end_time,
									  int id, const char user[])
{
	unsigned int src = ntohl(inet_addr(source));
	unsigned int dest = ntohl(inet_addr(destination));
	if (src == dest) {
		return 1;	// loopback
	}

	time_t current_time = time(0);
	if (start_time < current_time) {
		start_time = current_time;
	}
	if (end_time <= start_time) {
		dprintf(D_FULLDEBUG, "NetworkReservations::ReserveBandwidth "
				"called with end_time <= start_time (%d <= %d)!\n",
				(int)end_time, (int)start_time);
		return -1;
	}
	if (min_rate < 0.0) min_rate = 0.0;
	if (*max_rate < min_rate) {
		dprintf(D_FULLDEBUG, "NetworkReservations::ReserveBandwidth: "
				"*max_rate < min_rate (%.0f < %.0f).\n",
				*max_rate, min_rate);
		return -1;
	}

	int rval = 1;
	for (unsigned int current_addr = src;
		 rval > 0;
		 rval = router->NextHop(current_addr, src, dest)) {
		double capacity = 0.0;
		if (capacityinfo->GetCapacity(current_addr, capacity) < 0)
			continue;	// capacity not managed for this hop
		if (capacity < min_rate) {
			dprintf(D_FULLDEBUG,
					"netman: bandwidth request (%s --> %s) "
					"refused because subnet %s capacity %.0f is "
					"less than min rate %.0f\n", source, destination,
					inet_itoa(current_addr), capacity, min_rate);
			return 0;	// no chance of success
		}
		NetworkRecord *current = GetNetworkRecord(current_addr);
		if (!current) return -1;
		current->RemoveOldReservations(current_time);
		current->resiter.ToBeforeFirst();
		NetworkRecord::Reservation *r = NULL;
		current->bw_in_use = 0.0;	// running total during search
		while ((r = current->resiter.Next()) != NULL) {
			current->bw_in_use += r->bw;
			if (r->ts < start_time) continue;
			if (r->ts >= end_time) break;
			double bw_avail = capacity - current->bw_in_use;
			if (bw_avail < min_rate) {
				dprintf(D_FULLDEBUG,
						"netman: insufficient bandwidth on subnet %s for "
						"request of %.0f Mbps (%s --> %s)\n",
						inet_itoa(current_addr), min_rate,
						source, destination);
				return 0;
			}
			if (bw_avail < *max_rate) {
				*max_rate = bw_avail;
			}
		}
	}
	if (rval < 0) return -1;	// routing error

	AddReservation(src, dest, start_time, end_time, *max_rate, id, user);
	
	return 1;
}

int
NetworkReservations::CancelReservation(int id)
{
	NetworkRecord *current = NULL;
	network.startIterations();
	while (network.iterate(current)) {
		current->reslist.Rewind();
		NetworkRecord::Reservation *r = NULL;
		while ((r = current->reslist.Next()) != NULL) {
			if (r->id == id) {
				current->reslist.DeleteCurrent();
				delete r;
			}
		}
	}
	return 0;
}

int
NetworkReservations::TransferDuration(double bytes, const char source[],
									  const char destination[], double *rate)
{
	unsigned int src = ntohl(inet_addr(source));
	unsigned int dest = ntohl(inet_addr(destination));
	return TransferDuration(bytes, src, dest, rate);
}

int
NetworkReservations::TransferDuration(double bytes, unsigned int src,
									  unsigned int dest, double *u_rate)
{
	if (src == dest) {
		if (u_rate) *u_rate = bytes/131072.0; // loopback
		return 1;
	}

	double rate = Bandwidth(src, dest);

	if (rate == 0.0) {
		return -1;
	}

	if (u_rate) *u_rate = rate;
	return (int)(bytes/131072.0/(rate))+1;
}

double
NetworkReservations::Bandwidth(const char source[], const char destination[])
{
	unsigned int src = ntohl(inet_addr(source));
	unsigned int dest = ntohl(inet_addr(destination));
	return Bandwidth(src, dest);
}

double
NetworkReservations::Bandwidth(unsigned int src, unsigned int dest)
{
	double rate = 0.0;
	int rval = 1;
	for (unsigned int current_addr = src;
		 rval > 0;
		 rval = router->NextHop(current_addr, src, dest)) {
		double capacity = 0.0;
		if (capacityinfo->GetCapacity(current_addr, capacity) < 0) continue;
		if (capacity < rate || rate == 0.0) {
			rate = capacity;
		}
	}
	if (rval < 0) return 0.0;
	return rate;
}

void
NetworkReservations::AddReservation(unsigned int src, unsigned int dest,
									time_t start_time, time_t end_time,
									double rate, int id, const char user[])
{
	int rval = 1;
	for (unsigned int current_addr = src;
		 rval > 0;
		 rval = router->NextHop(current_addr, src, dest)) {
		NetworkRecord *current = NULL;
		if (network.lookup(current_addr, current) < 0) continue;
		current->reslist.Rewind();
		NetworkRecord::Reservation *r = NULL;
		NetworkRecord::Reservation *start =
			new NetworkRecord::Reservation(start_time, rate, id);
		NetworkRecord::Reservation *end =
			new NetworkRecord::Reservation(end_time, -(rate), id);
		start->other = end;
		end->other = start;
			// We always insert start transfer records after all other
			// records with the same timestamp so we are guaranteed
			// that for each timestamp usage first decreases then
			// increases.  Reservations are always for >0 seconds, so
			// we allocate capacity between each second (i.e., between
			// each timestamp).
		while (((r = current->reslist.Next()) != NULL) &&
			   (r->ts <= start_time));
		if (r) current->reslist.Insert(start);
		else current->reslist.Append(start);
			// Likewise, we insert end transfer records before all
			// other records with the same timestamp. Notice also that
			// these records have negative bandwidth.
		while ((r) && (r->ts < end_time)) {
			r = current->reslist.Next();
		}
		if (r) current->reslist.Insert(end);
		else current->reslist.Append(end);
		dprintf(D_BANDWIDTH, "netres: reserved %.0f Mbps on %s from %d "
				"to %d\n", rate, inet_itoa(current_addr), start_time,
				end_time);
		if (netusage) {
			double capacity = 0.0;
			if (capacityinfo->GetCapacity(current_addr, capacity) < 0) {
				continue;
			}
			netusage->AddUsage(user, current_addr,
							   double(end_time-start_time)*(rate/capacity));
		}
	}
}

void
NetworkReservations::NetworkRecord::RemoveOldReservations(time_t current_time)
{
	if (current_time <= expiration_date) return;
	expiration_date = current_time;
	Reservation *r = NULL;
	reslist.Rewind();
	while (((r = reslist.Next()) != NULL) &&
		   (r->ts < current_time)) {
		if (r->bw < 0) {
			reslist.DeleteCurrent();
			Reservation *start = r->other;
			delete r;
			reslist.Rewind();
			while (((r = reslist.Next()) != NULL) &&
				   r != start);
			if (r) {
				reslist.DeleteCurrent();
				delete r;
			}
		}
	}
}

NetworkReservations::NetworkRecord *
NetworkReservations::GetNetworkRecord(unsigned int addr)
{
	NetworkRecord *rec = NULL;
	if (network.lookup(addr, rec) < 0) {
		rec = new NetworkRecord(addr);
		ASSERT(network.insert(addr, rec) == 0);
	}
	return rec;
}
