/***************************Copyright-DO-NOT-REMOVE-THIS-LINE**
 * CONDOR Copyright Notice
 *
 * See LICENSE.TXT for additional notices and disclaimers.
 *
 * Copyright (c)1990-1998 CONDOR Team, Computer Sciences Department, 
 * University of Wisconsin-Madison, Madison, WI.  All Rights Reserved.  
 * No use of the CONDOR Software Program Source Code is authorized 
 * without the express consent of the CONDOR Team.  For more information 
 * contact: CONDOR Team, Attention: Professor Miron Livny, 
 * 7367 Computer Sciences, 1210 W. Dayton St., Madison, WI 53706-1685, 
 * (608) 262-0856 or miron@cs.wisc.edu.
 *
 * U.S. Government Rights Restrictions: Use, duplication, or disclosure 
 * by the U.S. Government is subject to restrictions as set forth in 
 * subparagraph (c)(1)(ii) of The Rights in Technical Data and Computer 
 * Software clause at DFARS 252.227-7013 or subparagraphs (c)(1) and 
 * (2) of Commercial Computer Software-Restricted Rights at 48 CFR 
 * 52.227-19, as applicable, CONDOR Team, Attention: Professor Miron 
 * Livny, 7367 Computer Sciences, 1210 W. Dayton St., Madison, 
 * WI 53706-1685, (608) 262-0856 or miron@cs.wisc.edu.
****************************Copyright-DO-NOT-REMOVE-THIS-LINE**/
#ifndef __NETWORK_RESERVATIONS_H__
#define __NETWORK_RESERVATIONS_H__

#include "router.h"
#include "netcapacity.h"
#include "netusage.h"
#include "list.h"

class NetworkReservations {
public:
	NetworkReservations() : next_id(1), network(255, addrhash),
		router(NULL), capacityinfo(NULL), netusage(NULL) {}
	~NetworkReservations();

	void BindRouter(Router *r) { router = r; }
	void BindCapacity(NetworkCapacity *c) { capacityinfo = c; } 
	void BindUsage(NetworkUsage *n) { netusage = n; }

	// reserve network capacity to transfer the given number of bytes
	// from the source to the destination network endpoints (IP
	// addresses in dotted notation) assuming the network is the
	// bottleneck (and so the network sets the transfer rate) at the
	// earliest time (if earliest is true) or latest time (if
	// earliest is false) after *start_time but before
	// *end_time; if *start_time is less than the current time, it
	// will be ignored and the current time will be used instead; if
	// (*end_time <= *start_time), *end_time will be ignored and no
	// end time boundary will be enforced; an end time boundary is
	// required when search_forward is false; it is good practice to
	// set *start_time and/or *end_time to 0 if you want that boundary
	// to be ignored; id may be specified to allow reservations to be
	// cancelled at a later time (multiple reservations may have the
	// same id); the actual begin and end times will be returned in
	// *start_time and *end_time; the available network bandwidth (in
	// Mb/s) will be returned in rate (if non-null); returns 1 if
	// request is granted, 0 if denied, -1 if error 
	// Note: Requesters with the ability to police their bandwidth
	// usage should prefer the ReserveBandwidth() methods, because
	// they allow greater scheduling flexibility.
	int ReserveCapacity(double bytes,
						const char source[], const char destination[],
						bool search_forward,
						time_t *start_time, time_t *end_time,
						int id=0, const char user[]="", double *rate=NULL);

	// in contrast to ReserveCapacity(), ReserveBandwidth() allows
	// minimum and maximum transfer rates to be specified in min_rate
	// and *max_rate, so the network is not necessarily assumed to be
	// the bottleneck; requesters therefore must limit themselves to
	// the allocated bandwidth, returned in *max_rate; if bytes is
	// non-zero, *start_time, *end_time, and id work as in
	// ReserveCapacity() above; returns 1 if request is granted, 0 if
	// denied, -1 if error
	enum reservation_preference { pref_first, pref_shortest,
								  pref_earliest_completion,
								  pref_latest_start };
	int ReserveBandwidth(double bytes,
						 double min_rate, double *max_rate,
						 const char source[], const char destination[],
						 bool search_forward, reservation_preference pref,
						 time_t *start_time, time_t *end_time,
						 int id=0, const char user[]="");

	// this version of ReserveBandwidth() requests a bandwidth
	// reservation between min_rate and *max_rate from start_time to
	// end_time; if start_time is less than the current time, it will
	// be ignored and the current time will be used instead; end_time
	// must be greater than start_time (and the current time) or the
	// request will return an error; the allocated bandwidth is
	// returned in *max_rate; id works as in ReserveCapacity() above;
	// returns 1 if request is granted, 0 if denied, -1 if error
	int ReserveBandwidth(double min_rate, double *max_rate,
						 const char source[], const char destination[],
						 time_t start_time, time_t end_time,
						 int id=0, const char user[]="");
	
	// cancel reservation(s) made via ReserveCapacity; returns 0 on
	// success, -1 on error
	int CancelReservation(int id);

	// get a new reservation id for use in ReserveCapacity
	int NewID() { return next_id++; }

	// return the duration (in seconds) of a transfer of the specified
	// number of bytes from source to destination, based on current
	// network capacity, assuming zero network congestion; the network
	// bandwidth (in Mb/s) used in the computation will be returned in
	// rate (if non-null); returns -1 on error
	int TransferDuration(double bytes, const char source[],
						 const char destination[], double *rate=NULL);

	// returns the max. bandwidth (in Mbps) available for a transfer
	// between the source and destionation, assuming zero network
	// congestion; returns 0.0 on error
	double Bandwidth(const char source[], const char destination[]);

private:
	// helper function which implements TransferDuration(double, const
	// char [], const char [], double *, int *) from above.
	int TransferDuration(double bytes, unsigned int src,
						 unsigned int dest, double *rate=NULL);

	// helper function which implements Bandwidth(const char source[],
	// const char destionation[]) from above.
	double Bandwidth(unsigned int src, unsigned int dest);

	void AddReservation(unsigned int src, unsigned int dest,
						time_t start_time, time_t end_time,
						double rate, int id, const char user[]="");

	struct NetworkRecord {
		NetworkRecord() : addr(0), resiter(reslist), bw_in_use(0.0),
			expiration_date(0) {}
		NetworkRecord(unsigned int a) : addr(a), resiter(reslist),
			bw_in_use(0.0), expiration_date(0) {}
		unsigned int addr;
		struct Reservation {
			Reservation(time_t timestamp, double bandwidth, int res_id) :
				ts(timestamp), bw(bandwidth), id(res_id), other(NULL) {}
			time_t ts;
			double bw;			// >0 for start of res; <0 for end of res
			int id;
			Reservation *other;	// corresponding start/end reservation
		};
		List<Reservation> reslist; // reservations of future capacity
		ListIterator<Reservation> resiter;
		double bw_in_use;
		void RemoveOldReservations(time_t current_time);
		time_t expiration_date;	// used by RemoveOldReservations()
	};
	int next_id;				// id counter for reservations
	HashTable<unsigned int, NetworkRecord *> network;

	NetworkRecord *GetNetworkRecord(unsigned int addr);

	Router *router;
	NetworkCapacity *capacityinfo;
	NetworkUsage *netusage;
};

#endif
