/***************************Copyright-DO-NOT-REMOVE-THIS-LINE**
 * CONDOR Copyright Notice
 *
 * See LICENSE.TXT for additional notices and disclaimers.
 *
 * Copyright (c)1990-1998 CONDOR Team, Computer Sciences Department, 
 * University of Wisconsin-Madison, Madison, WI.  All Rights Reserved.  
 * No use of the CONDOR Software Program Source Code is authorized 
 * without the express consent of the CONDOR Team.  For more information 
 * contact: CONDOR Team, Attention: Professor Miron Livny, 
 * 7367 Computer Sciences, 1210 W. Dayton St., Madison, WI 53706-1685, 
 * (608) 262-0856 or miron@cs.wisc.edu.
 *
 * U.S. Government Rights Restrictions: Use, duplication, or disclosure 
 * by the U.S. Government is subject to restrictions as set forth in 
 * subparagraph (c)(1)(ii) of The Rights in Technical Data and Computer 
 * Software clause at DFARS 252.227-7013 or subparagraphs (c)(1) and 
 * (2) of Commercial Computer Software-Restricted Rights at 48 CFR 
 * 52.227-19, as applicable, CONDOR Team, Attention: Professor Miron 
 * Livny, 7367 Computer Sciences, 1210 W. Dayton St., Madison, 
 * WI 53706-1685, (608) 262-0856 or miron@cs.wisc.edu.
****************************Copyright-DO-NOT-REMOVE-THIS-LINE**/

#include "condor_common.h"
#include "condor_debug.h"
#include "netsharealloc.h"

int reqhash(const int &id, int numBuckets) {
	return (id % numBuckets); 
}

NetworkShareAllocator::~NetworkShareAllocator()
{
	BandwidthRequest *req = NULL;
	requests.startIterations();
	while (requests.iterate(req)) {
		delete req;
	}
	requests.clear();
	NetworkRecord *current = NULL;
	network.startIterations();
	while (network.iterate(current)) {
		delete current;
	}
	network.clear();
}

int
NetworkShareAllocator::SubmitRequest(const char source[],
									 const char destination[],
									 double rate, const char user[])
{
	unsigned int src = ntohl(inet_addr(source));
	unsigned int dest = ntohl(inet_addr(destination));
	BandwidthRequest *req = new BandwidthRequest(src, dest, rate, user);
	int id = NewID();
	ASSERT(requests.insert(id, req) == 0);
	int rval = 1;
	for (unsigned int current_addr = src;
		 rval > 0;
		 rval = router->NextHop(current_addr, src, dest)) {
		NetworkRecord *current = GetNetworkRecord(current_addr);
		current->Add(req);
	}
	return id;
}

int
NetworkShareAllocator::CancelRequest(int id)
{
	BandwidthRequest *req = NULL;
	if (requests.lookup(id, req) < 0 ||
		requests.remove(id) < 0) {
		return -1;
	}
	int rval = 1;
	for (unsigned int current_addr = req->src;
		 rval > 0;
		 rval = router->NextHop(current_addr, req->src, req->dest)) {
		NetworkRecord *current = GetNetworkRecord(current_addr);
		current->Cancel(req);
		if (current->reqlist.IsEmpty()) {
			ASSERT(network.remove(current->addr) == 0);
			delete current;
		}
	}
	delete req;
	return 0;
}

int
NetworkShareAllocator::UpdateAllocations()
{
		// Allocate Max-Min Fair Share to each connection.
		// Eventually, we'll want to allocate fair share per user,
		// based on priorities.
		// Algorithm adapted from Q. Ma, P. Steenkiste, and
		// H. Zhang. "Routing High-Bandwidth Traffic in Max-Min Fair
		// Share Networks". In ACM SIGCOMM96, pages 206-217, Stanford,
		// CA, August 1996.

		// Initialize all requests to 0.0 and unsatisfied.
	int tot_num_unsat = requests.getNumElements();
	BandwidthRequest *req = NULL;
	requests.startIterations();
	while (requests.iterate(req)) {
		req->allocated_rate = 0.0;
		req->satisfied = false;
	}
	NetworkRecord *current = NULL;
	network.startIterations();
	while (network.iterate(current)) {
		current->num_unsat = current->reqlist.Number();
		current->usage = 0.0;
	}

		// Incrementally increase allocations until all requests are
		// satisfied.
	do {
			// Find the minimum increment (inc) it's safe to add to
			// all unsatisfied requests.
		double inc = 0.0;
		network.startIterations();
		while (network.iterate(current)) {
			if (!current->num_unsat) continue; // No unsatisfied requests here.
			double capacity = 0.0;
			if (capacityinfo->GetCapacity(current->addr, capacity) < 0)
				continue;
			double this_inc =
				(capacity-(current->usage))/((double)current->num_unsat);
			if (inc == 0.0 || this_inc < inc) {
				inc = this_inc;
			}
		}
			// Increase allocations for all unsatisfied requests by inc.
			// Mark requests that have been allocated their requested rate
			// as satisfied.
		requests.startIterations();
		while (requests.iterate(req)) {
			if (!req->satisfied) {
				req->allocated_rate += inc;
				if (req->allocated_rate >= req->requested_rate || inc == 0.0) {
					req->allocated_rate = req->requested_rate;
					req->satisfied = true;
					tot_num_unsat--;
				}
			}
		}
			// Update usage totals in each network record and mark all
			// requests that traverse fully utilized networks as
			// satisfied.
		network.startIterations();
		while (network.iterate(current)) {
			if (!current->num_unsat) continue; // No unsatisfied requests here.
			double capacity = 0.0;
			if (capacityinfo->GetCapacity(current->addr, capacity) < 0)
				continue;
			current->usage = 0.0;
			current->reqlist.Rewind();
			while (current->reqlist.Next(req)) {
				current->usage += req->allocated_rate;
			}
			if (capacity-(current->usage) <= 0.0) {
				current->reqlist.Rewind();
				while (current->reqlist.Next(req)) {
					if (!req->satisfied) {
						req->satisfied = true;
						tot_num_unsat--;
					}
				}
			}
		}
			// Update num_unsat totals in each network record.
		network.startIterations();
		while (network.iterate(current)) {
			if (!current->num_unsat) continue; // No unsatisfied requests here.
			current->num_unsat = 0;
			current->reqlist.Rewind();
			while (current->reqlist.Next(req)) {
				if (!req->satisfied) {
					current->num_unsat++;
				}
			}
		}
	} while (tot_num_unsat);

	if (DebugFlags&D_BANDWIDTH) {
		network.startIterations();
		while (network.iterate(current)) {
			if (current->usage > 0.0) {
				dprintf(D_BANDWIDTH, "netshare: %.2f Mbps allocated on %s\n",
						current->usage, inet_itoa(current->addr));
			}
		}
	}

	return 0;
}

double
NetworkShareAllocator::GetAllocation(int id)
{
	BandwidthRequest *req = NULL;
	if (requests.lookup(id, req) < 0) {
		return 0.0;
	}
	return req->allocated_rate;
}

void
NetworkShareAllocator::NetworkRecord::Add(BandwidthRequest *req)
{
	ASSERT(reqlist.Append(req));
}

void
NetworkShareAllocator::NetworkRecord::Cancel(BandwidthRequest *req)
{
	BandwidthRequest *r;
	reqlist.Rewind();
	while ((r = reqlist.Next()) != NULL) {
		if (r == req) {
			reqlist.DeleteCurrent();
		}
	}
}

NetworkShareAllocator::NetworkRecord *
NetworkShareAllocator::GetNetworkRecord(unsigned int addr)
{
	NetworkRecord *rec = NULL;
	if (network.lookup(addr, rec) < 0) {
		rec = new NetworkRecord(addr);
		ASSERT(network.insert(addr, rec) == 0);
	}
	return rec;
}
