/***************************Copyright-DO-NOT-REMOVE-THIS-LINE**
 * CONDOR Copyright Notice
 *
 * See LICENSE.TXT for additional notices and disclaimers.
 *
 * Copyright (c)1990-1998 CONDOR Team, Computer Sciences Department, 
 * University of Wisconsin-Madison, Madison, WI.  All Rights Reserved.  
 * No use of the CONDOR Software Program Source Code is authorized 
 * without the express consent of the CONDOR Team.  For more information 
 * contact: CONDOR Team, Attention: Professor Miron Livny, 
 * 7367 Computer Sciences, 1210 W. Dayton St., Madison, WI 53706-1685, 
 * (608) 262-0856 or miron@cs.wisc.edu.
 *
 * U.S. Government Rights Restrictions: Use, duplication, or disclosure 
 * by the U.S. Government is subject to restrictions as set forth in 
 * subparagraph (c)(1)(ii) of The Rights in Technical Data and Computer 
 * Software clause at DFARS 252.227-7013 or subparagraphs (c)(1) and 
 * (2) of Commercial Computer Software-Restricted Rights at 48 CFR 
 * 52.227-19, as applicable, CONDOR Team, Attention: Professor Miron 
 * Livny, 7367 Computer Sciences, 1210 W. Dayton St., Madison, 
 * WI 53706-1685, (608) 262-0856 or miron@cs.wisc.edu.
****************************Copyright-DO-NOT-REMOVE-THIS-LINE**/

#include "condor_common.h"
#include "condor_debug.h"
#include "router.h"

static bool routing_debug = false;

int
addrhash(const unsigned int &addr, int numBuckets)
{
	int hashval = 0;
	const unsigned char *byte = (const unsigned char *)&addr;
	for (unsigned int i=0; i < sizeof(unsigned int); i++) {
		hashval += *byte;
		byte++;
	}
	return ((unsigned int)hashval % numBuckets);
}

static int
maskcompar(unsigned int *a, unsigned int *b)
{
	if (*a > *b) return -1;
	if (*a < *b) return 1;
	return 0;
}

char *
inet_itoa(unsigned int addr)
{
	struct in_addr sin;
	addr = htonl(addr);
	memcpy(&sin, &addr, sizeof(struct in_addr));
	return inet_ntoa(sin);
}

ostream& operator<<(ostream &s, Router::NetworkRecord::NextHop hop)
{
	return s << '(' << inet_itoa(hop.net) << ", " << inet_itoa(hop.addr)
			 << ", " << inet_itoa(hop.mask) << ')';
}

int
Router::LoadRoutingTable(const char filename[])
{
	FILE *fp = NULL;
	if ((fp = fopen(filename, "r")) == NULL) {
		dprintf(D_ALWAYS, "router: can't open \"%s\"\n", filename);
		return -1;
	}

		// start with a clean routing table
	Cleanup();

	NetworkRecord *current = NULL;
	char line[80];
	int lc = 0;
	while (fgets(line, 80, fp)) {
		lc++;
		char *tok;
		if ((tok = strtok(line, " \t\n")) == NULL) continue;
		if (tok[0] == '#') continue;
		if (strcmp(tok, "-->") == 0) {
			NetworkRecord::NextHop *hop = new NetworkRecord::NextHop;
			if ((tok = strtok(NULL, " \t\n")) == NULL) {
					// parse error
				delete current;
				delete hop;
				dprintf(D_ALWAYS, "router: parse error on line %d of \"%s\"\n",
						lc, filename);
				return -1;
			}
			hop->net = ntohl(inet_addr(tok));
			if ((tok = strtok(NULL, " \t\n")) == NULL) {
					// parse error
				delete current;
				delete hop;
				dprintf(D_ALWAYS, "router: parse error on line %d of \"%s\"\n",
						lc, filename);
				return -1;
			}
			hop->addr = ntohl(inet_addr(tok));
			if ((tok = strtok(NULL, " \t\n")) == NULL) {
					// parse error
				delete current;
				delete hop;
				dprintf(D_ALWAYS, "router: parse error on line %d "
						"of \"%s\"\n", lc, filename);
				return -1;
			}
			hop->mask = ntohl(inet_addr(tok));
			if (hop->addr != (hop->addr&hop->mask)) {
					// error: masked out bits must be 0
				delete current;
				delete hop;
				dprintf(D_ALWAYS, "router: parse error on line %d "
						"of \"%s\"\n", lc, filename);
				return -1;
			}
			if (hop->addr == 0 && hop->mask != 0) {
				dprintf(D_ALWAYS,"router: parse error on line %d of \"%s\":"
						"invalid IP-ADDR 0.0.0.0\n", lc, filename);
			}
			NetworkRecord::NextHop *oldhop = NULL;
			if (current->hoplist.lookup(hop->addr, oldhop) == 0) {
				delete current;
				delete hop;
				dprintf(D_ALWAYS,"router: parse error on line %d of \"%s\":"
						"duplicate hop address\n", lc, filename);
				return -1;
			}
			current->hoplist.insert(hop->addr, hop);
			bool newmask = true;
			for (int i=0; i < nummasks && newmask; i++) {
				if (masklist[i] == hop->mask) {
					newmask = false;
				}
			}
			if (newmask) {
				masklist[nummasks++] = hop->mask;
			}
		} else {
			if (current) {
				network.insert(current->addr, current);
			}
			unsigned int addr = ntohl(inet_addr(tok));
			if ((tok = strtok(NULL, " \t\n")) == NULL) {
					// parse error
				delete current;
				dprintf(D_ALWAYS, "router: parse error on line %d of \"%s\"\n",
						lc, filename);
				return -1;
			}
			unsigned int mask = ntohl(inet_addr(tok));
			current = new NetworkRecord(addr, mask);
			bool newmask = true;
			for (int i=0; i < nummasks && newmask; i++) {
				if (masklist[i] == mask) {
					newmask = false;
				}
			}
			if (newmask) {
				masklist[nummasks++] = mask;
			}
		}
	}
	if (current) {
		network.insert(current->addr, current);
	}

	fclose(fp);

	// sort the masks in order of precedence
	qsort((void *)masklist, nummasks, sizeof(unsigned int),
		  (int (*)(const void *, const void *))maskcompar);

	return 0;
}

int
Router::NextHop(unsigned int &current, unsigned int src,
				unsigned int dest)
{
	NetworkRecord *rec = NULL;

	if (current == dest) {
		return 0;	// already reached the destination: no more hops
	}

	if (current == src) {
			// find the NetworkRecord for the source network
		for (int i=0; i < nummasks && !rec; i++) {
			unsigned int maskedsrc = src & masklist[i];
			network.lookup(maskedsrc, rec);
		}
		if (rec && rec->addr != src) { // found the second hop
			current = rec->addr;
			return 1;
		}
	} else {
		network.lookup(current, rec);
	}

	if (!rec) {
			// error: can't find source network in our table!
		dprintf(D_ALWAYS, "router: failed to find network for %s\n",
				inet_itoa(src));
		return -1;
	}

	if (last_dest != dest) {
		// find the NetworkRecord for the destination network
		destnet = NULL;
		for (int i=0; i < nummasks && !destnet; i++) {
			unsigned int maskeddest = dest & masklist[i];
			network.lookup(maskeddest, destnet);
		}

		if (!destnet) {
			// error: can't find destination network in our table!
			dprintf(D_ALWAYS, "router: failed to find network for %s\n",
					inet_itoa(dest));
			return -1;
		}

		last_dest = dest;
	}

	if (rec == destnet) {
		current = dest;
		return 1; // last hop
	}

	// find next hop
	NetworkRecord::NextHop *hop = NULL;
	if (rec->hopcache.addr == dest) {
		hop = rec->hopcache.nexthop;
	} else {
		for (int i=0; i < nummasks && !hop; i++) {
			unsigned int maskeddest = dest & masklist[i];
			rec->hoplist.lookup(maskeddest, hop);
		}
		rec->hopcache.addr = dest;
		rec->hopcache.nexthop = hop;
	}
	if (!hop) {
		// routing error!
		dprintf(D_ALWAYS, "router: failed to find hop from %s ",
				inet_itoa(rec->addr));
		dprintf(D_ALWAYS|D_NOHEADER, "for %s --> ", inet_itoa(src));
		dprintf(D_ALWAYS|D_NOHEADER, "%s connection\n", inet_itoa(dest));
		return -1;
	}
	else if (routing_debug) {
		dprintf(D_FULLDEBUG, "router: routing %s", inet_itoa(rec->addr));
		dprintf(D_FULLDEBUG|D_NOHEADER, " --> %s\n", inet_itoa(hop->net));
	}

	current = hop->net;
	return 1;
}

void
Router::Cleanup()
{
	NetworkRecord *current;
	network.startIterations();
	while (network.iterate(current)) {
		NetworkRecord::NextHop *hop = NULL;
		current->hoplist.startIterations();
		while (current->hoplist.iterate(hop)) {
			delete hop;
		}
		delete current;
	}
	network.clear();
	network.startIterations();
	if (network.iterate(current)) {
		abort();
	}
	last_dest = 0;
	destnet = NULL;
	nummasks = 0;
}
