/***************************Copyright-DO-NOT-REMOVE-THIS-LINE**
 * CONDOR Copyright Notice
 *
 * See LICENSE.TXT for additional notices and disclaimers.
 *
 * Copyright (c)1990-2001 CONDOR Team, Computer Sciences Department, 
 * University of Wisconsin-Madison, Madison, WI.  All Rights Reserved.  
 * No use of the CONDOR Software Program Source Code is authorized 
 * without the express consent of the CONDOR Team.  For more information 
 * contact: CONDOR Team, Attention: Professor Miron Livny, 
 * 7367 Computer Sciences, 1210 W. Dayton St., Madison, WI 53706-1685, 
 * (608) 262-0856 or miron@cs.wisc.edu.
 *
 * U.S. Government Rights Restrictions: Use, duplication, or disclosure 
 * by the U.S. Government is subject to restrictions as set forth in 
 * subparagraph (c)(1)(ii) of The Rights in Technical Data and Computer 
 * Software clause at DFARS 252.227-7013 or subparagraphs (c)(1) and 
 * (2) of Commercial Computer Software-Restricted Rights at 48 CFR 
 * 52.227-19, as applicable, CONDOR Team, Attention: Professor Miron 
 * Livny, 7367 Computer Sciences, 1210 W. Dayton St., Madison, 
 * WI 53706-1685, (608) 262-0856 or miron@cs.wisc.edu.
 ****************************Copyright-DO-NOT-REMOVE-THIS-LINE**/

/* File:    dag_property.C
   Author:  ningl
   Purpose: Class implementation for DagProperty
   Date:    10/15/02
*/

#include "condor_common.h"
#include "dag_property.h"

extern char* ignore_spaces (char * cur);
bool isSpace (char * c);
char * getFraction (char* cur, float & fraction, bool fraction_first);


Level::Level (string name) 
{
    this->name = name;
    number_of_nodes = 0;

    info.reserve (4);
    
    for (int i=0; i<4; i++) {
	info [i] = new list<level_nf_pair>;
    }
}

Level::~Level ()
{
    for (int i=0; i<4; i++) {
	delete info[i];
    }
}

DagProperty::DagProperty ()
{
    level_info = new level_map;
}


DagProperty::~DagProperty ()
{
    if(level_info) {
	level_info->clear();
	delete level_info;
    }
}

bool isSpace (char * c)
{
    if (*c == ' ' || *c == '\t') return true;
    return false;
}

bool DagProperty::ProcessConstraintString (string constraint)
{

    char * cur = strdup (constraint.c_str());
    string lhs, rhs;
    
    cur = ignore_spaces (cur);
    
    if (*cur != '[') {
	fprintf(stderr, "ERROR: dag constraint should be specified within square brackets\n");
	return false;
    }
    cur++;
    while (*cur != '\0') {
	// parse by line
	cur = ignore_spaces (cur);

	if (cur == '\0') return false;
	
	// Gather left hand side information
	lhs = "";
	while (*cur!='=' && !isSpace (cur)) {
	    lhs+= *cur;
	    cur++;
	}
	cur = ignore_spaces(cur);
	
	if (*cur != '=') {
	    fprintf(stderr, "ERROR: missing = at dag constraint\n");
	    return false;
	}
	cur ++;

	cur = ignore_spaces (cur);

	rhs = "";

	if (lhs.find('.') == lhs.npos) {
	    // This is a property definition
	    if (*cur!='\"') {
		fprintf(stderr, "ERROR: missing \" when specifying node contraint\n");
		return false;
	    }
	    cur ++;
	    while (cur != '\0') {
		if (*cur == '\n') {
		    char * tmp = cur;
		    while (*tmp==' ' || *tmp=='\t') {tmp--;}
		    if (*tmp != '\\') {
			fprintf(stderr, "ERROR: no line escape char found at the end of the line\n");
			return false;
		    }
		}
		if (*cur == '\"') {
		    if (*(cur-1)!='\\') {
			// Look for ";" even though it's not necessary, need to skip it
			cur = ignore_spaces (cur);
			cur ++;
			if (*cur == ';') {
			    cur++;
			}
			// found back quotes ", can stop now
			break;
		    }
		}
		if (*cur == ';') {
		    cerr<<"ERROR: Unterminated quotes for definition of "<<lhs<<endl;
		}
		rhs += *cur;
		cur++;
	    }
	    if (!ProcessPropertyDefinition (lhs, rhs)) {
		cerr<<"ERROR: Could not process property definition "<<
		    lhs<<" = "<<rhs<<endl;
		return false;
	    }
	} 
	else {
	    // This is a level definition 
	    // Do not allow new line in the level definitions 
	    while (*cur!=';') {
		rhs+= *cur;
		cur++;
	    }
	    rhs+=*cur;
	    cur++;
	    if (!ProcessLevelDefinition (lhs, rhs)) {
		cerr<<"ERROR: Could not process level definition "<<
		    lhs<<" = "<<rhs<<endl;
		return false;
	    }
	}
	
	cur = ignore_spaces (cur);
	if (cur == '\0') {
	    fprintf(stderr, "ERROR: dag constraint must be terminated using right bracket ]\n");
	}
	if (*cur == ']') {
	    cur++;
	    cur = ignore_spaces (cur);
	    if (*cur != '\0') {
		fprintf(stderr, "ERROR: parse error after ]\n");
		return false;
	    }
	}
    }
    
    return true;
}


bool DagProperty::ProcessPropertyDefinition (string lhs, string rhs)
{
    if (properties.find(lhs)!=properties.end()) {
	printf("WARNING: duplicated property definition of %s\n", lhs.c_str());
    }

    // Take out quotes from this
    const char * c = rhs.c_str();
    string req ="";

    while (*c!='\0') {
	if (!(*c== '\\' && *(c+1) == '"')) {
	    req+= *c;
	}
	c++;
    }
    properties[lhs] = req;
    return true;
}
    
bool DagProperty::ProcessLevelDefinition (string lhs, string rhs)
{

    //cout<<"\nDagProperty::ProcessLevelDefinition "<<lhs<<" = "<<rhs<<endl;

    qualifier q; 
    
    Level * current_level;
    string level_name = "";
    string level_qualifier = "";

    int node_size = 0;
    int pos = lhs.find (".");
    
    level_name = lhs.substr(0, pos);
    level_qualifier = lhs.substr(pos+1, lhs.size()-1);

    float fraction;
    string name; // temporary variable to get name of the level that 
                 // a fraction is referring to
        
    char * cur = strdup (rhs.c_str());
    
    cur = ignore_spaces (cur);


    if (!strcasecmp (level_qualifier.c_str(), "number")) {
	q = NUMBER;
    }
    else if (!strcasecmp (level_qualifier.c_str(), "parent")) {
	q = PARENT;
    }
    else if (!strcasecmp (level_qualifier.c_str(), "child")) {
	q = CHILD;
    }
    else if (!strcasecmp (level_qualifier.c_str(), "property")) {
	q = PROPERTY;
    }  
    else {
	fprintf(stderr, "ERROR: %s is not a valid level qualifer\n",
		level_qualifier.c_str()); 
	free ((void*) cur); //????
	return false;
    }

    //cout<<"\tLevel name = "<<level_name<<endl;
    //cout<<"\tLevel qualifier = "<<level_qualifier<<endl;

    // Create new level with level_name if one does not exist
    if (level_info->find(level_name) == level_info->end()) {
	current_level = new Level (level_name);
	(*level_info)[level_name] = current_level;
	//cout<<"Created new level in level_info-> "<<level_name<<endl;
    } else {
	current_level = (level_info->find(level_name))->second;
    }
    
    // Check which type of level qualifier we are processing
    if (q==NUMBER) {
	if (*cur <'0' || *cur>'9') {
	    fprintf(stderr, "ERROR: The max number of nodes at a level must be a number\n");
	    free ((void*) cur); // ??
	    return false;
	}
	
	string node_num = "";
	while (*cur!= ';' && !isSpace(cur) && *cur!='\n') {
	    node_num += *cur;
	    cur++;
	}
	node_size = atoi (node_num.c_str());
	current_level->set_node_num (node_size);

	//cout<<"DagProperty::ProcessLevelDefinition set node-num = "<<
	//((level_info->find(level_name))->second)->number_of_nodes<<endl;
    
    }
    else {
	// Qualifier is not a number

	while (cur) {
	    // Loop through all the characters in current rhs
	   
	    name="";

	    if ((*cur>='0' && *cur<='9')||*cur == '.') {
		// This is a fraction * name format 

		//cout <<"\tThis is a fraction * name format "<<endl;
		cur = getFraction (cur, fraction, true);

		if (!cur) {
		    fprintf(stderr, "ERROR: The specification for level %s is not correct\n",
			    level_name.c_str());
		    return false;
		}
		cur = ignore_spaces (cur);
		if (*cur!='*') {
		    fprintf(stderr, "ERROR: The format should be fraction * level_name\n");
		    return false;
		}
		cur++;
		cur = ignore_spaces (cur);
		
		while (*cur != ';' && !isSpace(cur) && *cur!='+') {
		    name += *cur;
		    cur ++;
		}
		//cout<<"\treferenced level/property name = "<<name<<endl;
		
	    }
	    else {
		// This is a name * fraction format
		
		//cout<<"\tThis is a name * fraction format"<<endl;
		while (*cur!='*' && !isSpace (cur)) {
		    name += *cur;
		    cur++;
		}

		cur = ignore_spaces (cur);
		if (*cur!='*') {
		    fprintf(stderr, "ERROR: The format should be level name * fraction\n");
		    return false;
		}
		cur ++;
		cur = ignore_spaces (cur);
		if ((*cur<'0' || *cur>'9') && *cur != '.') {
		    fprintf(stderr, "ERROR: incomplete specification for level %s\n",
			    level_name.c_str());
		    return false;
		}
		cur = getFraction (cur, fraction, false);
		if (!cur) {
		    fprintf(stderr, "ERROR: The specification for level %s is not correct\n",
			    level_name.c_str());
		    return false;
		}
		//cout<<"\treferenced level/property name = "<<name<<endl;
	    }
	    
	    // The following logic to check whether there are more specifications 
	    // for this level
	    cur = ignore_spaces (cur);
	    
	    //cout<<"Referenced level = "<<name<<" fraction = "<<fraction<<endl;
	    // Create level_info
	    struct lf_pair lf;
	    lf.name = name;
	    lf.fraction = fraction;


	    if (q==PROPERTY) {
		// check whether referenced property exists 
		if (properties.find(name) == properties.end()) {
		    cout<<"WARNING: property "<<name<<" not found yet"<<endl;
		}
	    }
	    else {
		// Check whether referenced level exists
		// Create it if it does not exists
		if (level_info->find(name) == level_info->end()) {
		    Level * newl = new Level (name);
		    (*level_info)[name] = newl;
		    //cout<<"*.*Created new level "<<level_name<<endl;
		}
	    }

	    // Insert new level info to the level info vector
	    current_level->info[q]->push_back(lf);
		
	    
	    if (*cur == ';') 
		break;
	    else if (*cur == '+') {
		    cur++;
		    cur = ignore_spaces (cur);
	    }
	    else {
		cerr<<"ERROR: Wrong specification for "<<lhs<<endl;
		return false;
	    }
	}
    }

    return true;
}

// Given property name, return actual requirement value
string DagProperty::GetRequirement (string pname) 
{
    property_map::iterator itr = properties.find(pname);
    if (itr!=properties.end()) {
	return itr->second;
    }
    else {
	return "";
    }
}


char * getFraction (char* cur, float & fraction, bool fraction_first)
{
    string int_part = "";
    string fl_part = "";

    bool float_p = true;

    
    while (*cur >= '0' && *cur<='9') {
	int_part += *cur;
	cur++;
    }
    
    if (*cur == '.') {
	// process float part
	cur ++; 
	
	while (*cur >= '0' && *cur<='9') {
	    fl_part+=*cur;
	    cur ++;
	}
    }
    
    cur = ignore_spaces (cur);
    if (fraction_first && *cur != '*') {
	cerr<<"ERROR: no * found between fraction and name\n";
	return NULL;
    }
    
    int intPart=0, flPart=0;
    if (int_part != "") {
	intPart = atoi(int_part.c_str());
    }
    if (fl_part != "") {
	flPart = atoi(fl_part.c_str());
    }
    fraction = (float) flPart;
    fraction = fraction / (pow (10.0, fl_part.size()));
    fraction += (float) intPart;

    return cur;
}

void DagProperty::PrintLevelInfo ()
{
    cout<<"DagProperty::PrintLevelInfo"<<endl;
    level_map::iterator itr = level_info->begin();
    
    
    for (; itr!=level_info->end(); itr++) {
	Level * l= itr->second;

	cout<<"Level "<<l->name <<" contains " << l->number_of_nodes<<" nodes"<<endl;
    }
    
}
