#include "condor_common.h"
#include <stdio.h>
#include <stdlib.h>
#include <iostream.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <string.h>
#include <time.h>
#include "condor_md.h"


/* The following are error exit codes */
#define OUT_EXISTS      127  /* Output file already exists */
#define NO_LOGGING      126  /* Cannot open any log file */
#define START_TIME_ERR  125  /* Cannot get start time */
#define END_TIME_ERR    124  /* Cannot get start time */

#define TIME_ARG        123  /* Incorrect argument after -t*/ 
#define FILESIZE_ARG    122  /* Incorrect argument for file size after -max*/
#define INCORRECT_USAGE 121  /* Incorrect usage */
#define OUTFILE_ERROR   120  /* Cannot open output file for writing*/
#define NO_OUTFILE      119  /* No output file specified after -o */
#define INPUT_ERROR     118  /* Incorrect input */

pid_t pid; 
FILE * mdfile_log;

void writelog (const char*);
void printusage (const char*);
void WriteMDtoLog (unsigned char []);

void WriteOutputFile (FILE * file, const char * filename, unsigned int size);
bool MDFile (char*);
void MDPrint (unsigned char[16]);
bool MDInput (char *filename, FILE * file);
bool MDCompare (unsigned char[], unsigned char[]);

int main(int argc, char** argv)
{
    unsigned int MAXBYTES = 5;
    bool error = false;
    int time_to_run = 0, i=1, fix_time=0, gone_time =0; 
    struct timeval start_tv, end_tv;

    FILE * output = NULL, * input = NULL;
    char * currentInput;

    char out [35];
    pid = getpid();
    
    /* Open log for writing. 
       Default log file goes to mdfile.log, otherwise mdlog1.log or mdlog2.log 
       Will exit if none of the log files can be opened for writing
    */

    mdfile_log = fopen ("mdfile.log", "a"); 
    if (mdfile_log==NULL)
    {
	char logname[35];
	sprintf(logname, "%d.log", pid);
	mdfile_log = fopen(logname, "a");
	fprintf (mdfile_log, "%d: This log is generated by %s\n",
		 pid, argv[0]);

	if (mdfile_log==NULL)
        {
	    cerr<<"Could not open log file for logging"<<endl;
	    exit(NO_LOGGING);
	}
    }

    writelog("\nCommand invoked is ");
    for (int i=0; i<argc; i++)
    {
	fprintf(mdfile_log, "%s ", argv[i]);
	fprintf(stderr, "%s ", argv[i]);
    }
    fprintf(mdfile_log, "\n");
    fprintf(stderr, "\n");

    if (argc<4)
    {
	printusage(argv[0]);
	writelog("Incorrect Usage\n\n");
	cerr<<"Closing log"<<endl;
	fclose (mdfile_log);
	exit(INCORRECT_USAGE);
    }
    
    if (gettimeofday (&start_tv, NULL)==-1)
    {
	perror("couldn't get start time");
	writelog("Couldn't get start time\n\n");
	cerr<<"Closing log"<<endl;
	fclose (mdfile_log);
	exit (START_TIME_ERR);
    }
  
 read_args:
    while (i<argc)
    {
	if (!strcmp (argv[i], "-t"))
	{
	    //if this argument is given, there must be at least 6 args
	    if (argc<6)
            {
		cerr<<"Not enough number of arguments"<<endl;
		printusage(argv[0]);
		writelog("Not enough number of arguments\n");
		exit(TIME_ARG); 
	    }
	    //record the time 
	    time_to_run = atoi(argv[i+1]);
	    printf("Will run for %d seconds\n", time_to_run);
	    fprintf(mdfile_log, "%d: Will run for %d seconds\n", pid, time_to_run);
	    i+=2;
	    goto read_args;
	}
    
	if (!strcmp (argv[i], "-o"))
        {
	    sprintf(out, "%s.out", argv[i+1]);
	   
	    //Check if output already exists 
	    struct stat filestat; 
	    if (!stat (out, &filestat))
            {
		writelog ("Output already exists!");
		exit (OUT_EXISTS);
	    }
	    output = fopen (out, "w");
	    if (output==NULL)
            {
		perror("output unsuccessfull");
		writelog("Cannot open output for writing\n");
		exit (OUTFILE_ERROR);
	    }
	    i+=2;
	    goto read_args; 
	}

	if (!strcmp (argv[i], "-max"))
        {
	    if (argc<6)
	    {
		cerr<<"Not enough number of arguments"<<endl;
		printusage(argv[0]);
		writelog("Not enough number of arguments\n");
		exit (FILESIZE_ARG); 
	    }
     
	    MAXBYTES = atoi(argv[i+1]);
	    if (MAXBYTES <1 || MAXBYTES>10000000) //1-10MB ok, else set it to just 10K
		MAXBYTES = 10001;
	    i+=2;
	    goto read_args;
	}

	/* Opens input file */
	input = NULL; 
	currentInput = argv[i];
	input = fopen(currentInput, "r"); 
	if (input == NULL)
        {
	    fprintf(stderr, "Could not open input %s \n", argv[i]);
	    fprintf(mdfile_log, "%d: Could not open input %s \n", pid, argv[i]);
	    break;
	}

	/* If file can be opened, process input file 
	   - Reads in input file
	   - Calculate MD5 checksum
	   - Compare it with MD5 checksum supplied 
	*/
	else
	{
	    if (MDInput (argv[i], input))
            {
		fprintf(mdfile_log, "Successfully MD5 checksumed %s\n", argv[i]);
		fprintf (mdfile_log, "%d SUCCESS: %s MD5 checksum matched\n", pid, argv[i]);
	    }
	    else {
		error = true;
		if (output) {
		    writelog ("Writing MD checksum msg to output file\n");
		    fprintf(mdfile_log, "%d ERROR: %s failed MD 5 checksum\n", pid, argv[i]);
		}
	    }
	}
    
	i++;
    }
  
    if (output==NULL)
    {
	error = true; 
	fprintf(stderr, "Please specify output file using -o option\n");
	writelog("Please specify output file using -o option\n");
	exit (NO_OUTFILE);
    }
  
    else 
    {
	if (error)
	{
	    writelog ("Writing output for error case\n");
	    if (input == NULL)
	    {
		fprintf(mdfile_log, "%d ERROR: %s could not be opened or did not exists\n",
			pid, currentInput);
	    }
	    WriteOutputFile (output, out, MAXBYTES);
	    fflush (output);
	    if (output) fclose(output);
	    exit (INPUT_ERROR);
	    
	}
	else
	{
	    writelog ("Writing output for success case\n");
	    WriteOutputFile (output, out, MAXBYTES);
	}
    }
    cerr<<"Closing output"<<endl;
    if (output)
      fclose (output);
    if (gettimeofday (&end_tv, NULL)==-1)
  {
    perror("couldn't get end time");
    writelog("Couldn't get end time\n");
    exit (END_TIME_ERR);
  }
  
  gone_time = (end_tv.tv_sec-start_tv.tv_sec)*1000000
    +end_tv.tv_usec-start_tv.tv_usec;
  
  //printf("Gone time = %d\n", gone_time);
  if (time_to_run >=0) //test how long it takes to run 10000 loops 
  {
    int counter = 10000;
    
    if (gettimeofday (&start_tv, NULL)==-1)
    {
      perror("couldn't get start time");
      writelog("Couldn't get start time\n");
      exit (START_TIME_ERR);
    }
    while (counter>0)
    {
      counter--; 
    }
    
    if (gettimeofday (&end_tv, NULL)==-1)
    {
      perror("couldn't get end time");
      writelog("Couldn't get end time\n");
      exit (END_TIME_ERR);
    }

    fix_time = (end_tv.tv_sec-start_tv.tv_sec)*1000000 +end_tv.tv_usec-start_tv.tv_usec;
    time_to_run*=1000000;
    time_to_run=time_to_run-gone_time-fix_time; 
    int loopnumber = (int)((time_to_run)/fix_time);

    // printf("fix time = %d for 10000 loops, Now we need %d loops for 
    // %d seconds\n", fix_time, loopnumber, time_to_run);
    for (i=0; i<loopnumber; i++)
    {
      counter = 10000;
      while (counter>0)
	counter--;
    }
  }
  if (!error)
    writelog("Successfully finished!\n\n");
  else
    writelog("Failed\n\n");
  cerr<<"Closing log"<<endl;
  fclose (mdfile_log);
  return 0; 
}


  


void writelog (const char * logmsg)
{
    /*
      struct time_t timenow; 
      
      if (gettimeofday (&timenow, NULL)==-1)
      {
      fprintf(mdfile_log, "%d: %s", pid,"Couldn't get current time\n");
      }
      char * curtime = ctime(&timenow);
      
      fprintf(mdfile_log, "%d: %s %s", pid, curtime, logmsg);
    */
    
    fprintf(stderr, "%d: %s", pid, logmsg);
    fprintf(mdfile_log, "%d: %s", pid, logmsg);
}

void printusage(const char * argv0)
{

  fprintf(stderr, "Usage: %s \n\t -o\t<output file name> <list of input files>\n", argv0);
  cerr<<"\t[-t\t<time_in_secs to run>]\n\t[-max\t<max number of bytes per output file>]\n\n";
}


bool MDInput (char* filename, FILE*file)
{
    Condor_MD_MAC context;
    unsigned long len = 1024;
    unsigned char uline[1024], oldsum[16];
    unsigned char * digest = new unsigned char [16];
    char md[strlen(filename)+3];
    FILE * mdfile;

    sprintf (md, "%s.md", filename);

    if ((file = fopen (filename, "r")) == NULL) {
	fprintf(mdfile_log, "Could not open input file %s\n", filename);
	return false;
    }

    if ((mdfile = fopen (md, "r")) == NULL) {
	fprintf(mdfile_log, "%d: Could not open md file %s\n", 
		pid, md);
	return false;
    }	
    
    //Find the "Start:" token
    while (fread(uline, 1, len, file))
    {
	context.addMD(uline, len);
    }
    digest = context.computeMD();
    
    if (fread (oldsum, 16, 1, mdfile)!=16)
	writelog("Old mdsum is not 16 bytes!\n"); 
    bool outcome = false;
    outcome = MDCompare (oldsum, digest);
    fclose (mdfile);
    if (outcome) {
	writelog ("removed md file\n");
	unlink (md);
    }
    return outcome;
}

/* Process input file 
   - Reads in input file
   - Calculate MD5 checksum
   - Compare it with MD5 checksum supplied 
*/
bool MDInput2 (char* filename, FILE* file)
{
    Condor_MD_MAC context;
    int len = 1024;
    unsigned char uline[1024], uline2[1024];
    unsigned char * digest = new unsigned char [16];
    int nbytes, sumbytes; //Number of bytes read
    
    file = fopen (filename, "r");
  
    //Find the "Start:" token
    
    while ((nbytes =fread(uline, 1, len, file))) {
	fprintf(stderr, " bytes read = %d \n", nbytes);
	if (nbytes != 17) {
	    writelog("nbytes !=17 case\n");
	    context.addMD (uline, (unsigned long) nbytes);
	}

	else {
	    writelog("nbytes == 17 case\n");
	    // This is not yet the last line, uline2 is the real old MD sum
	    
	    if ((sumbytes = fread(uline2, 1, len, file))!=0) {
		fprintf(mdfile_log, "I'm reading another %d bytes\n", sumbytes);
		
		context.addMD(uline, (unsigned long) nbytes);
	        digest = context.computeMD();
		
		if (sumbytes !=17) {
		    writelog ("The last line of the input isn't 16 bytes long!");
		    return false;
		}
		else {
		    writelog ("Calling MDCompare ... \n");
		    return MDCompare (uline2, digest);
		}
	    }
	    else {
		digest =  context.computeMD();
		return MDCompare (uline, digest);
		break;
	    }
	}
    }
    writelog("No reading??\n\n");
    return false;
}

bool MDCompare (unsigned char oldMDSum[], unsigned char digest [])
{
    for (int i=0; i<16; i++)
    {
	if (oldMDSum[i]!=digest[i])
	{
	    writelog ("MDSum read in is  ");
	    WriteMDtoLog  (oldMDSum);
	    
	    writelog ("MDSum calcaulated ");
	    WriteMDtoLog  (digest);
	    return false;
	}
    }
    writelog("MDSum matched\n");
    printf("MDSum matched!\n");
    return true;
}

/* Process input file 
   - Reads in input file
   - Calculate MD5 checksum
   - Compare it with MD5 checksum supplied 
*/

/*
bool MDInput2 (char *filename, FILE * file)
{
  MD5_CTX context;
  int len = 1024;
  char line[1024];
  unsigned char uline[1024], digest[16], oldMDSum[16];
  bool found = false;
  FILE *tempfile;
  char newfile [50];
  
  sprintf (newfile, "%s.temp", filename);
  if ((tempfile = fopen(newfile, "a"))==NULL)
  {
      writelog ("Could not open temporary file for MD5 checksum\n");
      return false;
  }

  if (file)
      fclose (file);
  file = fopen (filename, "r");
  //Find the "Start:" token
  while (fgets (line, len, file))
  {
      if (!strcmp (strtok(line, "\n"), "Start:"))
      {
	  printf("Found start token\n");
	  writelog ("Found start token in input file \n");
	  found = true;
	  break;
      }
  }
  if (!found)
  {
    writelog ("Could not find start token in input\n");
    return false;
  
  }
  found = false;
  
  
  char *templine =  NULL;
  //Process the middle portion until reached for end
  while (fgets(line, len, file))
  {
      fprintf(stderr, "Looking at %s\n", line);
      strtok (line, "\n");
      if (line) {
	  if (!strcmp(line, "End")) {
	      printf("Found end token\n");
	      writelog ("Found end token in input\n");
	      found = true;
	      break;
	  }
      }

      //Copy middle portion to a temporary file
      fprintf(tempfile, "%s", line);
  }
  
  if (!found)
  {
    writelog ("Could not find end token in input\n");
    fclose (file);
    fclose (tempfile);
    return false;
  }

  //MD5 checksum the temporary file
  MD5Init (&context);

  if (tempfile)
      fclose (tempfile);
  tempfile = fopen (newfile, "r");
  while (fread(uline, 1, len, tempfile))
      MD5Update (&context, uline, len);

  MD5Final (digest, &context);
  //printf ("MD5 (%s) = ", filename); 
  //MDPrint (digest);
  
  //Read out and compare the MD checksum from the input file
  
  
  if (fread (oldMDSum, 16, 1, mdfile)==0)
  {
      writelog ("Error reading MD checksum!\n");
      return false;
  }
  
  for (int i=0; i<16; i++)
  {
      if (oldMDSum[i]!=digest[i])
      {
	  fprintf(mdfile_log, "Old MDSum = ");
	  fprintf(stderr, "Old MDSum = ");
	  MDPrint(oldMDSum);
	  fprintf(mdfile_log, "New MDSum = ");
	  fprintf(stderr, "New MDSum = ");
	  MDPrint(digest);
	  return false;
      }
  }
  writelog("MDSum matched\n");
  printf("MDSum matched!\n");
  
  //Clean up remove temp file
  fclose (tempfile);
  //unlink (newfile);
  return true;
    
}
*/

void WriteOutputFile (FILE * output, const char * ofilename, unsigned int MAXBYTES) 
{
    unsigned char uline[1024]; 
    unsigned char * digest = new unsigned char [16];
    unsigned int len = 0;
    char line [1024];
    Condor_MD_MAC context; 
            
    char md [strlen (ofilename)+3];
    
    sprintf(md, "%s.md", ofilename);
    FILE * mdout = fopen(md, "w");

    srand (time(NULL));
    unsigned int bytes = (unsigned int) (rand()%MAXBYTES+1);
    fprintf(stderr, "%d: Writing output file of size: %d\n", pid, bytes);
    fprintf(mdfile_log, "%d: Writing output file of size: %d\n", pid, bytes);
    
    fprintf(mdfile_log, "%d Writing output file %s with total %d bytes\n", 
	    pid, ofilename, bytes);
    
    for (unsigned int i=0; i<bytes; i++) {
	int randnum = (int)(10.0*rand()/(RAND_MAX+1.0));
	if (!i%1023 && i!=0) {
	    fprintf (output, "\n");
	}
	else 
	    fprintf (output, "%d", randnum);
    }
    
    fclose (output);
    output = fopen (ofilename, "r");

    while (fread (uline, 1, len, output))
	context.addMD (uline, (unsigned long) len);
    digest = context.computeMD();
    
    //Append the tempfile content to output
    fprintf(mdfile_log, "%d End writing content to output %s\n",
	    pid, ofilename);

    writelog ("MDSum = ");
    WriteMDtoLog (digest);
    fwrite (digest, 16, 1, mdout); 
    fclose (mdout);
    delete [] digest;
}



//MD5 checksum output file
bool MDFile (char * filename)
{
    FILE *file;
    Condor_MD_MAC context;
    unsigned long len;
    unsigned char buffer[1024];
    unsigned char * digest = new unsigned char[16];
    
    if ((file = fopen (filename, "rb")) == NULL)
    {
	fprintf (mdfile_log, "%s can't be opened\n", filename);
	return 0;
    }
  
    while ((len = fread (buffer, 1, 1024, file)))
        context.addMD (buffer, len);
    digest = context.computeMD();
    cerr<<"Closing file"<<endl;
    fclose (file);
    
    printf ("MD5 (%s) = ", filename);
    fprintf(mdfile_log, "MD5 (%s) = ", filename);
    
    MDPrint (digest);
    printf ("\n");
    delete [] digest;
    return true;
}

void MDPrint (unsigned char * digest)
{
  unsigned int i;
  
  for (i = 0; i < 16; i++)
  {
    printf ("%02x", digest[i]);
    fprintf(mdfile_log, "%02x", digest[i]);
  }
  printf("\n");
  fprintf(mdfile_log, "\n");
}

void WriteMDtoLog (unsigned char digest[16])
{
    for (unsigned int i=0; i<16; i++)
    {
	fprintf(mdfile_log, "%02x", digest[i]);
    }
    fprintf(mdfile_log, "\n");
}
