/***************************Copyright-DO-NOT-REMOVE-THIS-LINE**
 * CONDOR Copyright Notice
 *
 * See LICENSE.TXT for additional notices and disclaimers.
 *
 * Copyright (c)1990-2001 CONDOR Team, Computer Sciences Department, 
 * University of Wisconsin-Madison, Madison, WI.  All Rights Reserved.  
 * No use of the CONDOR Software Program Source Code is authorized 
 * without the express consent of the CONDOR Team.  For more information 
 * contact: CONDOR Team, Attention: Professor Miron Livny, 
 * 7367 Computer Sciences, 1210 W. Dayton St., Madison, WI 53706-1685, 
 * (608) 262-0856 or miron@cs.wisc.edu.
 *
 * U.S. Government Rights Restrictions: Use, duplication, or disclosure 
 * by the U.S. Government is subject to restrictions as set forth in 
 * subparagraph (c)(1)(ii) of The Rights in Technical Data and Computer 
 * Software clause at DFARS 252.227-7013 or subparagraphs (c)(1) and 
 * (2) of Commercial Computer Software-Restricted Rights at 48 CFR 
 * 52.227-19, as applicable, CONDOR Team, Attention: Professor Miron 
 * Livny, 7367 Computer Sciences, 1210 W. Dayton St., Madison, 
 * WI 53706-1685, (608) 262-0856 or miron@cs.wisc.edu.
 ****************************Copyright-DO-NOT-REMOVE-THIS-LINE**/

/* File:    runall.C
   Author:  ningl
   Purpose: This program makes calls to the dag_gen program to randomly
            create dags used by dagman.  The program creates as many 
	    dags as supplied by the last argument to the program. 
	    Each dag is then submitted to the condor system with 5 secs
	    interval (otherwise, there would be too many AFS accesses
	    per second)
*/

using namespace std;

# include "condor_common.h"

#if (__GNUC__<3)
#include <hash_map>
#else
#include <ext/hash_map>
#endif

#include "ddag.h"
#include "dag_info.h"
#include "constraint.h"
#include "dag_property.h"

char *mySubSystem = "CONDOR_DAGMAN_TEST";
# define FAIL    false
# define SUCCEED true
# define CONDOR_CONFIG "condor_config"

bool INTELLIGENCE;
int  dag_num;
int  job_num;

bool process_log (const char * path);
void printusage  (char * argv0);
bool setup_config (char * config_file_name, char* &old_config_path);
bool revert_config_changes (char* old_env_value);

int  GenerateDagFiles (const char* path, PoolInfo * pool);
bool submitdag   (const char* dagname);
bool removedags  ();
void writelog    (const char * msg);
bool clean       (const char* path); //deletes path
void logerror    (char * fmt, ...);
void printtime   (FILE * file);
void sleepfor    (int sec);

// implementation in condor_c++_utils/condor_config.C
extern "C"{
    void init_tilde();
    char* find_file(const char *env_name, const char *file_name);
}

void printads (ClassAdList &);
void sortads (ClassAdList &);
void parseStartAd (char * start_string);
bool isAvailable (ClassAd * ad);

typedef hash_map <const char*, DagNode*, hash<const char*>, eqstr> jobs_hash_map;

// Global variables
FILE * runall_log = NULL; 
FILE * LOG;
FILE * SUMMARY_LOG;
set    <dag_info *> dags;
TESTMODE test_mode;

/* These variables are shared by all the files*/
PoolInfo * pool; 
DagConstraints * constraints;
DagProperty * dag_property;

int main(int argc, char** argv)
{
    int         attempt = 0, i=0, rand_num;

    //number of minutes before exiting and after max # dags reached
    int         MAX_ATTEMPT = 30; 
    struct stat stats, mystat;
    bool        status; 

    char *      exec_name1;
    char *      dagpath;    
    char *      log_dir = NULL;
    char *      old_config_path;

    bool        checked_proxy = false;          //needed for globus jobs
    bool        checked_condor_version = false; //needed for globus jobs
    if (argc !=4)
    {
	printusage(argv[0]);
	exit (-1);
    }
    
    // initialize to read from config file
    myDistro->Init( argc, argv );
    
    old_config_path = NULL;

    // Set up config file & location
    if (!setup_config (argv[2], old_config_path)) {
	printf("Could not set up config file\n");
	exit (-1);
    }
    
    config();

    // Revert configuration changes 
    if (!revert_config_changes(old_config_path)) {
	printf("Could not revert env settings for CONDOR_CONFIG\n");
	exit (-1);
    }
    
 
    // Check if the log directory exists
    // if not, it will be created
    log_dir = param ("LOG_DIRECTORY"); 

    if (!log_dir) {
	log_dir = "log";
	printf("Using default log directory: ./log\n");
    }

    if (stat (log_dir, &stats) != 0) {
	if (mkdir (log_dir, 0700) != 0) {
	    fprintf(stderr, "Could not mkdir %s\n", log_dir);
	    exit(-1);
	}
    }
    
    
    char  logname [strlen(log_dir)+35];
   
    // Opens log file for current process
    sprintf(logname, "%s/%d.log", log_dir, getpid());

    if (log_dir) {
	free (log_dir);
    }

    if ((runall_log = fopen (logname, "w"))==NULL) {
	perror ("Could not open log!");
	exit (-1);
    }
    
    
    // Settings for "intelligence"    
    char * intelligence = param ("INTELLIGENCE");
    pool = NULL;
    if (intelligence) {
	if (!strcasecmp (intelligence, "TRUE")) {
	    INTELLIGENCE = true;
	}
	free (intelligence);
    }
    
    exec_name1 = param ("TEST_PROGRAM_NAME");
    if (!exec_name1) {
	    exec_name1= "testjob";
	    printf("Warning: using default testjob name: %s\n", exec_name1);
    }
    
    // Get the working directory name & number of dags 
    dagpath = argv[1];
    
    int max_dag_num = atoi(argv[3]);
    if (max_dag_num < 0) max_dag_num = 10;
    
    //Find out whether all the executables are in place
    if (stat (exec_name1, &mystat)!=0) {
	writelog("Could not find testjob");
	printf("Cannot open %s\n", exec_name1);
	return -1;
    }
        
     // Make or verify existance of the directory to put all the dags
    if (stat (dagpath, &stats)==-1)
    {
	printtime(runall_log);
	fprintf(runall_log, "Creating directory %s\n", dagpath);
	if (mkdir(dagpath, 0722)!=0)
        {
	    printtime(runall_log);
	    fprintf(runall_log, "Could not mkdir %s\n", dagpath);
	    printf("Could not mkdir %s\n", dagpath);
	    exit (-1);
	}
    }
    else 
    {
	printtime(runall_log);
	fprintf(runall_log, "The directory %s already exists\n", dagpath);
    }
 

    // Set the max number of attempts to clean up dags in directory 
    // after the max quota met
    char * attempt_num = param("TIME_OUT");    
    if (attempt_num) {
	int attempts = atoi (attempt_num);
	if (attempts> 0) {
	    MAX_ATTEMPT = attempts;
	}
	free (attempt_num);
    }


    char * dag_constraint = param ("DAG_CONSTRAINT");
    dag_property = NULL;
    if (dag_constraint) {
	INTELLIGENCE = false;
	string dag_constr = dag_constraint;
	dag_property = new DagProperty ();
	
	if (!dag_property->ProcessConstraintString (dag_constr)) {
	    fprintf(stderr, "Could not successfully parse dag constraint\n");
	    delete dag_property;
	    dag_property = NULL;
	}
    }


    if (INTELLIGENCE) {
	cout<<"INTELLIGENCE DEFINED"<<endl;
 	// Record initial machines ads that are present in the pool
	ClassAdList startdAds;
	
	if (!obtainAdsFromCollector(startdAds)) {
	    dprintf(D_ALWAYS, "Could not fetch startd ads\n");
	     fprintf(stderr, "Could not fetch startd ads\n");
	     exit (-1);
	} 
	
	pool = new PoolInfo (&startdAds, 5);

	int loops = 3;
	int wait_time = 60;
	
	char * loop_num = param ("ITERATION"); 
	char * wait_t = param ("WAIT_TIME");

	if (loop_num) {
	    loops = atoi (loop_num);
	    free (loop_num);
	}
	if (loops <0) loops = 3;

	if (wait_t)  {
	    wait_time = atoi (wait_t);
	    free (wait_t);
	}
	if (wait_time <0) wait_time = 60;

	//ningl -- 10/10 temporarily removed findstatic
	//pool->FindStaticAttributes (wait_time, loops);
    }

    // Set up dag constraints specified by user
    constraints = NULL;
    
    char * constr = param("CONSTRAINT");

    if (constr && !dag_constraint) {
	constraints = new DagConstraints(constr);
        
	if (!constraints) {
	    cerr<<"Could not process constraints on dag node requirements\n";
	    exit (-1);
	}
    }
    
    if (dag_constraint) {
	free (dag_constraint);
    }
    dags.clear();
    
    job_num = 0; 
    dag_num = 0;

    
    // Set job to have low priority before infinite loop
    nice (20);    
    
    i = 0;
    // This loop will continuously create dags and submit them
    
    while (1) {
	// while (1) loop

	i++;
	writelog ("\n");
	
	char    dag_name[50];
	char *  dag_path = new char [strlen(dagpath)+52];
	
	//Preparing information to generate dag
	sprintf(dag_name, "dag_%d", i);
	
	sprintf (dag_path, "%s/%s", dagpath, dag_name);
	
	srand(time(NULL));

	attempt = 0;
	
	// start trying to remove dags once size reached 10
	if (dags.size()>= 10) {
	    removedags();
	}
	// Stop after max_dag_num of dags have been submitted
	if (dags.size()>=max_dag_num) {
	    printf("Number of dags, currently %d has reached the max number allowed\n",
		   dags.size ());
	    
	    while (!removedags()) {
		if (attempt >= MAX_ATTEMPT) {
		    printtime(runall_log);
		    fprintf(runall_log, "Failed to reduce directory size after 50 minutes, aborting\n");
		    break;
		}
		sleepfor (60);
		attempt++;
	    }
 	    if (attempt<50) {
		printf("Succeeded in removing some dags, continuing on\n");
		attempt = 0;
	    }
	    else {
		printtime (runall_log);
		fprintf(runall_log, "There are %d dags running at a time\n",
			max_dag_num); 
		break;
	    }
	}
	
	// Generate dags of size rand_num into directory dag_path
	rand_num = rand()%5;
	printtime(runall_log);
	fprintf(runall_log, "Generating dag %s\n", dag_path);

	int dag_gen_status = 0;
	
	
	if (INTELLIGENCE) {
	    // Look at current machine ads in the pool
	    // wait_time = 0
	    // loops = 1
	    pool->FindStaticAttributes (0, 1);
	    
	    dag_gen_status = GenerateDagFiles (dag_path, pool);
	    
	} else {

	    //cout<<"Generating dag files"<<endl;
	    if (job_num && (job_num%100 == 0) || (dag_num % 30 == 0)) {
		printtime(runall_log);
		fprintf(runall_log, " **** %d jobs were submitted in %d dags ****\n", job_num, dag_num);
	    }
	    dag_gen_status = GenerateDagFiles (dag_path, NULL);
	}	
	

      	if (!dag_gen_status) {
	    printtime(runall_log);
	    fprintf(runall_log, "Could not generate dag %s\n", dag_path);
	    printf("Could not generate dag %s\n", dag_path);
	    return -1;
	}
	// New dag is not created
	if (stat(dag_path, &stats)!=0) {
	    printtime(runall_log);
	    fprintf(runall_log, "%s was not created\n", dag_path);
	    printf("%s was not created\n", dag_path);
	    continue;
	}

	// Submit dag to condor 
	if (!submitdag(dag_path))
        {
	    printtime(runall_log);
	    fprintf(runall_log, "Could not submit dag %s\n", dag_name);
	    printf("Could not submit dag %s\n", dag_name);
	    fflush (runall_log);
	}
	
	// Add dag_path to data structure
	else 
        {
	    
	    char logname[strlen(dag_path)+30];
	    char line[1024];
	    char * _universe = param ("UNIVERSE");
	    if (_universe) {
		if (strcasecmp(_universe, "globus")) {
		    checked_condor_version = true; 
		    checked_proxy = true; 
		}
		free (_universe);
	    }

	    // Check if current condor version supports globus univ jobs
	    if (!checked_condor_version) {
		sleepfor(3);
		sprintf(logname, "%s/random.dag.dagman.out", dag_path);
		FILE * grep3; 
		char grep_version[strlen(logname)+70];
		sprintf(grep_version, "grep \"This version of Condor doesn't support Globus Universe jobs.\" %s", 
			logname);
		
		if ((grep3 = popen(grep_version, "r"))!=NULL) {
		    if (fgets(line, 1024, grep3)>0) {
			fflush (runall_log);
			pclose(grep3);
			fprintf(stderr, "ERROR: This version of Condor does not support globus jobs\n");
			exit (-1);
		    }
		    pclose(grep3);
		}
		checked_condor_version = true;
	    }

	    // check if proxy has been initialized 
	    if (!checked_proxy) {
		FILE * grep4; 
		char grep_proxy[strlen(logname)+27];
		sprintf(grep_proxy, "grep \"proxy has expired\" %s", logname);
		
		if ((grep4 = popen(grep_proxy, "r"))!=NULL) {
		    if (fgets(line, 1024, grep4)>0) {
			fflush (runall_log);
			pclose(grep4);
			fprintf(stderr, "ERROR: proxy has expired, please run grid-proxy-init\n");
			exit (-1);
		    }
		    pclose(grep4);
		} 
		checked_proxy = true;
	    }
	    // Update data structure
	    dag_info * newinfo = new dag_info();
	    newinfo->name = dag_path;
	    newinfo->status = status;
	    printtime(runall_log);
	    fprintf(runall_log, "Successfully submitted dag %s\n", dag_path);
	    dags.insert(newinfo);
	    printf("There are currently %d dags in progress, this number may include those failed dags\n",
		   dags.size());

	    assert (dags.size()!=0);
	}
        sleepfor (10);
	
    }
    
    return 0;
}

void printusage (char * argv0)
{
    cerr<<argv0<<" -- version "<<__TIME__ <<" "<<__DATE__<<endl;
    fprintf(stderr, "Usage: %s <dir> <config_file> <max_dag_num>\n", argv0);
    fprintf(stderr, "       <dir>             name of dir to place all the data files\n");
    fprintf(stderr, "       <config_file>     config file name for %s\n", argv0);
    fprintf(stderr, "       <max_dag_num>     number of dags that will be generated\n");
}


/* Add the local config file to the list of config files read by condor 
 */
bool setup_config (char * config_file_name, char * &old_conf) 
{
    char * path; // path for condor_config

    // check for the environ variable CONDOR_CONFIG
    path = getenv ("CONDOR_CONFIG");
    
    // if present, make a copy of the location 
    if (path) {
	old_conf = new char[strlen(path)+1];
	strncpy (old_conf, path, strlen(path)+1);
    } 
    // if not present, search the default paths 
    else {
	init_tilde();
	path = find_file("CONDOR_CONFIG", CONDOR_CONFIG);
    }

    if (!path) {
	fprintf(stderr, "Could not find %s\n", CONDOR_CONFIG);
	return false;
    }
    printf("Path for condor_config is %s\n", path);
    
    // set CONDOR_CONFIG environment variable to be in the current directory
    char cwd[1024];
    if (getcwd (cwd, 1024) == NULL) {
	perror ("could not get current directory");
	return false;
    }

    char local_condor_config [strlen(cwd)+strlen(CONDOR_CONFIG)+2];
    sprintf(local_condor_config, "%s/%s", cwd, CONDOR_CONFIG);

    if (setenv ("CONDOR_CONFIG", local_condor_config, 1) != 0) {
	perror ("Could not setenv CONDOR_CONFIG");
	return false;
    }
    
    // Copy the condor config file over 
    // Make changes to the local condor config directory
    // Need to add a local config file to condor
    FILE * config = fopen (path, "r");
    
    if (config == NULL) {
	perror ("Could not open condor_config file");
	return false;
    }
    
    FILE * local_config = fopen (local_condor_config, "w");
     
    if (local_config == NULL) {
	perror ("Could not open local condor_config file");
	return false;
    }

    char * buffer = new char[1024];
    
    char * macro_name = ", $(CONDOR_DAGMAN_TEST_CONFIG)";
	
    while (fgets (buffer, 1024, config)) {	

	if (strstr (buffer, "LOCAL_CONFIG_FILE")) {
	    
	    fprintf(local_config, "CONDOR_DAGMAN_TEST_CONFIG = %s\n\n", config_file_name);
	    
	    int buflen = strlen(buffer);
	    int macrolen = strlen(macro_name);
	    
	    char tempbuf [buflen+macrolen];

	    strncpy (tempbuf, buffer, buflen-1);

	    strncpy (tempbuf+buflen-1, macro_name, macrolen);

	    tempbuf[buflen+macrolen-1]='\0';
	    
	    fprintf(local_config, "%s\n", tempbuf);
	}
	else {
	    fprintf(local_config, "%s", buffer);
	}
	delete [] buffer;
	buffer = new char[1024];
	
    }
    delete [] buffer;
    
    
    if (local_config) 
	fclose(local_config);

    if (config)
	fclose (config);

    return true;
}

/* revert changes to  condor config */
bool revert_config_changes (char * old_env)
{
    // if old_env is NULL, remove it from ENV
    if (old_env == NULL) {
	unsetenv ("CONDOR_CONFIG");
}
    // if old_env is not NULL, change it back to old_env
    else {
	setenv ("CONDOR_CONFIG", old_env, 1);
    }
    // delete local copy of condor_config
    return true;
}

/* Given dag name, generate directory structure and dag*/
int GenerateDagFiles (const char * dag_path, PoolInfo * pool)
{
    char * jobname;
    int nodenum = 0; 

    // Setting number of nodes in the dag
    char * nodes = param ("DAGNODE_NUMBER"); 

    if (nodes) {
	nodenum = atoi (nodes);
	free (nodes);
    }

    if (nodenum <=0) {
	srand (time(NULL));
	// Randomly generate some parameters of the dag

	char * randnodes = param ("DAGNODE_MAX_NUMBER");
	int rand_num = 0;
	if (randnodes) {
	    rand_num = atoi (randnodes);
	    free (randnodes);
	}
	if (rand_num > 0) {
	    nodenum = min (rand_num, rand()%5+5);
	}
	else {
	    nodenum = rand()%30+5;
	}
    }
    
    printf("Generating dag with %d nodes in directory: %s\n", 
	   nodenum, dag_path);

    // Creating a new dag with nodenum nodes
    DDag * new_dag = new DDag(nodenum, dag_path, false);
    
    if (!new_dag) {
	printf("ERROR: Failed to generate dag files with %d nodes in directory: %s\n", 
	       nodenum, dag_path);
    }
    if (constraints) {
	cout<<"-------------- Initializing constraints for the dag"<<endl;
	constraints->InitializeForDag (nodenum);
    }

  
    jobname = param ("TEST_PROGRAM_NAME"); 
    if (jobname == NULL) {
	jobname = "testjob";
    }

    char * testmode = param ("TEST_MODE");
    
    if (!testmode) testmode = "normal";

    if (!strcasecmp (testmode, "md5")) {
	test_mode = MDSUM;
	new_dag->SetTestMode (MDSUM, jobname);
    } else if (!strcasecmp (testmode, "dap")) {
	test_mode = DAP;
	new_dag->SetTestMode (DAP, jobname);

    } else if (!strcasecmp (testmode, "longfile")) {
	test_mode = LONGFILE;
	new_dag->SetTestMode (LONGFILE, jobname);
    } else {
	test_mode = NORMAL;
	new_dag->SetTestMode (NORMAL, jobname);
    }
    
    if (dag_property) {
	new_dag->GenerateDagByLevels();
    } else {
	new_dag->GenerateDag ();
    }
    printf("Generating Dag File\n");
    if (!new_dag->GenerateDagFile()) {
	fprintf(stderr, "Genereate dag files failed \n");
	return 0;
    }

    //If the new directory was created successfully, cpy necessary files    
    struct stat stats;
    if (stat(dag_path, &stats)!=0) {
	printf("Could not stat %s\n", dag_path);
	return 0;
    }

    // Path of the test program to copy from 
    char * job_path = param ("TEST_PROGRAM_PATH");
    
    // Decision on whether to copy test program 
    char * copy = param ("COPY_EXECUTABLE");

    
    // Copy executables to the directory if it's in /scratch
    if (copy && !strcasecmp (copy, "true")) {
	int len;
	if (job_path) {
	    len = strlen(job_path)+strlen(jobname)+strlen(dag_path)+10;
	} else {
	    len = strlen(jobname)+strlen(dag_path)+15;
	}
	char command[len];
	if (job_path) {
	    sprintf(command, "cp %s/%s %s/.", job_path, jobname, dag_path);
	    printf("echo command: %s\n", command);
	    
	}
	else {
	    sprintf(command, "cp %s %s/.", jobname, dag_path);
	}
	if (system (command)!=0) {
	    printf("Could not do command %s\n");
	    if (jobname) {
		free (jobname);
	    }
	    return -1;
	}
    }
    
    
    // Generate dag submit files
    if (new_dag->GenerateSubmitFiles(pool))
	cout<<"Great! Finished generating dagman and condor submit files"<<endl<<endl;
    else {
	cerr<<"Could not generate submit files successfully"<<endl;
	return 0;
    }
    
    dag_num ++;
    job_num += nodenum; 

    if (jobname) {
	free (jobname);
    }
    delete new_dag;
    return nodenum;
}


bool submitdag(const char* dagpath)
{
    //writelog ("Submitting dag ... \n");
    char * condor_submit_dag = param ("CONDOR_SUBMIT_DAG");
    
    if (!condor_submit_dag) {
	condor_submit_dag = "condor_submit_dag";
    }
    if (test_mode == DAP) {
	condor_submit_dag = "/scratch/ningl/dap/condor_submit_dag -daplog ../../dap_log/dap.log -dapserver ophelia";
    }

    char path  [1024];
    getcwd (path, 1024);
    if (!path) {
	printf("ERROR: could not get current working directory\n");
	return false;
    }

    if (chdir (dagpath))
    {
	printtime(runall_log);
	fprintf(runall_log, "Could not cd %s\n", dagpath);
	printf("Could not cd %s\n", dagpath);
	return false;
    }
    
    pid_t child = fork ();
    if (child ==-1) {
	printtime(runall_log);
	fprintf(runall_log, "Could not submit dag in directory %s\n", dagpath);
	printf("Could not submit dag in directory %s\n", dagpath);
	return false;
    }
    if (child == 0) {
        writelog(condor_submit_dag);
	if (execlp(condor_submit_dag, condor_submit_dag,
		   "-notification", "error", "random.dag", NULL)==-1) {
	    printtime(runall_log);
	    fprintf(runall_log, "Could not execute condor_submit_dag random.dag from %s\n",
		    dagpath);
	    printf("Could not execute condor_submit_dag random.dag from %s\n",
		   dagpath);
	    return -1;
	}
    }
    
    wait (NULL);

    if (condor_submit_dag) {
	free (condor_submit_dag);
    }

    if (chdir (path)) {
        printtime(runall_log);
	fprintf(runall_log, "Could not cd back to %s\n", path);
	printf("Could not cd back to %s\n", path);
	exit (-1);
    }
    return true;
}

/* Clean contents of a dag that resides in the directory dag_path*/
bool 
clean (const char* dag_path)
{  
    int attempt = 0;
    struct stat mystat;

    char command [strlen(dag_path)+20];
    sprintf(command, "/bin/rm -rf %s", dag_path);
    
    while (attempt < 3 && (stat(dag_path, &mystat) != -1)) {
	system (command);
	attempt++;
    }
    printtime(runall_log);
    if (attempt < 3) {
	fprintf(runall_log, "Removed directory %s\n", dag_path);
    }
    else {
	fprintf(runall_log, "Could not remove directory %s\n", dag_path);
    }
    return true;
}

/* Decides whether dag can be removed from directory dag_path 
   Checks to see if there's an error in the dag execution 
   If there is returns false - dag will stay in the directory
   otherwise, dag can be marked for removal
*/
bool 
remove_dag (const char * dag_path, const bool status)
{
    if (!dag_path) {
	printf("ERROR: dag_path is null in remove_dag\n");
	return false;
    }
    char logname[strlen(dag_path)+30];
    sprintf(logname, "%s/random.dag.dagman.out", dag_path);
    
    struct stat mystats;

    int namelen = strlen(logname);
    char command0[namelen+20];
    char command1[namelen+16];
    char command2[namelen+25];
    char command3[namelen+70];
    char command4[namelen+27];

    sprintf(command0, "grep \"cycle exists\" %s", logname);
    sprintf(command1, "grep \"ERROR\" %s", logname);
    sprintf(command2, "grep \"All jobs Completed!\" %s", logname);
    sprintf(command3, "grep \"This version of Condor doesn't support Globus Universe jobs.\" %s", logname);
    sprintf(command4, "grep \"proxy has expired\" %s", logname);

    if (stat(logname, &mystats)!=0) {
	printtime(runall_log);
	fprintf (runall_log, "ERROR: could not open %s\n", logname);
	fflush (runall_log);
	return false;
    }

    FILE * grep0, * grep1, * grep2, *grep3, *grep4;
    char line[1024];

    if (status == FAIL) {
	if ((grep0=popen(command0, "r"))!=NULL) {
	    if (fgets(line, 1024, grep0)>0) {
		printtime(runall_log);
		fprintf(runall_log, "%s exited with cycle as expected\n", dag_path);
		pclose(grep0);
		return true;
	    }
	    pclose(grep0);
	}
    }
    if ((grep1=popen(command1, "r"))!=NULL) 
    {
	if (fgets(line, 1024, grep1)>0) {
	    printtime(runall_log);
	    fprintf(runall_log, "%s exited with error\n", dag_path);
	    pclose(grep1);
	    return false;
	}
	pclose(grep1);
    }
    if ((grep2= popen(command2, "r"))!=NULL)
    {
	if (fgets(line, 1024, grep2)>0) {
	    fflush (runall_log);
	    pclose(grep2);
	    return (process_log (dag_path));
	}
	pclose(grep2);
    } 

    if ((grep3 = popen(command3, "r"))!=NULL)
    {
	if (fgets(line, 1024, grep3)>0) {
	    fflush (runall_log);
	    pclose(grep3);
	    fprintf(stderr, "ERROR: This version of Condor does not support globus jobs\n");
	    exit (-1);
	}
	pclose(grep3);
    } 
    if ((grep4 = popen(command4, "r"))!=NULL)
    {
	if (fgets(line, 1024, grep4)>0) {
	    fflush (runall_log);
	    pclose(grep4);
	    fprintf(stderr, "ERROR: proxy has expired, please run grid-proxy-init\n");
	    exit (-1);
	}
	pclose(grep4);
    } 
    return false;
}


/* This function reconstructs dag as read from the dag description
   file in the directory dag_path & check that everything is in place:
   eg: all output files exist and all output files are generated in 
       the correct order
*/
bool 
process_log (const char * dag_path) 
{
    cout<<"Process_log at "<<dag_path<<endl;
    char * filename, * parentname, *job;	
    bool   bad = false, no_log = false;
    int    MAXLINE = 256, jobnum = 0; //number of jobs
    char   line [MAXLINE];
    char   parent [35];
    char   child [35];
    struct stat mystat, parentstat; 

    jobs_hash_map jobs;
    jobs_hash_map::iterator itr;
    
    set <DagNode*>::iterator parent_itr;	
    set <DagNode*> * parents;

    printtime (runall_log);
    fprintf(runall_log, "Processing log of %s\n", dag_path);

   // Open log files
    if (!LOG) {
	if ((LOG = fopen ("current_log.log", "w")) == NULL) {
	    perror("Error opening log");
	    printtime (runall_log);
	    fprintf(runall_log, "Could not open current_log.log for writing\n");
	    return false;
	}
    }
    if (!SUMMARY_LOG) {
	if ((SUMMARY_LOG = fopen ("summary.log", "w")) == NULL) {
	    perror("Error opening log");
	    printtime(runall_log);
	    fprintf(runall_log, "Could not open summary.log for writing\n");
	    return false;
	}
    }
    
    // Read in the random.dag file to count number of jobs
    char dagfilename [strlen(dag_path)+15];
    sprintf(dagfilename, "%s/random.dag", dag_path);
    
    printtime(runall_log);
    fprintf(runall_log, "Opening dag file: %s/random.dag\n", dag_path);

    FILE * dagfile = fopen (dagfilename, "r");
    
    if (dagfile == NULL) {
	perror ("Could not open dagfile\n");
	printtime(runall_log);
	printf("could not open dagfile: %s\n", dagfilename);
	fprintf(LOG, "could not open dagfile: %s\n", dagfilename);
	return false;
    }
    

    /*Read dag file line by line */
    while (fgets(line, MAXLINE, dagfile)!=NULL)
    {
	job = new char[30];	
	if (sscanf(line, "Job %s", job) !=0)
	{
	    jobnum ++; 
	}
	delete [] job;
    }
    
    //Create new dag if not done so already
    DDag newdag (jobnum, NULL, false);
    
    fprintf(runall_log, "There are %d jobs in %s\n", jobnum, dagfilename);
    
    cout<<"There are "<<jobnum<<" jobs"<<endl;

    /* establish relationships between nodes*/ 
    rewind (dagfile);
    while (fgets(line, MAXLINE, dagfile)!=NULL)
    {
	job = new char[30];
	if (sscanf(line, "Job %s", job) !=0)
	{
	    if (jobs.find(job)==jobs.end())
		jobs[job]=newdag.AddNodeToDag(job);
	}			
	if (sscanf(line, "PARENT %s CHILD %s", parent, child)!=0)
	{
	    //printf("Dependency: parent = %s, child = %s\n", parent, child);
	    
	    //Establish relationship between the parent and child nodes
	    if (jobs.find(parent)==jobs.end())
	    {
		fprintf(stderr, "Parent %s node not found!\n", parent);
		break;
	    }

	    if (jobs.find(child)==jobs.end())
	    {
		fprintf(stderr, "Child %s node not found!\n", child);
		break;
	    }
	    jobs[parent]->addChild (jobs[child]);
	    jobs[child]->addParent(jobs[parent]);
	    
	}
    }
    
    /* check if every job has been run - is there an output file 
       created by jobs? */
    fprintf(LOG, "Job");

    for (itr=jobs.begin(); itr!=jobs.end(); itr++) {
	no_log = false;

	DagNode * current = (*itr).second;
	filename = new char [strlen(dag_path)+30];
	sprintf(filename, "%s/%s.out", dag_path, current->GetName());
	
	if (stat(filename, &mystat) == -1) {
	    fprintf(LOG, " %s", current->GetName());
	    printf("Error in getting file stat of %s\n", filename);
	    no_log = true;
	}
	delete [] filename; 
	filename = NULL;
    }

    if (no_log) 
	fprintf(LOG, " some output files are missing\n"); 
    else 
	fprintf(LOG, " logs are all in the directory\n");
    
    /*Check every parent has run before every child 
      and wrote the correct output */	
    for (itr = jobs.begin(); itr!=jobs.end(); itr++) {
	DagNode * current = (*itr).second;
	
	parents = current->GetParents();
	if (parents->size()!=0) {
	    //Get current timestamp first 
	    filename = new char [strlen(dag_path)+30];
	    sprintf(filename, "%s/%s.out", dag_path, current->GetName());
	    if (stat(filename, &mystat) == -1) {
		fprintf(stderr, "Cannot stat file %s\n", filename);
		perror("because: ");
		fprintf(LOG, "Cannot find %s, current job failed\n", filename);
		bad = true; 
		delete [] filename; 
		break;
	    }
	    
	    time_t mytime = mystat.st_mtime; 
	    delete [] filename;
	    
	    
	    for (parent_itr=parents->begin(); parent_itr!=parents->end(); 
		 parent_itr++) 
	    {
		DagNode * parent = *parent_itr;
		
		parentname = new char [strlen(dag_path)+30];
		//Check the timestamp of parent comparing it with mine
		sprintf(parentname, "%s/%s.out", dag_path, parent->GetName());
		if (stat (parentname, &parentstat) == -1) 
		{	
	
		    fprintf(stderr, "Cannot stat file %s", parentname);
		    perror (": ");
		    fprintf(LOG, "Cannot stat file %s\n", parentname);
		    bad = true; 
		    delete [] parentname; 
		    break;
		}
		time_t parenttime = parentstat.st_mtime;
		delete [] parentname;
				
		//parent's file didn't get created before mine!
		if (mytime < parenttime)
	        {
		    fprintf(stderr, "Parent %s's output has been created after child %s's output!\n",
			    parent->GetName(), current->GetName());
		    fprintf(LOG, "Parent %s's output has been created after child %s's output!\n",
			    parent->GetName(), current->GetName());
		    
		    bad = true; 
		}
	    }		
	}
    }
    if (!bad) {
	
	printf ("Dags in %s ran successfully as expected.\n", dag_path);
    	fprintf(SUMMARY_LOG, "Dags in %s ran successfully as expected.\n", dag_path);
	
	// remove the directory
	fprintf(SUMMARY_LOG, "Removing directory %s\n", dag_path);
	sleepfor (10);  // wait 10 secs for condor clean up
	clean (dag_path);
    }
    else { 
	printf("The dag in %s failed\n", dagfilename);	
	fprintf(SUMMARY_LOG, "The dag in %s failed\n", dagfilename);
    }
    if (dagfile) 
	fclose (dagfile);
    
    return !bad;
}


/* Tries to remove dags returns true if any dag has been removed 
   or if no dags are running
*/
bool
removedags()
{
    bool deleted = false;
    writelog ("Starting removal of dags\n");
    set<dag_info *>::iterator itr=dags.begin(), itr_next;
        
    for (int i=0; i<dags.size(); i++) {
	itr_next = ++itr;
	itr--;
	if (itr == dags.end()) 
	    break;
	if (*itr) {
	    cout<<"Checking status of dag "<<(*itr)->name<<endl;
	    if (remove_dag((*itr)->name, (*itr)->status)) {
		delete *itr;
		deleted = true;
		dags.erase(itr);
	    }
	}	
    	itr = itr_next;
    }
    return (deleted || dags.size()==0);
}


void
writelog(const char* msg)
{
    //fprintf(stderr, "Writing log: %s\n", msg);
    printtime(runall_log);
    fprintf(runall_log, "%s\n", msg);
    fflush(runall_log);
}

void
printtime(FILE * log)
{
    time_t curtime = time (NULL);
    struct tm * timeptr = localtime(&curtime);
    
    asctime (timeptr);
    fprintf(log, "%02d/%02d %02d:%02d:%02d ", 
	    timeptr->tm_mon+1, 
	    timeptr->tm_mday,
	    timeptr->tm_hour,
	    timeptr->tm_min,
	    timeptr->tm_sec);
}

void 
sleepfor (int sec)
{
    for (int i=0; i<sec; i++) {
	sleep(1);
	fprintf(stderr, ".");
    }
    fprintf(stderr, "\n");
}

