# batch_test.pl - Condor Test Suite Batch Tester
#
# V2.0 / 2000-May-31 / Peter Couvares / pfc@cs.wisc.edu

require 5.0;
use FileHandle;
use POSIX "sys_wait_h";

# configuration options
$test_dir = ".";            # directory in which to find test programs
$test_timeout = 60*60*18;   # seconds until we decide a test is hung

# setup
STDOUT->autoflush();   # disable command buffering of stdout
STDERR->autoflush();   # disable command buffering of stderr
$num_success = 0;
$num_failed = 0;

#
# the args:
# -d[irectory <dir>: just test this directory
# -f[ile] <filename>: use this file as the list of tests to run
# -s[kip] <filename>: use this file as the list of tests to skip
#
while( $_ = shift( @ARGV ) ) {
  SWITCH: {
	if( /-d.*/ ) {
		push(@compilers, shift(@ARGV));
		last SWITCH;
	}
	if( /-f.*/ ) {
		$testfile = shift(@ARGV);
		last SWITCH;
	}
	if( /-s.*/ ) {
		$skipfile = shift(@ARGV);
		last SWITCH;
	}
  }
}

# find all compiler subdirectories in the test directory
if( ! @compilers ) {
	opendir( TEST_DIR, $test_dir ) || die "error opening \"$test_dir\": $!\n";
	@compilers = grep -d, readdir( TEST_DIR );
	# filter out . and ..
	@compilers = grep !/^\.\.?$/, @compilers;
	closedir TEST_DIR;
	die "error: no compiler subdirectories found\n" unless @compilers;
}

%test_suite = ();

# now let's grab the tests. If we weren't given a file to do, open up all the
# directories we know about and grab everything out of it.
if( ! $testfile ) {

	# find all test programs (all files ending in .run) for each compiler
	foreach $compiler (@compilers)
	{
   		opendir( COMPILER_DIR, $compiler )
		|| die "error opening \"$compiler\": $!\n";
		@{$test_suite{"$compiler"}} = grep /\.run$/, readdir( COMPILER_DIR );
   		closedir COMPILER_DIR;
		die "error: no test programs found for $compiler\n" unless @{$test_suite{"$compiler"}};
	}
}

# if we were given a file, let's read it in and use it.
else {
	open(TESTFILE, $testfile) || die "Can't open $testfile\n";
	while(<TESTFILE>) {
		chomp();
		($compiler, $test) = split('\/');
		push(@{$test_suite{"$compiler"}}, $test);
	}
}


# if we were given a skip file, let's read it in and use it.
if ($skipfile) {
	open(SKIPFILE, $skipfile) || die "Can't open $skipfile\n";
	while(<SKIPFILE>) {
		chomp();
		($compiler, $test) = split('\/');
		$skip_hash{"$compiler"}->{"$test"} = 1;
	}
	close(SKIPFILE);
}

# Now we'll run each test. 
foreach $compiler (@compilers)
{
    chdir $compiler || die "error switching to directory $compiler: $!\n";

    # fork a child to run each test program
    print "submitting $compiler tests";
    foreach $test_program (@{$test_suite{"$compiler"}})
    {
	print ".";
	next if $skip_hash{$compiler}->{$test_program};
	$pid = fork();
	die "error calling fork(): $!\n" unless defined $pid;

	# if we're the parent, record the child's pid and go on
	if( $pid > 0 )
	{
	    $test{$pid} = "$test_program";
	    # wait a moment so as not to overwhelm condor_submit
	    sleep 5;
	    next;
	}

	# if we're the child, start test program
	
	# kill ourselves with SIGALARM if we run for more than 18 hours
	alarm $test_timeout;

	# exec the test itself
	exec( "perl $test_program > $test_program.out 2>&1" );
    }

	# wait for each test to finish and print outcome
	print "\n";
	while( $child = wait() )
	{
    	# if there are no more children, we're done
    	last if $child == -1;

    	# record the child's return status
    	$status = $?;

    	($test_name) = $test{$child} =~ /(.*)\.run$/;
    	printf( "%-18s ", $test_name );

    	if( WIFEXITED( $status ) && WEXITSTATUS( $status ) == 0 )
    	{
		print "succeeded\n";
		$num_success++;
		@successful_tests = (@successful_tests, "$compiler/$test_name");
    	}
    	else
    	{
		$failure = `grep 'FAILURE' $test{$child}.out`;
		$failure =~ s/^.*FAILURE[: ]//;
		chomp $failure;
		$failure = "failed" if $failure =~ /^\s*$/;
		print "$failure\n";
		$num_failed++;
		@failed_tests = (@failed_tests, "$compiler/$test_name");
    	}
	}

    print "\n";
    chdir ".." || die "error switching to directory ..: $!\n";
}

print "$num_success successful, $num_failed failed\n";

open( OUTF, ">successful_tests" )
    || die "error opening \"successful_tests\": $!\n";
for $test_name (@successful_tests)
{
    print OUTF "$test_name\n";
}
close OUTF;

open( OUTF, ">failed_tests" )
    || die "error opening \"failed_tests\": $!\n";
for $test_name (@failed_tests)
{
    print OUTF "$test_name\n";
}
close OUTF;

exit $num_failed;
