#include "condor_common.h"
#include "gmr.h"

GMR::GMR( )
{
    parentAd = NULL;
    key = -1;
}


GMR::~GMR( )
{
    if( parentAd ) {
        delete parentAd;
    }
#if defined(DEBUG)
    printf( "Deleted GMR %d (in dtor)\n", key );
#endif
}


GMR *
GMR::MakeGMR( int key, ClassAd *ad )
{
    const ExprTree          *expr;
    ClassAd                 *portAd;
    const ExprList         	*el;
    ExprListIterator        eli;
    References::iterator  	rit;
    Value                   val;
    GMR::Port               port;
    ClassAd::iterator       citr;
    string                  attr;
    Ports::iterator         portItr;
	References				dependencies;
    GMR                     *gmr = new GMR( );

    if( !gmr ) return( NULL );
    gmr->parentAd = ad;

        // get the port list
    if( !ad->EvaluateAttr( "Ports", val ) ) {
        printf( "Failed to get 'Ports' attribute: %s\n", CondorErrMsg.c_str());
        delete gmr;
        return( NULL );
    }
    if( !val.IsListValue( el ) ) {
        printf( "'Ports' attribute not a list\n" );
        delete gmr;
        return( NULL );
    }

        // insert basic port info
    for( eli.Initialize( el ) ; !eli.IsAfterLast( ); eli.NextExpr( ) ) {
        if(!(expr=eli.CurrentExpr())||expr->GetKind()!=ExprTree::CLASSAD_NODE){
            printf( "Port is not a classad\n" );
            delete gmr;
            return( NULL );
        }
        portAd = (ClassAd*)expr;

        port.portAd = portAd;
        if( !portAd->EvaluateAttrString( "label", port.label ) ) {
            printf( "Bad or missing port label\n" );
            delete gmr;
            return( NULL );
        }

            // check if this port has already been declared
        for( portItr=gmr->ports.begin(); portItr!=gmr->ports.end(); portItr++ ){
            if( strcasecmp( portItr->label.c_str( ), port.label.c_str( ) )==0 ){
                printf( "Port %s already declared\n", port.label.c_str( ) );
                delete gmr;
                return( NULL );

           }
        }

            // get external references of all expressions in port ad
		dependencies.clear( );
        for( citr = portAd->begin( ); citr != portAd->end( ); citr++ ) {
			attr = citr->first;
			expr = citr->second;
            if( !ad->GetExternalReferences( expr, dependencies ) ) {
                printf( "Failed to get external refs for %s in port %s\n",
                    attr.c_str( ), port.label.c_str( ) );
                delete gmr;
                return( NULL );
            }
        }

            // make sure that the dependencies are not forwards
        for( rit=dependencies.begin(); rit!=dependencies.end(); rit++ ) {
                // dependency on self?
            if( strcasecmp( rit->c_str( ), port.label.c_str( ) ) == MATCH ) {
                continue;
            }
                // must be dependent on some previous port
            for(portItr=gmr->ports.begin();portItr!=gmr->ports.end();portItr++){
                if( strcasecmp( rit->c_str( ), portItr->label.c_str( ) )==0 ) {
						// insert port number into dependency set
					port.dependencies.insert( portItr-gmr->ports.begin() );
                    break;
                }
            }
            if( portItr == gmr->ports.end( ) ) {
                printf( "Illegal reference/dependency %s in port %s\n",
                    rit->c_str( ), port.label.c_str( ) );
                delete gmr;
                return( NULL );
            }
        }

            // stuff port into port list
        gmr->ports.push_back( port );
    }

    gmr->key = key;
    //printf( "Created GMR with key %d\n", key );
    return( gmr );
}


bool
GMR::GetExternalReferences( References &refs )
{
    ClassAd         ctx, *pctx;
	ClassAd::iterator	caitr;
    PortReferences  prefs;
    References      tmp;
    Ports::iterator itr;
    ExprTree        *tree;
	string			attr;

    ctx.Insert( "_ad", parentAd );
    for( itr = ports.begin( ); itr != ports.end( ); itr++ ) {
            // insert dummy context and register as a valid target scope
        pctx = new ClassAd( );
        ctx.Insert( itr->label, pctx );
        prefs[pctx] = tmp;

            // find external references of port
		for( caitr=itr->portAd->begin( ); caitr!=itr->portAd->end( ); caitr++){
			attr = caitr->first;
			tree = caitr->second;
        	if( !( itr->portAd->GetExternalReferences( tree, prefs ) ) ) {
            	ctx.Remove( "_ad" );
            	return( false );
        	}
		}			
    }

        // accumulate all external references in refs
    PortReferences::iterator    prItr;
    References::iterator        rItr;
    for( prItr = prefs.begin( ); prItr != prefs.end( ); prItr++ ) {
        for( rItr=prItr->second.begin( ); rItr!=prItr->second.end( ); rItr++ ){
            refs.insert( *rItr );
        }
    }

    ctx.Remove( "_ad" );
    return( true );
}

bool
GMR::AddRectangles( Rectangles &r )
{
    Ports::iterator     pitr;

    r.NewClassAd( );
    for( pitr = ports.begin( ); pitr != ports.end( ); pitr++ ) {
        r.NewPort( );
		if( !r.AddRectangles( pitr->portAd, pitr->label, true ) ) {
			return( false );
		}
    }

    return( true );
}

bool
GMR::AddRectangles( int portNum, Rectangles &r )
{
	return( AddRectangles( portNum, portNum, r ) );
}


bool
GMR::AddRectangles( int oPortNum, int dPortNum, Rectangles &r )
{
    Port		*oPort, *dPort;

	if( ( oPortNum < 0 || (unsigned)oPortNum >= ports.size( ) ) || 
		( dPortNum < 0 || (unsigned)dPortNum >= ports.size( ) ) ) {
		return( false );
	}
	oPort = &(ports[oPortNum]);
	dPort = &(ports[dPortNum]);

	r.NewClassAd( );
	r.NewPort( );
	if( !r.AddRectangles( dPort->portAd, oPort->label, false ) ) {
        return( false );
    }

    return( true );
}


void
GMR::Display( FILE *fp )
{
    Ports::iterator         itr;
    Dependencies::iterator  dit;

    for( itr = ports.begin( ); itr != ports.end( ); itr++ ) {
        fprintf( fp, "Port %s: Depends on ", itr->label.c_str( ) );
        for(dit=itr->dependencies.begin();dit!=itr->dependencies.end();dit++){
            fprintf( fp, " %d", *dit );
        }
        putc( '\n', fp );
    }
}
